package espresso;

import java.util.Arrays;

final class Jcofactor {

/*
    The cofactor of a cover against a cube "c" is a cover formed by the
    cofactor of each cube in the cover against c.  The cofactor of two
    cubes is null if they are distance 1 or more apart.  If they are
    distance zero apart, the cofactor is the restriction of the cube
    to the minterms of c.

    The cube list contains the following information:

	T[0] = pointer to a cube identifying the variables that have
		been cofactored against
	T[1] = pointer to just beyond the sentinel (i.e., T[n] in this case)
	T[2]
	  .
	  .  = pointers to cubes
	  .
	T[n-2]
	T[n-1] = NULL pointer (sentinel)


    Cofactoring involves repeated application of "cdist0" to check if a
    cube of the cover intersects the cofactored cube.  This can be
    slow, especially for the recursive descent of the espresso
    routines.  Therefore, a special cofactor routine "scofactor" is
    provided which assumes the cofactor is only in a single variable.
*/


	/** cofactor -- compute the cofactor of a cover with respect to a cube */
	static Jset[] cofactor(final Jset[] T, final int[] c, final int off_c) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] temp = cube.temp[0];

		final int listlen = Jespresso.CUBELISTSIZE( T ) + 5;

		/* Allocate a new list of cube pointers (max size is previous size) */
		final Jset[] Tc_save = new Jset[ listlen ];
		int Tc = 0;

		/* pass on which variables have been cofactored against */
		Jset.set_diff( temp, cube.fullset, c, off_c );
		Tc_save[ Tc++ ] = new Jset( Jset.set_or( Jespresso.new_cube(), T[0].base, T[0].offset, temp ), 0 );
		Tc++;

		/* Loop for each cube in the list, determine suitability, and save */
		Jset p;
false_label:
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			if( p.base != c || p.offset != off_c ) {// java p != c

if( Jespresso.NO_INLINE ) {
				if( ! Jsetc.cdist0( p.base, p.offset, c, off_c ) ) {
					continue;
				}
} else {
				int last;
				if( (last = cube.inword) != -1 ) {
					int x = p.get( last ) & c[ off_c + last ];
					if( (~(x | x >>> 1) & cube.inmask) != 0 ) {
						continue;
					}
					for( int w = 1; w < last; w++ ) {
						x = p.get( w ) & c[ off_c + w ];
						if( (~(x | x >>> 1) & Jespresso.DISJOINT) != 0 ) {
							continue false_label;
						}
					}
				}
nextvar_label:
				for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
					final int[] mask = cube.var_mask[ var ];
					last = cube.last_word[ var ];
					for( int w = cube.first_word[ var ], ci = off_c + w; w <= last; w++, ci++ ) {
						if( (p.get( w ) & c[ ci ] & mask[ w ]) != 0 ) {
							continue nextvar_label;
						}
					}
					continue false_label;
// nextvar:			;
				}
}
				Tc_save[ Tc++ ] = p;
	// false:		;
			}
		}
		Tc_save[ Tc++ ] = null;/* sentinel */
		Tc_save[ 1 ] = new Jset( null, Tc );/* save pointer to last */
		return Tc_save;
	}

	/**
	scofactor -- compute the cofactor of a cover with respect to a cube,
	where the cube is "active" in only a single variable.

	This routine has been optimized for speed.
	*/
	static Jset[] scofactor(final Jset[] T, final int[] c, final int var) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] mask = cube.temp[ 1 ];
		final int first = cube.first_word[ var ], last = cube.last_word[ var ];

		final int listlen = Jespresso.CUBELISTSIZE( T ) + 5;

		/* Allocate a new list of cube pointers (max size is previous size) */
		final Jset[] Tc_save = new Jset[ listlen ];
		int Tc = 0;// java Tc_save[ Tc ]

		/* pass on which variables have been cofactored against */
		Jset.set_diff( mask, cube.fullset, c );
		Tc_save[ Tc++ ] = new Jset( Jset.set_or( Jespresso.new_cube(), T[ 0 ].base, T[ 0 ].offset, mask ), 0 );
		Tc++;

		/* Setup for the quick distance check */
		Jset.set_and( mask, cube.var_mask[ var ], c );

		/* Loop for each cube in the list, determine suitability, and save */
		Jset p;
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			if( p.base != c || p.offset != 0 ) {// FIXME is always != ? if( p != c ) {
				int i = first;
				do {
					if( (p.get( i ) & mask[ i ]) != 0 ) {
						Tc_save[ Tc++ ] = p;
						break;
					}
				} while( ++i <= last );
			}
		}

		Tc_save[ Tc++ ] = null;/* sentinel */
		Tc_save[ 1 ] = new Jset( null, Tc );/* save pointer to last */
		return Tc_save;
	}

	static void massive_count(final Jset[] T) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] count = Jglobals.cdata.part_zeros;

		/* Clear the column counts (count of # zeros in each column) */
		for( int i = cube.size - 1; i >= 0; i-- ) {
			count[ i ] = 0;
		}

		/* Count the number of zeros in each column */
		Jset p;
		final Jset cof = T[0];
		final int[] full = cube.fullset;
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			for( int i = Jespresso.LOOP( p ); i > 0; i-- ) {
				int val;
				if( (val = full[i] & ~ (p.get( i ) | cof.get( i ))) != 0 ) {
					final int cnt = ((i - 1) << Jespresso.LOGBPI);// java count[ cnt ]
// if( Jespresso.BPI == 32 ) {
					if( (val & 0xFF000000) != 0 ) {
						if( (val & 0x80000000) != 0 ) {
							count[ cnt + 31 ]++;
						}
						if( (val & 0x40000000) != 0 ) {
							count[ cnt + 30 ]++;
						}
						if( (val & 0x20000000) != 0 ) {
							count[ cnt + 29 ]++;
						}
						if( (val & 0x10000000) != 0 ) {
							count[ cnt + 28 ]++;
						}
						if( (val & 0x08000000) != 0 ) {
							count[ cnt + 27 ]++;
						}
						if( (val & 0x04000000) != 0 ) {
							count[ cnt + 26 ]++;
						}
						if( (val & 0x02000000) != 0 ) {
							count[ cnt + 25 ]++;
						}
						if( (val & 0x01000000) != 0 ) {
							count[ cnt + 24 ]++;
						}
					}
					if( (val & 0x00FF0000) != 0 ) {
						if( (val & 0x00800000) != 0 ) {
							count[ cnt + 23 ]++;
						}
						if( (val & 0x00400000) != 0 ) {
							count[ cnt + 22 ]++;
						}
						if( (val & 0x00200000) != 0 ) {
							count[ cnt + 21 ]++;
						}
						if( (val & 0x00100000) != 0 ) {
							count[ cnt + 20 ]++;
						}
						if( (val & 0x00080000) != 0 ) {
							count[ cnt + 19 ]++;
						}
						if( (val & 0x00040000) != 0 ) {
							count[ cnt + 18 ]++;
						}
						if( (val & 0x00020000) != 0 ) {
							count[ cnt + 17 ]++;
						}
						if( (val & 0x00010000) != 0 ) {
							count[ cnt + 16 ]++;
						}
					}
// }// if( Jespresso.BPI == 32 )
					if( (val & 0xFF00) != 0 ) {
						if( (val & 0x8000) != 0 ) {
							count[ cnt + 15 ]++;
						}
						if( (val & 0x4000) != 0 ) {
							count[ cnt + 14 ]++;
						}
						if( (val & 0x2000) != 0 ) {
							count[ cnt + 13 ]++;
						}
						if( (val & 0x1000) != 0 ) {
							count[ cnt + 12 ]++;
						}
						if( (val & 0x0800) != 0 ) {
							count[ cnt + 11 ]++;
						}
						if( (val & 0x0400) != 0 ) {
							count[ cnt + 10 ]++;
						}
						if( (val & 0x0200) != 0 ) {
							count[ cnt + 9 ]++;
						}
						if( (val & 0x0100) != 0 ) {
							count[ cnt + 8 ]++;
						}
					}
					if( (val & 0x00FF) != 0 ) {
						if( (val & 0x0080) != 0 ) {
							count[ cnt + 7 ]++;
						}
						if( (val & 0x0040) != 0 ) {
							count[ cnt + 6 ]++;
						}
						if( (val & 0x0020) != 0 ) {
							count[ cnt + 5 ]++;
						}
						if( (val & 0x0010) != 0 ) {
							count[ cnt + 4 ]++;
						}
						if( (val & 0x0008) != 0 ) {
							count[ cnt + 3 ]++;
						}
						if( (val & 0x0004) != 0 ) {
							count[ cnt + 2 ]++;
						}
						if( (val & 0x0002) != 0 ) {
							count[ cnt + 1 ]++;
						}
						if( (val & 0x0001) != 0 ) {
							count[ cnt + 0 ]++;
						}
					}
				}
			}
		}

		/*
		* Perform counts for each variable:
		*    cdata.var_zeros[var] = number of zeros in the variable
		*    cdata.parts_active[var] = number of active parts for each variable
		*    cdata.vars_active = number of variables which are active
		*    cdata.vars_unate = number of variables which are active and unate
		*
		*    best -- the variable which is best for splitting based on:
		*    mostactive -- most # active parts in any variable
		*    mostzero -- most # zeros in any variable
		*    mostbalanced -- minimum over the maximum # zeros / part / variable
		*/
		final Jcdata_struct cdata = Jglobals.cdata;// java
		int lastbit, active, maxactive;
		int best = -1;
		int mostactive = 0, mostzero = 0;
		int mostbalanced = 32000;
		cdata.vars_unate = cdata.vars_active = 0;

		for( int var = 0, n = cube.num_vars; var < n; var++ ) {
			if( var < cube.num_binary_vars ) { /* special hack for binary vars */
				int i = (var << 1) + 1;
				lastbit = count[ i ];// (var << 1) + 1
				i = count[ --i ];// (var << 1)
				active = 0;
				if( i > 0 ) {
					active++;
				}
				if( lastbit > 0 ) {
					active++;
				}
				cdata.var_zeros[ var ] = i + lastbit;
				maxactive = Math.max( i, lastbit );
			} else {
				maxactive = active = cdata.var_zeros[ var ] = 0;
				lastbit = cube.last_part[ var ];
				for( int i = cube.first_part[ var ]; i <= lastbit; i++ ) {
					cdata.var_zeros[ var ] += count[ i ];
					if( count[ i ] > 0 ) {
						active++;
					}
					if( active > maxactive ) {
						maxactive = active;
					}
				}
			}

			/* first priority is to maximize the number of active parts */
			/* for binary case, this will usually select the output first */
			if( active > mostactive ) {
				best = var;
				mostactive = active;
				mostzero = cdata.var_zeros[ best ];
				mostbalanced = maxactive;
			} else if( active == mostactive ) {
				if( cdata.var_zeros[ var ] > mostzero ) {
					best = var;
					mostzero = cdata.var_zeros[ best ];
					mostbalanced = maxactive;
				} else if( cdata.var_zeros[ var ] == mostzero ) {
					if( maxactive < mostbalanced ) {
						best = var;
						mostbalanced = maxactive;
					}
				}
			}
			cdata.parts_active[ var ] = active;
			cdata.is_unate[ var ] = (active == 1);
			if( active > 0 ) {
				cdata.vars_active++;
			}
			if( active == 1 ) {
				cdata.vars_unate++;
			}
		}
		cdata.best = best;
	}

	@SuppressWarnings("boxing")
	static int binate_split_select(final Jset[] T, final int[] cleft, final int[] cright, final int debug_flag) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int best = Jglobals.cdata.best;
		final int lastbit = cube.last_part[ best ];
		int halfbit = 0;
		final Jset cof = T[0];

		/* Create the cubes to cofactor against */
		Jset.set_diff( cleft, cube.fullset, cube.var_mask[ best ] );
		Jset.set_diff( cright, cube.fullset, cube.var_mask[ best ] );
		int i;
		for( i = cube.first_part[ best ]; i <= lastbit; i++ ) {
			if( ! Jespresso.is_in_set( cof.base, cof.offset, i ) ) {
				halfbit++;
			}
		}
		for( i = cube.first_part[ best ], halfbit = halfbit >>> 1; halfbit > 0; i++ ) {
			if( ! Jespresso.is_in_set( cof.base, cof.offset, i ) ) {
				halfbit--;
				Jespresso.set_insert( cleft, i );
			}
		}
		for( ; i <= lastbit; i++ ) {
			if( ! Jespresso.is_in_set( cof.base, cof.offset, i ) ) {
				Jespresso.set_insert( cright, i );
			}
		}

		if( (Jglobals.debug & debug_flag) != 0 ) {
			System.out.printf("BINATE_SPLIT_SELECT: split against %d\n", best );
			if( Jglobals.verbose_debug ) {
				System.out.printf("cl=%s\ncr=%s\n", Jcvrout.pc1( cleft, 0 ), Jcvrout.pc2( cright ) );
			}
		}
		return best;
	}

	static Jset[] cube1list(final Jset_family A) {
		final Jset[] list = new Jset[ A.count + 3 ];
		int plist = 0;// java list[ plist ]
		list[ plist++ ] = new Jset( Jespresso.new_cube(), 0 );
		plist++;
		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			list[ plist++ ] = new Jset( data, p );
		}
		list[ plist++ ] = null;/* sentinel */
		list[ 1 ] = new Jset( null, plist );
		return list;
	}

	static Jset[] cube2list(final Jset_family A, final Jset_family B) {
		final Jset[] list = new Jset[ A.count + B.count + 3 ];
		int plist = 0;// java list[ plist ]
		list[ plist++ ] = new Jset( Jespresso.new_cube(), 0 );
		plist++;
		int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			list[ plist++ ] = new Jset( data, p );
		}
		data = B.data;// java
		for( int p = 0, last = p + B.count * B.wsize; p < last; p += B.wsize ) {
			list[ plist++ ] = new Jset( data, p );
		}
		list[ plist++ ] = null;
		list[ 1 ] = new Jset( null, plist );
		return list;
	}

	static Jset[] cube3list(final Jset_family A, final Jset_family B, final Jset_family C) {
		final Jset[] list = new Jset[ A.count + B.count + C.count + 3 ];
		int plist = 0;// java list[ plist ]
		list[ plist++ ] = new Jset( Jespresso.new_cube(), 0 );
		plist++;
		int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			list[ plist++ ] = new Jset( data, p );
		}
		data = B.data;// java
		for( int p = 0, last = p + B.count * B.wsize; p < last; p += B.wsize ) {
			list[ plist++ ] = new Jset( data, p );
		}
		data = C.data;// java
		for( int p = 0, last = p + C.count * C.wsize; p < last; p += C.wsize ) {
			list[ plist++ ] = new Jset( data, p );
		}
		list[ plist++ ] = null;
		list[ 1 ] = new Jset( null, plist );
		return list;
	}

	static Jset_family cubeunlist(final Jset[] A1) {
		final Jset cof = A1[ 0 ];

		final Jset_family A = Jespresso.new_cover( Jespresso.CUBELISTSIZE( A1 ) );
		Jset p;
		for( int i = 2, ai = 0; (p = A1[ i ]) != null; i++, ai += A.wsize ) {
			Jset.set_or( A.data, ai, p.base, p.offset, cof.base, cof.offset );
		}
		A.count = Jespresso.CUBELISTSIZE( A1 );
		return A;
	}

	static void simplify_cubelist(final Jset[] T) {
		Jset.set_copy( Jglobals.cube.temp[0], T[0].base, T[0].offset );// retrieve cofactor

		final int ncubes = Jespresso.CUBELISTSIZE( T );
		// qsort( (char *) (T + 2), ncubes, sizeof(pset), d1_order );
		Arrays.sort( T, 2, ncubes + 2, new Jdistance_order() );

		int Tdest = 2;// java T[ Tdest ]
		/*   *Tdest++ = T[2]; */
		for( int i = 3; i < ncubes; i++ ) {
			if( Jdistance_order.d1_order( T[ i - 1 ], T[ i ] ) != 0 ) {
				T[ Tdest++ ] = T[ i ];
			}
		}

		T[ Tdest++ ] = null;// sentinel
		T[ Tdest + 1 ] = new Jset( null, Tdest );// save pointer to last
	}
}