package espresso;

import java.util.Arrays;
import java.util.Comparator;

/*
 *  module: compl.c
 *  purpose: compute the complement of a multiple-valued function
 *
 *  The "unate recursive paradigm" is used.  After a set of special
 *  cases are examined, the function is split on the "most active
 *  variable".  These two halves are complemented recursively, and then
 *  the results are merged.
 *
 *  Changes (from Version 2.1 to Version 2.2)
 *      1. Minor bug in compl_lifting -- cubes in the left half were
 *      not marked as active, so that when merging a leaf from the left
 *      hand side, the active flags were essentially random.  This led
 *      to minor impredictability problem, but never affected the
 *      accuracy of the results.
 */

final class Jcompl {

	private static final int USE_COMPL_LIFT               = 0;
	private static final int USE_COMPL_LIFT_ONSET         = 1;
	private static final int USE_COMPL_LIFT_ONSET_COMPLEX = 2;
	private static final int NO_LIFTING                   = 3;

	private static int compl_level = 0;

	/** complement -- compute the complement of T
	 * @param T T will be disposed of
	 */
	static Jset_family complement(final Jset[] T) {

		if( (Jglobals.debug & Jespresso.COMPL) != 0 ) {
			Jcvrout.debug_print( T, "COMPLEMENT", compl_level++ );
		}

		final Jset_family[] Tbar = new Jset_family[ 1 ];// java holder
		if( compl_special_cases( T, Tbar ) == Jespresso.MAYBE ) {

			/* Allocate space for the partition cubes */
			int[] cl = Jespresso.new_cube();
			int[] cr = Jespresso.new_cube();
			final int best = Jcofactor.binate_split_select( T, cl, cr, Jespresso.COMPL );

			/* Complement the left and right halves */
			final Jset_family Tl = complement( Jcofactor.scofactor( T, cl, best ) );
			final Jset_family Tr = complement( Jcofactor.scofactor( T, cr, best ) );

			int lifting;
			if( Tr.count * Tl.count > (Tr.count + Tl.count) * Jespresso.CUBELISTSIZE( T ) ) {
				lifting = USE_COMPL_LIFT_ONSET;
			} else {
				lifting = USE_COMPL_LIFT;
			}
			Tbar[ 0 ] = compl_merge( T, Tl, Tr, cl, cr, best, lifting );

			cl = null;
			cr = null;
			T[0] = null;
		}

		if( (Jglobals.debug & Jespresso.COMPL) != 0 ) {
			Jcvrout.debug1_print( Tbar[ 0 ], "exit COMPLEMENT", --compl_level );
		}
		return Tbar[ 0 ];
	}

	/**
	 * @param T will be disposed if answer is determined
	 * @param Tbar returned only if answer determined
	 * @return
	 */
	private static int compl_special_cases(final Jset[] T, final Jset_family[] Tbar) {
		final Jset cof = T[0];

		/* Check for no cubes in the cover */
		if( T[2] == null ) {
			Tbar[0] = Jset.sf_addset( Jespresso.new_cover( 1 ), Jglobals.cube.fullset );
			T[0] = null;
			return Jespresso.TRUE;
		}

		/* Check for only a single cube in the cover */
		if( T[3] == null ) {
			final Jset temp = Jset.set_or( cof, cof.base, cof.offset, T[2].base, T[2].offset );
			Tbar[0] = compl_cube( temp.base, temp.offset );
			T[0] = null;
			return Jespresso.TRUE;
		}

		/* Check for a row of all 1's (implies complement is null) */
		Jset p;
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			if( Jsetc.full_row( p, cof ) ) {
				Tbar[0] = Jespresso.new_cover( 0 );
				T[0] = null;
				return Jespresso.TRUE;
			}
		}

		/* Check for a column of all 0's which can be factored out */
		int[] ceil = Jset.set_save( cof );
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			Jset.set_or( ceil, ceil, p.base, p.offset );
		}
		if( ! Jset.setp_equal( ceil, Jglobals.cube.fullset ) ) {
			final Jset_family ceil_compl = compl_cube( ceil, 0 );
			Jset.set_diff( ceil, Jglobals.cube.fullset, ceil );
			Jset.set_or( cof.base, cof.offset, cof.base, cof.offset, ceil );
			ceil = null;
			Tbar[0] = Jset.sf_append( complement( T ), ceil_compl );
			return Jespresso.TRUE;
		}
		ceil = null;

		/* Collect column counts, determine unate variables, etc. */
		Jcofactor.massive_count( T );

		/* If single active variable not factored out above, then tautology ! */
		if( Jglobals.cdata.vars_active == 1 ) {
			Tbar[0] = Jespresso.new_cover( 0 );
			T[0] = null;
			return Jespresso.TRUE;

			/* Check for unate cover */
		} else if( Jglobals.cdata.vars_unate == Jglobals.cdata.vars_active ) {
			Jset_family A = Junate.map_cover_to_unate( T );
			T[0] = null;
			A = Junate.unate_compl( A );
			Tbar[0] = Junate.map_unate_to_cover( A );
			Jset.sf_free( A );
			return Jespresso.TRUE;

			/* Not much we can do about it */
		} else {
			return Jespresso.MAYBE;
		}
	}

	/**
	*  compl_merge -- merge the two cofactors around the splitting
	*  variable
	*
	*  The merge operation involves intersecting each cube of the left
	*  cofactor with cl, and intersecting each cube of the right cofactor
	*  with cr.  The union of these two covers is the merged result.
	*
	*  In order to reduce the number of cubes, a distance-1 merge is
	*  performed (note that two cubes can only combine distance-1 in the
	*  splitting variable).  Also, a simple expand is performed in the
	*  splitting variable (simple implies the covering check for the
	*  expansion is not full containment, but single-cube containment).
	*
	* @param T1 Original ON-set
	* @param L Complement from each recursion branch
	* @param R Complement from each recursion branch
	* @param cl cubes used for cofactoring
	* @param cr cubes used for cofactoring
	* @param var splitting variable
	* @param lifting whether to perform lifting or not
	*/
	@SuppressWarnings("boxing")
	private static Jset_family compl_merge(final Jset[] T1, final Jset_family L, final Jset_family R, final int[] cl, final int[] cr, final int var, final int lifting) {

		if( (Jglobals.debug & Jespresso.COMPL) != 0 ) {
			System.out.printf("compl_merge: left %d, right %d\n", L.count, R.count );
			System.out.printf("%s (cl)\n%s (cr)\nLeft is\n", Jcvrout.pc1( cl, 0 ), Jcvrout.pc2( cr ) );
			Jcvrout.cprint( L );
			System.out.printf("Right is\n");
			Jcvrout.cprint( R );
		}

		/* Intersect each cube with the cofactored cube */
		int[] data = L.data;// java
	    for( int p = 0, last = p + L.count * L.wsize; p < last; p += L.wsize ) {
			Jset.set_and( data, p, data, p, cl );
			Jespresso.SET( data, p, Jespresso.ACTIVE );
		}
	    data = R.data;// java
	    for( int p = 0, last = p + R.count * R.wsize; p < last; p += R.wsize ) {
			Jset.set_and( data, p, data, p, cr );
			Jespresso.SET( data, p, Jespresso.ACTIVE );
		}

		/* Sort the arrays for a distance-1 merge */
	    Jset.set_copy( Jglobals.cube.temp[0], Jglobals.cube.var_mask[var] );
		Jset[] L1, R1;
		final Comparator<Jset> comparator = new Jdistance_order();
		Arrays.sort( L1 = Jcontain.sf_list( L ), 0, L.count, comparator );
		Arrays.sort( R1 = Jcontain.sf_list( R ), 0, R.count, comparator );

		/* Perform distance-1 merge */
		compl_d1merge( L1, R1 );

		/* Perform lifting */
		switch( lifting ) {
		case USE_COMPL_LIFT_ONSET: {
			final Jset_family T = Jcofactor.cubeunlist( T1 );
			compl_lift_onset( L1, T, cr, var );
			compl_lift_onset( R1, T, cl, var );
			Jespresso.free_cover( T );
			break; }
		case USE_COMPL_LIFT_ONSET_COMPLEX: {
			final Jset_family T = Jcofactor.cubeunlist( T1 );
			compl_lift_onset_complex( L1, T, var );
			compl_lift_onset_complex( R1, T, var );
			Jespresso.free_cover( T );
			break; }
		case USE_COMPL_LIFT:
			compl_lift( L1, R1, cr, var );
			compl_lift( R1, L1, cl, var );
			break;
		case NO_LIFTING:
			break;
		}
		L1 = null;
		R1 = null;

		/* Re-create the merged cover */
		final Jset_family Tbar = Jespresso.new_cover( L.count + R.count );
		final int[] pt_base = Tbar.data;// java
		int pt = 0;// pt_base[ pt ]
		data = L.data;// java
	    for( int p = 0, last = p + L.count * L.wsize; p < last; p += L.wsize ) {
			Jset.set_copy( pt_base, pt, data, p );
			Tbar.count++;
			pt += Tbar.wsize;
		}
	    data = R.data;// java
	    for( int p = 0, last = p + R.count * R.wsize; p < last; p += R.wsize ) {
	    	if( Jespresso.TESTP( data, p, Jespresso.ACTIVE ) ) {
				Jset.set_copy( pt_base, pt, data, p );
				Tbar.count++;
				pt += Tbar.wsize;
	    	}
		}

		if( (Jglobals.debug & Jespresso.COMPL) != 0 ) {
			System.out.printf("Result %d\n", Tbar.count );
			if( Jglobals.verbose_debug ) {
				Jcvrout.cprint( Tbar );
			}
		}

		Jespresso.free_cover( L );
		Jespresso.free_cover( R );
		return Tbar;
	}

	/**
	*  compl_lift_simple -- expand in the splitting variable using single
	*  cube containment against the other recursion branch to check
	*  validity of the expansion, and expanding all (or none) of the
	*  splitting variable.
	*/
	private static void compl_lift(final Jset[] A1, final Jset[] B1, final int[] bcube, final int var) {
		final Jcube_struct cube = Jglobals.cube;// java

		Jset a, b;
		final int[] lift = cube.temp[4], liftor = cube.temp[5];
		final int[] mask = cube.var_mask[ var ];

		Jset.set_and( liftor, bcube, mask );

		/* for each cube in the first array ... */
		for( int pa = 0; (a = A1[ pa++ ]) != null; ) {
			if( Jespresso.TESTP( a, Jespresso.ACTIVE ) ) {

				/* create a lift of this cube in the merging coord */
				Jset.set_merge( lift, bcube, a.base, a.offset, mask );

				/* for each cube in the second array */
				for( int B2 = 0; (b = B1[ B2++ ]) != null; ) {
					if( ! Jset.setp_implies( lift, b.base, b.offset ) ) {
						continue;
					}
					/* cube of A1 was contained by some cube of B1, so raise */
					Jset.set_or( a.base, a.offset, a.base, a.offset, liftor );
					break;
				}
			}
		}
	}

	/**
	*  compl_lift_onset -- expand in the splitting variable using a
	*  distance-1 check against the original on-set; expand all (or
	*  none) of the splitting variable.  Each cube of A1 is expanded
	*  against the original on-set T.
	*/
	private static void compl_lift_onset(final Jset[] A1, final Jset_family T, final int[] bcube, final int var) {
		Jset a;
		final int[] lift = Jglobals.cube.temp[4], mask = Jglobals.cube.var_mask[var];

		/* for each active cube from one branch of the complement */
nolift_label:
		for( int pa = 0; (a = A1[ pa++ ]) != null; ) {
			if( Jespresso.TESTP( a, Jespresso.ACTIVE ) ) {

				/* create a lift of this cube in the merging coord */
				Jset.set_and( lift, bcube, mask );	/* isolate parts to raise */
				Jset.set_or( lift, a.base, a.offset, lift );	/* raise these parts in a */

				/* for each cube in the ON-set, check for intersection */
				final int[] data = T.data;// java
			    for( int p = 0, last = p + T.count * T.wsize; p < last; p += T.wsize ) {
					if( Jsetc.cdist0( data, p, lift, 0 ) ) {
						continue nolift_label;
					}
				}
				Jset.set_copy( a.base, a.offset, lift );		/* save the raising */
				Jespresso.SET( a, Jespresso.ACTIVE );
//	nolift : ;
			}
		}
	}

	/**
	*  compl_lift_complex -- expand in the splitting variable, but expand all
	*  parts which can possibly expand.
	*  T is the original ON-set
	*  A1 is either the left or right cofactor
	*
	* @param A1 array of pointers to new result
	* @param T original ON-set
	* @param var which variable we split on
	*/
	private static void compl_lift_onset_complex(final Jset[] A1, final Jset_family T, final int var) {
		final Jcube_struct cube = Jglobals.cube;// java
		Jset a;

		/* for each cube in the complement */
		int[] xlower = Jespresso.new_cube();
		for( int pa = 0; (a = A1[ pa++ ]) != null; ) {

			if( Jespresso.TESTP( a, Jespresso.ACTIVE ) ) {

				/* Find which parts of the splitting variable are forced low */
				Jset.set_clear( xlower, cube.size );

				final int[] data = T.data;// java
			    for( int p = 0, last = p + T.count * T.wsize; p < last; p += T.wsize ) {
			    	int dist;
					if( (dist = Jsetc.cdist01( data, p, a.base, a.offset )) < 2 ) {
						if( dist == 0 ) {
							Jcvrmisc.fatal("compl: ON-set and OFF-set are not orthogonal");
						} else {
							Jsetc.force_lower( xlower, 0, data, p, a.base, a.offset );
						}
					}
				}

				Jset.set_diff( xlower, cube.var_mask[ var ], xlower );
				Jset.set_or( a.base, a.offset, a.base, a.offset, xlower );
				xlower = null;
			}
		}
	}

	/**
	*  compl_d1merge -- distance-1 merge in the splitting variable
	*/
	private static void compl_d1merge(final Jset[] L1, final Jset[] R1) {
		int pL1 = 0, pR1 = 0;// java

		/* Find equal cubes between the two cofactors */
		for( Jset pl = L1[ 0 ], pr = R1[ 0 ]; (pl != null) && (pr != null); ) {
			switch( Jdistance_order.d1_order( L1[ pL1 ], R1[ pR1 ] ) ) {
				case 1:
					pr = R1[ ++pR1 ]; break;/* advance right pointer */
				case -1:
					pl = L1[ ++pL1 ]; break;/* advance left pointer */
				case 0:
					Jespresso.RESET( pr, Jespresso.ACTIVE );
					Jset.set_or( pl.base, pl.offset, pl.base, pl.offset, pr.base, pr.offset );
					pr = R1[ ++pR1 ];
			}
		}
	}

	/** compl_cube -- return the complement of a single cube (De Morgan's law) */
	private static Jset_family compl_cube(final int[] p, final int off) {
		final Jcube_struct cube = Jglobals.cube;// java

		final int[] diff = cube.temp[7];
		final int[] full = cube.fullset;

		/* Allocate worst-case size cover (to avoid checking overflow) */
		final Jset_family R = Jespresso.new_cover( cube.num_vars );

		/* Compute bit-wise complement of the cube */
		Jset.set_diff( diff, full, p, off );

		for( int var = 0; var < cube.num_vars; var++ ) {
			final int[] mask = cube.var_mask[ var ];
			/* If the bit-wise complement is not empty in var ... */
			if( ! Jset.setp_disjoint( diff, mask ) ) {
				Jset.set_merge( R.data, R.wsize * R.count++, diff, full, mask );
			}
		}
		return R;
	}

	private static int simp_comp_simplify_level = 0;

	/** simp_comp -- quick simplification of T
	 *
	 * @param T will be disposed of
	 */
	static void simp_comp(final Jset[] T, final Jset_family[] Tnew, final Jset_family[] Tbar) {

		if( (Jglobals.debug & Jespresso.COMPL) != 0 ) {
			Jcvrout.debug_print( T, "SIMPCOMP", simp_comp_simplify_level++ );
		}

		if( simp_comp_special_cases( T, Tnew, Tbar ) == Jespresso.MAYBE ) {

			/* Allocate space for the partition cubes */
			int[] cl = Jespresso.new_cube();
			int[] cr = Jespresso.new_cube();
			final int best = Jcofactor.binate_split_select( T, cl, cr, Jespresso.COMPL );

			/* Complement the left and right halves */
			final Jset_family[] Tl = new Jset_family[ 1 ], Tr = new Jset_family[ 1 ], Tlbar = new Jset_family[ 1 ], Trbar = new Jset_family[ 1 ];// java holders
			simp_comp( Jcofactor.scofactor( T, cl, best ), Tl, Tlbar );
			simp_comp( Jcofactor.scofactor( T, cr, best ), Tr, Trbar );

			int lifting = USE_COMPL_LIFT;
			Tnew[ 0 ] = compl_merge( T, Tl[ 0 ], Tr[ 0 ], cl, cr, best, lifting );

			lifting = USE_COMPL_LIFT;
			Tbar[ 0 ] = compl_merge( T, Tlbar[ 0 ], Trbar[ 0 ], cl, cr, best, lifting );

			/* All of this work for nothing ? Let's hope not ... */
			if( Tnew[ 0 ].count > Jespresso.CUBELISTSIZE( T ) ) {
				Jset.sf_free( Tnew[ 0 ] );
				Tnew[ 0 ] = Jcofactor.cubeunlist( T );
			}

			cl = null;
			cr = null;
			T[0] = null;
		}

		if( (Jglobals.debug & Jespresso.COMPL) != 0 ) {
			Jcvrout.debug1_print( Tnew[ 0 ], "exit SIMPCOMP (new)", simp_comp_simplify_level );
			Jcvrout.debug1_print( Tbar[ 0 ], "exit SIMPCOMP (compl)", simp_comp_simplify_level );
			simp_comp_simplify_level--;
		}
	}

	/**
	 * @param T will be disposed if answer is determined
	 * @param Tnew returned only if answer determined
	 * @param Tbar returned only if answer determined
	 * @return 3 state value
	 */
	private static int simp_comp_special_cases(final Jset[] T, final Jset_family[] Tnew, final Jset_family[] Tbar) {
		final Jcube_struct cube = Jglobals.cube;// java
		final Jset cof = T[ 0 ];

		/* Check for no cubes in the cover (function is empty) */
		if( T[2] == null ) {
			Tnew[ 0 ] = Jespresso.new_cover( 1 );
			Tbar[ 0 ] = Jset.sf_addset( Jespresso.new_cover( 1 ), cube.fullset );
			T[0] = null;
			return Jespresso.TRUE;
		}

		/* Check for only a single cube in the cover */
		if( T[3] == null ) {
			Jset.set_or( cof.base, cof.offset, cof.base, cof.offset, T[2].base, T[2].offset );
			Tnew[ 0 ] = Jset.sf_addset( Jespresso.new_cover( 1 ), cof.base, cof.offset );
			Tbar[ 0 ] = compl_cube( cof.base, cof.offset );
			T[0] = null;
			return Jespresso.TRUE;
		}

		/* Check for a row of all 1's (function is a tautology) */
		Jset p;
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			if( Jsetc.full_row( p, cof ) ) {
				Tnew[ 0 ] = Jset.sf_addset( Jespresso.new_cover( 1 ), cube.fullset );
				Tbar[ 0 ] = Jespresso.new_cover( 1 );
				T[0] = null;
				return Jespresso.TRUE;
			}
		}

		/* Check for a column of all 0's which can be factored out */
		int[] ceil = Jset.set_save( cof );
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			Jset.set_or( ceil, ceil, p.base, p.offset );
		}
		if( ! Jset.setp_equal( ceil, cube.fullset ) ) {
			int[] pc = Jespresso.new_cube();
			Jset.set_diff( pc, cube.fullset, ceil );
			Jset.set_or( cof.base, cof.offset, cof.base, cof.offset, pc );
			pc = null;
			simp_comp( T, Tnew, Tbar );

			/* Adjust the ON-set */
			final Jset_family A = Tnew[ 0 ];
			final int[] data = A.data;// java
			for( int pi = 0, last = pi + A.count * A.wsize; pi < last; pi += A.wsize ) {
				Jset.set_and( data, pi, data, pi, ceil );
			}

			/* Compute the new complement */
			Tbar[ 0 ] = Jset.sf_append( Tbar[ 0 ], compl_cube( ceil, 0 ) );
			ceil = null;
			return Jespresso.TRUE;
		}
		ceil = null;

		/* Collect column counts, determine unate variables, etc. */
		Jcofactor.massive_count( T );

		/* If single active variable not factored out above, then tautology ! */
		if( Jglobals.cdata.vars_active == 1 ) {
			Tnew[ 0 ] = Jset.sf_addset( Jespresso.new_cover( 1 ), cube.fullset );
			Tbar[ 0 ] = Jespresso.new_cover( 1 );
			T[0] = null;
			return Jespresso.TRUE;

		/* Check for unate cover */
		} else if( Jglobals.cdata.vars_unate == Jglobals.cdata.vars_active ) {
			/* Make the cover minimum by single-cube containment */
			Jset_family A = Jcofactor.cubeunlist( T );
			Tnew[ 0 ] = Jcontain.sf_contain( A );

			/* Now form a minimum representation of the complement */
			A = Junate.map_cover_to_unate( T );
			A = Junate.unate_compl( A );
			Tbar[ 0 ] = Junate.map_unate_to_cover( A );
			Jset.sf_free( A );
			T[0] = null;
			return Jespresso.TRUE;

			/* Not much we can do about it */
		} else {
			return Jespresso.MAYBE;
		}
	}

	private static int simplify_level = 0;

	/** simplify -- quick simplification of T
	 * @param T T will be disposed of
	 */
	static Jset_family simplify(final Jset[] T) {
		if( (Jglobals.debug & Jespresso.COMPL) != 0 ) {
			Jcvrout.debug_print( T, "SIMPLIFY", simplify_level++ );
		}

		final Jset_family[] Tbar = new Jset_family[ 0 ];// java helper
		if( simplify_special_cases( T, Tbar ) == Jespresso.MAYBE ) {

			/* Allocate space for the partition cubes */
			int[] cl = Jespresso.new_cube();
			int[] cr = Jespresso.new_cube();

			final int best = Jcofactor.binate_split_select( T, cl, cr, Jespresso.COMPL );

			/* Complement the left and right halves */
			final Jset_family Tl = simplify( Jcofactor.scofactor( T, cl, best ) );
			final Jset_family Tr = simplify( Jcofactor.scofactor( T, cr, best ) );

			final int lifting = USE_COMPL_LIFT;
			Tbar[ 0 ] = compl_merge( T, Tl, Tr, cl, cr, best, lifting );

			/* All of this work for nothing ? Let's hope not ... */
			if( Tbar[ 0 ].count > Jespresso.CUBELISTSIZE( T ) ) {
				Jset.sf_free( Tbar[ 0 ] );
				Tbar[ 0 ] = Jcofactor.cubeunlist( T );
			}

			cl = null;
			cr = null;
			T[0] = null;
		}

		if( (Jglobals.debug & Jespresso.COMPL) != 0 ) {
			Jcvrout.debug1_print( Tbar[ 0 ], "exit SIMPLIFY", --simplify_level );
		}
		return Tbar[ 0 ];
	}

	/**
	 * @param T will be disposed if answer is determined
	 * @param Tnew returned only if answer determined
	 * @return 3 state value
	 */
	private static int simplify_special_cases(final Jset[] T, final Jset_family[] Tnew) {
		final Jcube_struct cube = Jglobals.cube;// java
		final Jset cof = T[0];

		/* Check for no cubes in the cover */
		if( T[ 2 ] == null ) {
			Tnew[ 0 ] = Jespresso.new_cover( 0 );
			T[0] = null;
			return Jespresso.TRUE;
		}

		/* Check for only a single cube in the cover */
		if( T[ 3 ] == null ) {
			final Jset temp = Jset.set_or( cof, cof.base, cof.offset, T[2].base, T[2].offset );
			Tnew[ 0 ] = Jset.sf_addset( Jespresso.new_cover( 1 ), temp.base, temp.offset );
			T[0] = null;
			return Jespresso.TRUE;
		}

		/* Check for a row of all 1's (implies function is a tautology) */
		Jset p;
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			if( Jsetc.full_row( p, cof ) ) {
				Tnew[ 0 ] = Jset.sf_addset( Jespresso.new_cover( 1 ), cube.fullset );
				T[0] = null;
				return Jespresso.TRUE;
			}
		}

		/* Check for a column of all 0's which can be factored out */
		int[] ceil = Jset.set_save( cof );
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			Jset.set_or( ceil, ceil, p.base, p.offset );
		}
		if( ! Jset.setp_equal( ceil, cube.fullset ) ) {
			int[] pc = Jespresso.new_cube();
			Jset.set_diff( pc, cube.fullset, ceil );
			Jset.set_or( cof.base, cof.offset, cof.base, cof.offset, pc );
			pc = null;

			final Jset_family A = simplify( T );
			final int[] data = A.data;// java
		    for( int pi = 0, last = pi + A.count * A.wsize; pi < last; pi += A.wsize ) {
				Jset.set_and( data, pi, data, pi, ceil );
			}
			Tnew[ 0 ] = A;
			ceil = null;
			return Jespresso.TRUE;
		}
		ceil = null;

		/* Collect column counts, determine unate variables, etc. */
		Jcofactor.massive_count( T );

		/* If single active variable not factored out above, then tautology ! */
		if( Jglobals.cdata.vars_active == 1 ) {
			Tnew[ 0 ] = Jset.sf_addset( Jespresso.new_cover( 1 ), cube.fullset );
			T[ 0 ] = null;
			return Jespresso.TRUE;

			/* Check for unate cover */
		} else if( Jglobals.cdata.vars_unate == Jglobals.cdata.vars_active ) {
			final Jset_family A = Jcofactor.cubeunlist( T );
			Tnew[ 0 ] = Jcontain.sf_contain( A );
			T[ 0 ] = null;
			return Jespresso.TRUE;

			/* Not much we can do about it */
		} else {
			return Jespresso.MAYBE;
		}
	}
}