package espresso;

import java.util.Arrays;
import java.util.Comparator;

/**
    contain.c -- set containment routines

    These are complex routines for performing containment over a
    family of sets, but they have the advantage of being much faster
    than a straightforward n*n routine.

    First the cubes are sorted by size, and as a secondary key they are
    sorted so that if two cubes are equal they end up adjacent.  We can
    than quickly remove equal cubes from further consideration by
    comparing each cube to its neighbor.  Finally, because the cubes
    are sorted by size, we need only check cubes which are larger (or
    smaller) than a given cube for containment.
*/

final class Jcontain {

	/**
	 * sf_contain -- perform containment on a set family (delete sets which
	 * are contained by some larger set in the family).  No assumptions are
	 * made about A, and the result will be returned in decreasing order of
	 * set size.
	 *
	 * @param A disposes of A
	 */
	static Jset_family sf_contain(final Jset_family A) {
		final Comparator<Jset> descend = new Jdescend_order();
		final Jset[] A1 = sf_sort( A, descend );           /* sort into descending order */
		int cnt = rm_equal( A1, descend );        /* remove duplicates */
		cnt = rm_contain( A1 );               /* remove contained sets */
		final Jset_family R = sf_unlist( A1, cnt, A.sf_size ); /* recreate the set family */
		Jset.sf_free( A );
		return R;
	}

	/**
	 * sf_rev_contain -- perform containment on a set family (delete sets which
	 * contain some smaller set in the family).  No assumptions are made about
	 * A, and the result will be returned in increasing order of set size
	 *
	 *@param A disposes of A
	 */
	static Jset_family sf_rev_contain(final Jset_family A) {
		final Comparator<Jset> ascend = new Jascend_order();
		final Jset[] A1 = sf_sort( A, ascend );            /* sort into ascending order */
		int cnt = rm_equal( A1, ascend );         /* remove duplicates */
		cnt = rm_rev_contain( A1 );           /* remove containing sets */
		final Jset_family R = sf_unlist( A1, cnt, A.sf_size) ; /* recreate the set family */
		Jset.sf_free( A );
		return R;
	}

	/**
	 * sf_ind_contain -- perform containment on a set family (delete sets which
	 * are contained by some larger set in the family).  No assumptions are
	 * made about A, and the result will be returned in decreasing order of
	 * set size.  Also maintains a set of row_indices to track which rows
	 * disappear and how the rows end up permuted.
	 *
	 * @param A disposes of A
	 * @param row_indices updated with the new values
	 */
	static Jset_family sf_ind_contain(final Jset_family A, final int[] row_indices) {
		final Comparator<Jset> descend = new Jdescend_order();
		final Jset[] A1 = sf_sort( A, descend );           /* sort into descending order */
		int cnt = rm_equal( A1, descend );        /* remove duplicates */
		cnt = rm_contain( A1 );               /* remove contained sets */
		final Jset_family R = sf_ind_unlist( A1, cnt, A.sf_size, row_indices, A.data, 0 );// TODO java check the last parameter
		Jset.sf_free( A );
		return R;
	}


	/** sf_dupl -- delete duplicate sets in a set family
	 *
	 * @param A disposes of A
	 */
	static Jset_family sf_dupl(final Jset_family A) {
		final Comparator<Jset> descend = new Jdescend_order();
		final Jset[] A1 = sf_sort( A, descend );           /* sort the set family */
		final int cnt = rm_equal( A1, descend );        /* remove duplicates */
		final Jset_family R = sf_unlist( A1, cnt, A.sf_size ); /* recreate the set family */
		Jset.sf_free( A );
		return R;
	}


	/**
	 * sf_union -- form the contained union of two set families (delete
	 * sets which are contained by some larger set in the family).  A and
	 * B are assumed already sorted in decreasing order of set size (and
	 * the SIZE field is assumed to contain the set size), and the result
	 * will be returned sorted likewise.
	 *
	 * @param A disposes of A and B
	 */
	static Jset_family sf_union(final Jset_family A, final Jset_family B) {
		final Comparator<Jset> descend = new Jdescend_order();
		final Jset[] A1 = sf_list( A ), B1 = sf_list( B );

		final Jset[] E1 = new Jset[ Math.max( A.count, B.count ) + 1 ];
		int cnt = rm2_equal( A1, B1, E1, descend );
		cnt += rm2_contain( A1, B1 ) + rm2_contain( B1, A1 );
		final Jset_family R = sf_merge( A1, B1, E1, cnt, A.sf_size );
		Jset.sf_free( A );
		Jset.sf_free( B );
		return R;
	}


	/**
	 * dist_merge -- consider all sets to be "or"-ed with "mask" and then
	 * delete duplicates from the set family.
	 *
	 * @param A disposes of A
	 * @param mask defines variables to mask out
	 */
	static Jset_family dist_merge(final Jset_family A, final int[] mask) {
		final Comparator<Jset> d1_order = new Jdistance_order();
		Jset.set_copy( Jglobals.cube.temp[0], mask );
		final Jset[] A1 = sf_sort( A, d1_order );
		final int cnt = d1_rm_equal( A1, d1_order );
		final Jset_family R = sf_unlist( A1, cnt, A.sf_size );
		Jset.sf_free( A );
		return R;
	}


	/**	d1merge -- perform an efficient distance-1 merge of cubes of A
	 *
	 * @param A disposes of A
	 */
	static Jset_family d1merge(final Jset_family A, final int var) {
		return dist_merge( A, Jglobals.cube.var_mask[ var ] );
	}

	/** d1_rm_equal -- distance-1 merge (merge cubes which are equal under a mask)
	 *
	 * @param A1 array of set pointers
	 * @param compare comparison function
	 */
	static int d1_rm_equal(final Jset[] A1, final Comparator<Jset> c) {
		int dest = 0;
		if( A1[0] != null ) {
			int i, j;
			for( i = 0, j = 1; A1[ j ] != null; j++ ) {
				final Jset ai = A1[ i ];
				final Jset aj = A1[ j ];
				if ( c.compare( ai, aj ) == 0 ) {
					/* if sets are equal (under the mask) merge them */
					Jset.set_or( ai.base, ai.offset, ai.base, ai.offset, aj.base, aj.offset );
				} else {
					/* sets are unequal, so save the set i */
					A1[ dest++ ] = ai;
					i = j;
				}
			}
			A1[ dest++ ] = A1[ i ];
		}
		A1[ dest ] = null;
		return dest;
	}

	/** rm_equal -- scan a sorted array of set pointers for duplicate sets
	 *
	 * @param A1 updated in place
	 * @param compare
	 */
	static int rm_equal(final Jset[] A1, final Comparator<Jset> c) {
		int pdest = 0;// java A1[ pdest ];

		if( A1[ 0 ] != null ) {/* If more than one set */
			int p;// java A1[ p ];
			Jset s, s1;
			for( p = 1; (s = A1[ p ]) != null; p++ ) {
				s1 = A1[ p - 1 ];
				if( c.compare( s, s1 ) != 0 ) {
					A1[ pdest++ ] = s1;
				}
			}
			A1[ pdest++ ] = A1[ p - 1 ];
			A1[ pdest ] = null;
		}
		return pdest;// pdest - A1;
	}

	/** rm_contain -- perform containment over a sorted array of set pointers
	 *
	 * @param A1 updated in place
	 */
	static int rm_contain(final Jset[] A1) {
		int pcheck = 0;// java to suppress the var may have been unitialized
		Jset a;
		int pdest = 0;// java A1[ pdest ];
		int last_size = -1;

		/* Loop for all cubes of A1 */
lnext1_label:
		for( int pa = 0; (a = A1[ pa++ ]) != null; ) {// pa = A1
			/* Update the check pointer if the size has changed */
			if( Jespresso.SIZE( a ) != last_size ) {
				last_size = Jespresso.SIZE( a );
				pcheck = pdest;
			}
			for( int pb = 0; pb != pcheck; ) {// pb = A1
				final Jset b = A1[ pb++ ];
				if( ! Jset.setp_implies( a.base, a.offset, b.base, b.offset ) ) {
					continue;
				}
				continue lnext1_label;
			}
			/* set a was not contained by some larger set, so save it */
			A1[ pdest++ ] = a;
//	lnext1: ;
		}

		A1[ pdest ] = null;
		return pdest;// pdest - A1;
	}

	/** rm_rev_contain -- perform rcontainment over a sorted array of set pointers
	 *
	 * @param A1 updated in place
	 */
	static int rm_rev_contain(final Jset[] A1) {
		Jset a;
		int pcheck = 0;// java: to suppress warning
		final Jset[] pdest_base = A1;
		int pdest = 0;// java pdest_base[ pdest ]
		int last_size = -1;

		/* Loop for all cubes of A1 */
next_label1:
		for( int pa = 0; (a = A1[ pa++ ]) != null; ) {
			/* Update the check pointer if the size has changed */
			final int t = Jespresso.SIZE( a.base, a.offset );
			if( t != last_size ) {
				last_size = t;
				pcheck = pdest;
			}
			for( int pb = 0; pb != pcheck; ) {
				final Jset b = A1[ pb++ ];
				if( ! Jset.setp_implies( b.base, b.offset, a.base, a.offset ) ) {
					continue;
				}
				continue next_label1;
			}
			/* the set a did not contain some smaller set, so save it */
			pdest_base[ pdest++ ] = a;
//	lnext1: ;
		}

		pdest_base[ pdest ] = null;
		return pdest;// pdest - A1;
	}

	/** rm2_equal -- check two sorted arrays of set pointers for equal cubes
	 *
	 * @param A1 updated in place
	 * @param B1 updated in place
	 */
	static int rm2_equal(final Jset[] A1, final Jset[] B1, final Jset[] E1, final Comparator<Jset> c) {
		int pa1 = 0, pb1 = 0;// java A1, B1 offsets
		int pda = 0, pdb = 0, pde = 0;// java int[][] pda = A1, pdb = B1, pde = E1;

		/* Walk through the arrays advancing pointer to larger cube */
		for( ; A1[ pa1 ] != null && B1[ pb1 ] != null; ) {
			switch( c.compare( A1[ pa1 ], B1[ pb1 ] ) ) {
			case -1:    /* "a" comes before "b" */
				A1[ pda++ ] = A1[ pa1++ ]; break;
			case 0:     /* equal cubes */
				E1[ pde++ ] = A1[ pa1++ ]; pb1++; break;
			case 1:     /* "a" is to follow "b" */
				B1[ pdb++ ] = B1[ pb1++ ]; break;
			}
		}

		/* Finish moving down the pointers of A and B */
		while( A1[ pa1 ] != null ) {
			A1[ pda++ ] = A1[ pa1++ ];
		}
		while( B1[ pb1 ] != null ) {
			B1[ pdb++ ] = B1[ pb1++ ];
		}
		A1[ pda ] = B1[ pdb ] = E1[ pde ] = null;

		return pde;// java pde - E1;
	}


	/** rm2_contain -- perform containment between two arrays of set pointers
	 *
	 * @param A1 updated in place
	 * @param B1 unchanged
	 */
	static int rm2_contain(final Jset[] A1, final Jset[] B1) {
		Jset a, b;
		int pdest = 0;// java int[][] pdest = A1;

		/* for each set in the first array ... */
next_label1:
		for( int pa = 0; (a = A1[ pa++ ]) != null; ) {
			/* for each set in the second array which is larger ... */
			for( int pb = 0; (b = B1[ pb++ ]) != null && Jespresso.SIZE( b ) > Jespresso.SIZE( a ); ) {
				if( ! Jset.setp_implies( a.base, a.offset, b.base, b.offset ) ) {
					continue;
				}
				/* set was contained in some set of B, so don't save pointer */
				continue next_label1;
			}
			/* set wasn't contained in any set of B, so save the pointer */
			A1[ pdest++ ] = a;
	// lnext1: ;
		}

		A1[ pdest ] = null;/* sentinel */
		return pdest;// pdest - A1;/* # elements in A1 */
	}

	/* sf_sort -- sort the sets of A */
	static Jset[] sf_sort(final Jset_family A, final Comparator<Jset> c) {
		/* Create a single array pointing to each cube of A */
		final Jset[] A1 = new Jset[ A.count + 1 ];
		int pdest = 0;
		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			Jespresso.PUTSIZE( data, p, Jset.set_ord( data, p ) );/* compute the set size */
			A1[ pdest++ ] = new Jset( data, p );/* save the pointer */
		}
		A1[ pdest ] = null;/* Sentinel -- never seen by sort */

		/* Sort cubes by size */
		Arrays.sort( A1, 0, A.count, c );
		return A1;
	}


	/** sf_list -- make a list of pointers to the sets in a set family */
	static Jset[] sf_list(final Jset_family A) {
		/* Create a single array pointing to each cube of A */
		final Jset[] A1 = new Jset[ A.count + 1 ];
		int pdest = 0;
		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			A1[ pdest++ ] = new Jset( data, p );/* save the pointer */
		}
		A1[ pdest ] = null;/* Sentinel */
		return A1;
	}


	/** sf_unlist -- make a set family out of a list of pointers to sets */
	static Jset_family sf_unlist(final Jset[] A1, final int totcnt, final int size) {
		Jset p;
		final Jset_family R = Jset.sf_new( totcnt, size );

		R.count = totcnt;
		final int[] data = R.data;// java
		for( int pr = 0, pa = 0; (p = A1[ pa++ ]) != null; pr += R.wsize ) {
			Jset.set_copy( data, pr, p.base, p.offset );
		}
		// A1 = null;
		return R;
	}

	// TODO java: check pbfirst
	/** sf_ind_unlist -- make a set family out of a list of pointers to sets */
	static Jset_family sf_ind_unlist(final Jset[] A1, final int totcnt, final int size, final int[] row_indices, final int[] pbase, final int pfirst) {
		Jset p;
		final Jset_family R = Jset.sf_new( totcnt, size );

		R.count = totcnt;
		int[] new_row_indices = new int[ totcnt ];
		final int[] data = R.data;// java
		for( int pr = 0, pa = 0, i = 0; (p = A1[ pa++ ]) != null; pr += R.wsize, i++ ) {
			Jset.set_copy( data, pr, p.base, p.offset );
			// new_row_indices[ i ] = row_indices[ (p - pfirst) / R.wsize ];
			new_row_indices[ i ] = row_indices[ (p.offset - pfirst) / R.wsize ];
		}
		for( int i = 0; i < totcnt; i++ ) {
			row_indices[ i ] = new_row_indices[ i ];
		}
		new_row_indices = null;
		// A1 = null;
		return R;
	}

	/** sf_merge -- merge three sorted lists of set pointers
	 *
	 * @param A1 will be disposed of
	 * @param B1 will be disposed of
	 * @param E1 will be disposed of
	 */
	static Jset_family sf_merge(final Jset[] A1, final Jset[] B1, final Jset[] E1, final int totcnt, final int size) {
		final Jset[][] temp = new Jset[3][];

		/* Allocate the result set_family */
		final Jset_family R = Jset.sf_new( totcnt, size );
		R.count = totcnt;

		/* Quick bubble sort to order the top member of the three arrays */
		final int n = 3;  temp[0] = A1;  temp[1] = B1;  temp[2] = E1;
		for( int i = 0; i < n-1; i++ ) {
			for( int j = i + 1; j < n; j++ ) {
				Jset[] ti = temp[ i ];
				Jset[] tj = temp[ j ];
				if( Jsetc.desc1( ti[ 0 ], tj[ 0 ] ) > 0 ) {
					final Jset[] swap = tj;
					tj = ti;
					ti = swap;
				}
			}
		}
		Jset[] pmin_base = temp[ 0 ];
		Jset[] pmid_base = temp[ 1 ];
		Jset[] pmax_base = temp[ 2 ];
		int pmin = 0, pmid = 0, pmax = 0;// java

		final int[] data = R.data;// java
		int pr = 0;
		/* Save the minimum element, then update pmin, pmid, pmax */
		while( pmin_base[ pmin ] != null ) {
			final Jset ps = pmin_base[ pmin++ ];
			Jset.set_copy( data, pr, ps.base, ps.offset );
			pr += R.wsize;
			if( Jsetc.desc1( pmin_base[ pmin ], pmax_base[ pmax ] ) > 0 ) {
				final Jset[] swap_base = pmax_base;
				final int swap = pmax;
				pmax_base = pmin_base;
				pmax = pmin;
				pmin_base = pmid_base;
				pmin = pmid;
				pmid_base = swap_base;
				pmid = swap;
			} else if( Jsetc.desc1( pmin_base[ pmin ], pmid_base[ pmid ] ) > 0 ) {
				final Jset[] swap_base = pmin_base;
				final int swap = pmin;
				pmin_base = pmid_base;
				pmin = pmid;
				pmid_base = swap_base;
				pmid = swap;
			}
		}

		// A1 = null;
		// B1 = null;
		// E1 = null;
		return R;
	}
}