package espresso;

import java.util.Arrays;
import java.util.Comparator;

/**
module: cvrm.c
Purpose: miscellaneous cover manipulation
a) verify two covers are equal, check consistency of a cover
b) unravel a multiple-valued cover into minterms
c) sort covers
*/
final class Jcvrm {

	private static void cb_unravel(final int[] c, final int off_c, final int start, final int end, final int[] startbase, final Jset_family B1) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] base = cube.temp[0];
		int size;

		/* Determine how many cubes it will blow up into, and create a mask
		for those parts that have only a single coordinate
		*/
		int expansion = 1;
		Jset.set_copy( base, startbase );
		for( int var = start; var <= end; var++ ) {
			if( (size = Jset.set_dist( c, off_c, cube.var_mask[var] )) < 2 ) {
				Jset.set_or( base, base, cube.var_mask[var] );
			} else {
				expansion *= size;
			}
		}
		Jset.set_and( base, c, off_c, base );

		/* Add the unravelled sets starting at the last element of B1 */
		final int offset = B1.count;
		B1.count += expansion;
		final int[] data = B1.data;// java
		for( int p = B1.wsize * (offset - 1) + B1.wsize, last = B1.count * B1.wsize; p < last; p += B1.wsize ) {
			Jset.set_copy( data, p, base );
		}

		int place = expansion;
		for( int var = start; var <= end; var++ ) {
			if( (size = Jset.set_dist( c, off_c, cube.var_mask[var] )) > 1 ) {
				final int skip = place;
				place = place / size;
				int n = 0;
				for( int i = cube.first_part[var]; i <= cube.last_part[var]; i++ ) {
					if( Jespresso.is_in_set( c, off_c, i ) ) {
						for( int j = n; j < expansion; j += skip ) {
							for( int k = 0; k < place; k++ ) {
								Jespresso.set_insert( data, B1.wsize * (j + k + offset), i );
							}
						}
						n += place;
					}
				}
			}
		}
	}


	static Jset_family unravel_range(final Jset_family B, final int start, final int end) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] startbase = cube.temp[1];

		/* Create the starting base for those variables not being unravelled */
		Jset.set_copy( startbase, cube.emptyset );
		for( int var = 0; var < start; var++ ) {
			Jset.set_or( startbase, startbase, cube.var_mask[ var ] );
		}
		for( int var = end + 1; var < cube.num_vars; var++ ) {
			Jset.set_or( startbase, startbase, cube.var_mask[ var ] );
		}

		/* Determine how many cubes it will blow up into */
		int total_size = 0;
		int[] data = B.data;// java
		for( int p = 0, last = p + B.count * B.wsize; p < last; p += B.wsize ) {
			int expansion = 1;
			for( int var = start; var <= end; var++ ) {
				int size;
				if( (size = Jset.set_dist( data, p, cube.var_mask[var] )) >= 2 ) {
					if( (expansion *= size) > 1000000 ) {
						Jcvrmisc.fatal("unreasonable expansion in unravel");
					}
				}
			}
			total_size += expansion;
		}

		/* We can now allocate a cover of exactly the correct size */
		final Jset_family B1 = Jespresso.new_cover( total_size );
		data = B.data;// java
		for( int p = 0, last = p + B.count * B.wsize; p < last; p += B.wsize ) {
			cb_unravel( data, p, start, end, startbase, B1 );
		}
		Jespresso.free_cover( B );
		return B1;
	}

	static Jset_family unravel(final Jset_family B, final int start) {
		return unravel_range( B, start, Jglobals.cube.num_vars - 1 );
	}

	/** lex_sort -- sort cubes in a standard lexical fashion */
	static Jset_family lex_sort(final Jset_family T) {
		final Comparator<Jset> lex_order = new Jlex_order();
		final Jset_family T1 = Jcontain.sf_unlist( Jcontain.sf_sort( T, lex_order ), T.count, T.sf_size );
		Jespresso.free_cover( T );
		return T1;
	}

	/** size_sort -- sort cubes by their size */
	static Jset_family size_sort(final Jset_family T) {
		final Comparator<Jset> descend = new Jdescend_order();
		final Jset_family T1 = Jcontain.sf_unlist( Jcontain.sf_sort( T, descend ), T.count, T.sf_size);
		Jespresso.free_cover( T );
		return T1;
	}


	/**  mini_sort -- sort cubes according to the heuristics of mini */
	static Jset_family mini_sort(final Jset_family F, final Comparator<Jset> c) {
		final int n = Jglobals.cube.size;

		/* Perform a column sum over the set family */
		int[] count = Jset.sf_count( F );

		/* weight is "inner product of the cube and the column sums" */
		final int[] data = F.data;// java
		for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			int cnt = 0;
			for( int i = 0; i < n; i++ ) {
				if( Jespresso.is_in_set( data, p, i ) ) {
					cnt += count[ i ];
				}
			}
			Jespresso.PUTSIZE( data, p, cnt );
		}
		count = null;

		/* use qsort to sort the array */
		final Jset[] F1 = Jcontain.sf_list( F );
		Arrays.sort( F1, 0, F.count, c );
		final Jset_family F_sorted = Jcontain.sf_unlist( F1, F.count, F.sf_size );
		Jespresso.free_cover( F );

		return F_sorted;
	}


	/** sort_reduce -- Espresso strategy for ordering the cubes before reduction */
	static Jset_family sort_reduce(final Jset_family T) {
		if( T.count == 0 ) {
			return T;
		}
		int bestsize = -1, size;
		final int n = Jglobals.cube.num_vars;
		/* find largest cube */
		int largest = -1;// java
		final int[] t_data = T.data;// java
		for( int p = 0, last = p + T.count * T.wsize; p < last; p += T.wsize ) {
			if( (size = Jset.set_ord( t_data, p )) > bestsize ) {
				largest = p;
				bestsize = size;
			}
		}

		for( int p = 0, last = p + T.count * T.wsize; p < last; p += T.wsize ) {
			Jespresso.PUTSIZE( t_data, p, ((n - Jsetc.cdist( t_data, largest, t_data, p )) << 7) + Math.min( Jset.set_ord( t_data, p ), 127 ) );
		}
		final Comparator<Jset> descend = new Jdescend_order();
		final Jset[] T1 = Jcontain.sf_list( T );
		Arrays.sort( T1, 0, T.count, descend );
		final Jset_family T_sorted = Jcontain.sf_unlist( T1, T.count, T.sf_size );
		Jespresso.free_cover( T );

		return T_sorted;
	}

	static Jset_family random_order(final Jset_family F) {
		final int[] f_data = F.data;
		int[] temp = Jset.set_new( F.sf_size );
		for( int i = F.count - 1; i > 0; i-- ) {
			int k;
			/* Choose a random number between 0 and i */
if( Jespresso.RANDOM ) {
			k = Jespresso.random.nextInt() % i;
} else {
			/* this is not meant to be really used; just provides an easy
			 "out" if random() and srandom() aren't around
			 */
			k = (i * 23 + 997) % i;
}
			/* swap sets i and k */
			k *= F.wsize;
			final int fi = F.wsize * i;// java
			Jset.set_copy( temp, f_data, k );
			Jset.set_copy( f_data, k, f_data, fi );
			Jset.set_copy( f_data, fi, temp );
		}
		temp = null;
		return F;
	}

	/**
	 *  cubelist_partition -- take a cubelist T and see if it has any components;
	 *  if so, return cubelist's of the two partitions A and B; the return value
	 *  is the size of the partition; if not, A and B
	 *  are undefined and the return value is 0
	 *
	 * @param T a list of cubes
	 * @param A cubelist of partition
	 * @param B cubelist of remainder
	 */
	@SuppressWarnings("boxing")
	static int cubelist_partition(final Jset[] T, final Jset[][] A, final Jset[][] B, final int comp_debug) {
		final int numcube = Jespresso.CUBELISTSIZE( T );

		/* Mark all cubes -- covered cubes belong to the partition */
		Jset p;
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			Jespresso.RESET( p, Jespresso.COVERED );
		}

		/*
		*  Extract a partition from the cubelist T; start with the first cube as a
		*  seed, and then pull in all cubes which share a variable with the seed;
		*  iterate until no new cubes are brought into the partition.
		*/
		int[] seed = Jset.set_save( T[ 2 ] );
		final Jset cof = T[ 0 ];
		Jespresso.SET( T[ 2 ], Jespresso.COVERED );
		int count = 1;

		boolean change;
		do {
			change = false;
			for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
				if( ! Jespresso.TESTP( p, Jespresso.COVERED ) && Jsetc.ccommon( p.base, p.offset, seed, cof.base, cof.offset ) ) {
					Jset.set_and( seed, seed, p.base, p.offset );
					Jespresso.SET( p, Jespresso.COVERED );
					change = true;
					count++;
				}

			}
		} while( change );

		seed = null;

		if( comp_debug != 0 ) {
			System.out.printf("COMPONENT_REDUCTION: split into %d %d\n",
					count, numcube - count);
		}

		if( count != numcube ) {
			/* Allocate and setup the cubelist's for the two partitions */
			A[ 0 ] = new Jset[ numcube + 3 ];
			B[ 0 ] = new Jset[ numcube + 3 ];
			A[ 0 ][0] = new Jset( Jset.set_save( T[0] ), 0 );
			B[ 0 ][0] = new Jset( Jset.set_save( T[0] ), 0 );
			int A1 = 2;// java *A[ A1 ]
			int B1 = 2;// java *B[ B1 ]

			/* Loop over the cubes in T and distribute to A and B */
			for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
				if( Jespresso.TESTP( p, Jespresso.COVERED ) ) {
					A[ 0 ][ A1++ ] = p;
				} else {
					B[ 0 ][ B1++ ] = p;
				}
			}

			/* Stuff needed at the end of the cubelist's */
			A[ 0 ][ A1++ ] = null;
			A[ 0 ][ 1 ] = new Jset( null, A1 );// (*A)[1] = (int[]) A1;
			B[ 0 ][ B1++ ] = null;
			B[ 0 ][ 1 ] = new Jset( null, B1 );// (*B)[1] = (int[]) B1;
		}

		return numcube - count;
	}

	/**
	 *  quick cofactor against a single output function
	 */
	static Jset_family cof_output(final Jset_family T, final int i) {
		final int[] mask = Jglobals.cube.var_mask[ Jglobals.cube.output ];
		final Jset_family T1 = Jespresso.new_cover( T.count );
		final int[] data = T.data;// java
	    for( int p = 0, last = p + T.count * T.wsize; p < last; p += T.wsize ) {
			if( Jespresso.is_in_set( data, p, i ) ) {
				final int pdest = T1.wsize * T1.count++;
				Jset.set_or( T1.data, pdest, data, p, mask );
				Jespresso.RESET( T1.data, pdest, Jespresso.PRIME );
			}
		}
		return T1;
	}

	/**
	 *  quick intersection against a single output function
	 */
	static Jset_family uncof_output(final Jset_family T, final int i) {
		final Jcube_struct cube = Jglobals.cube;// java

		if( T == null ) {
			return T;
		}

		final int[] mask = cube.var_mask[ cube.output ];
		final int[] data = T.data;// java
		for( int p = 0, last = p + T.count * T.wsize; p < last; p += T.wsize ) {
			Jset.set_diff( data, p, data, p, mask );
			Jespresso.set_insert( data, p, i );
		}
		return T;
	}


	/**
	*  A generic routine to perform an operation for each output function
	*
	*  func() is called with a PLA for each output function (with the output
	*  part effectively removed).
	*  func1() is called after reforming the equivalent output function
	*
	*  Each function returns TRUE if process is to continue
	*/
	static void foreach_output_function(final JpPLA PLA, final Ido_func fdo, final Iout_func fout) {// int (*func)(), int (*func1)()) {
		final Jcube_struct cube = Jglobals.cube;// java
		JpPLA PLA1;
		int i;

		/* Loop for each output function */
		for( i = 0; i < cube.part_size[cube.output]; i++ ) {

			/* cofactor on the output part */
			PLA1 = Jcvrin.new_PLA();
			PLA1.F = cof_output( PLA.F, i + cube.first_part[cube.output] );
			PLA1.R = cof_output( PLA.R, i + cube.first_part[cube.output] );
			PLA1.D = cof_output( PLA.D, i + cube.first_part[cube.output] );

			/* Call a routine to do something with the cover */
			if( ! fdo.fdo( PLA1, i ) ) {
				Jcvrin.free_PLA( PLA1 );
				return;
			}

			/* intersect with the particular output part again */
			PLA1.F = uncof_output( PLA1.F, i + cube.first_part[cube.output] );
			PLA1.R = uncof_output( PLA1.R, i + cube.first_part[cube.output] );
			PLA1.D = uncof_output( PLA1.D, i + cube.first_part[cube.output] );

			/* Call a routine to do something with the final result */
			if( ! fout.fout( PLA1, i ) ) {
				Jcvrin.free_PLA( PLA1 );
				return;
			}

			/* Cleanup for next go-around */
			Jcvrin.free_PLA( PLA1 );
		}
	}

	static Jset_family Fmin;
	static int[] phase;

	/**
	*  minimize each output function individually
	*/
	static void so_espresso(final JpPLA PLA, final int strategy) {
		Fmin = Jespresso.new_cover( PLA.F.count );
		if( strategy == 0 ) {
			foreach_output_function( PLA, new Jso_do_espresso(), new Jso_save() );
		} else {
			foreach_output_function( PLA, new Jso_do_exact(), new Jso_save() );
		}
		Jset.sf_free( PLA.F );
		PLA.F = Fmin;
	}

	/**
	*  minimize each output function, choose function or complement based on the
	*  one with the fewer number of terms
	*/
	static void so_both_espresso(final JpPLA PLA, final int strategy) {
		phase = Jset.set_save( Jglobals.cube.fullset );
		Fmin = Jespresso.new_cover( PLA.F.count );
		if( strategy == 0 ) {
			foreach_output_function( PLA, new Jso_both_do_espresso(), new Jso_both_save() );
		} else {
			foreach_output_function( PLA, new Jso_both_do_exact(), new Jso_both_save() );
		}
		Jset.sf_free( PLA.F );
		PLA.F = Fmin;
		PLA.phase = phase;
	}
}