package espresso;

import java.util.Random;

/**
 *  Module: espresso.c
 *  Purpose: The main espresso algorithm
 *
 *  Returns a minimized version of the ON-set of a function
 *
 *  The following global variables affect the operation of Espresso:
 *
 *  MISCELLANEOUS:
 *      trace
 *          print trace information as the minimization progresses
 *
 *      remove_essential
 *          remove essential primes
 *
 *      single_expand
 *          if true, stop after first expand/irredundant
 *
 *  LAST_GASP or SUPER_GASP strategy:
 *      use_super_gasp
 *          uses the super_gasp strategy rather than last_gasp
 *
 *  SETUP strategy:
 *      recompute_onset
 *          recompute onset using the complement before starting
 *
 *      unwrap_onset
 *          unwrap the function output part before first expand
 *
 *  MAKE_SPARSE strategy:
 *      force_irredundant
 *          iterates make_sparse to force a minimal solution (used
 *          indirectly by make_sparse)
 *
 *      skip_make_sparse
 *          skip the make_sparse step (used by opo only)
 */

final class Jespresso {
	static final String VERSION = "UC Berkeley, Espresso Version #2.3, Release date 01/31/88";

	/** Name of package */
	//static final String PACKAGE "espresso";

	/** Version number of package */
	//static final String VERSION = "1.0";

	static final boolean RANDOM = false;
	static final boolean RANDOM_MINCOV = false;
	static final boolean FAST_AND_LOOSE = false;
	static final boolean COLS = false;
	static final boolean NO_INLINE = false;

	/* Define host machine characteristics of "unsigned int" */
	/** # bits per integer, 32 */
	static final int BPI = 32;// Integer.SIZE;

	/** log(BPI)/log(2) */
	static final int LOGBPI = 5;// BPI == 32 ? 5 : 4;

	/* Define the set type */
	// typedef unsigned int *pset;// java: int[] will be used

	//#ifdef lint
//	private static final int DISJOINT = 0x5555;
//#else
	static final int DISJOINT = 0x55555555;// BPI == 32 ? 0x55555555 : 0x5555;
//#endif

	static final Random random = new Random();

	/* espresso.h -- header file for Espresso-mv */

	/** Table for efficient bit counting */
	private static final int bit_count[/* 256 */] = {
		0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
		1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
		1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
		3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8
	};

	static final int count_ones(final int v) {
// if( Jespresso.BPI == 32 ) {
		return (bit_count[v & 255] + bit_count[(v >>> 8) & 255] + bit_count[(v >>> 16) & 255] + bit_count[v >>> 24]);
//} else {
//		return (bit_count[ v & 255 ] + bit_count[ (v >>> 8) & 255 ]);
//}
	}

/*-----THIS USED TO BE set.h----- */

/*
 *  set.h -- definitions for packed arrays of bits
 *
 *   This header file describes the data structures which comprise a
 *   facility for efficiently implementing packed arrays of bits
 *   (otherwise known as sets, cf. Pascal).
 *
 *   A set is a vector of bits and is implemented here as an array of
 *   unsigned integers.  The low order bits of set[0] give the index of
 *   the last word of set data.  The higher order bits of set[0] are
 *   used to store data associated with the set.  The set data is
 *   contained in elements set[1] ... set[LOOP(set)] as a packed bit
 *   array.
 *
 *   A family of sets is a two-dimensional matrix of bits and is
 *   implemented with the data type "set_family".
 *
 *   BPI == 32 and BPI == 16 have been tested and work.
 */

	/** Macros to set single elements */
	static final int WHICH_WORD(final int element) {
		return (element >>> LOGBPI) + 1;
	}

	/** Macros to test single elements */
	static final int WHICH_BIT(final int element) {
		return element & (BPI - 1);
	}

	/** # of ints needed to allocate a set with "size" elements */
	static final int SET_SIZE(final int size) {
// if( BPI == 32 ) {
		return (size <= BPI ? 2 : (WHICH_WORD( size - 1 ) + 1));
//} else {
//		return (size <= BPI ? 3 : (WHICH_WORD( size - 1 ) + 2));
//}
	}

	/*
	 *  Three fields are maintained in the first word of the set
	 *      LOOP is the index of the last word used for set data
	 *      LOOPCOPY is the index of the last word in the set
	 *      SIZE is available for general use (e.g., recording # elements in set)
	 *      NELEM retrieves the number of elements in the set
	 */
	static final int LOOP(final Jset set) {// TODO check if it needs to have 2 versions
		return (set.base[ set.offset ] & 0x03ff);
	}
	static final int LOOP(final int[] set, final int off) {
		return (set[ off ] & 0x03ff);
	}
	static final int LOOP(final int[] set) {
		return (set[ 0 ] & 0x03ff);
	}

	static final void PUTLOOP(final Jset set, final int i) {
		int s = set.base[ set.offset ];
		s &= ~0x03ff;
		s |= i;
		set.base[ set.offset ] = s;
	}
	static final void PUTLOOP(final int[] set, final int off, final int i) {
		int s = set[ off ];
		s &= ~0x03ff;
		s |= i;
		set[ off ] = s;
	}
	static final void PUTLOOP(final int[] set, final int i) {
		int s = set[ 0 ];
		s &= ~0x03ff;
		s |= i;
		set[ 0 ] = s;
	}

	static final int LOOPCOPY(final int[] set, final int off) {
// if( BPI == 32 ) {
		return LOOP( set, off );
//} else {
//		return (LOOP( set, off ) + 1);
//}
	}
	static final int LOOPCOPY(final int[] set) {
// if( BPI == 32 ) {
		return LOOP( set );
//} else {
//		return (LOOP( set ) + 1);
//}
	}
	static final int LOOPCOPY(final Jset set) {
// if( BPI == 32 ) {
		return LOOP( set );
//} else {
//		return (LOOP( set ) + 1);
//}
	}

	static final int SIZE(final Jset set) {
// if( BPI == 32 ) {
		return (set.base[ set.offset ] >>> 16);
//} else {
//		return set.base[ set.offset + LOOP( set ) + 1 ];
//}
	}
	static final int SIZE(final int[] set, final int off) {
//if( BPI == 32 ) {
		return (set[ off ] >>> 16);
//} else {
//		return set[ off + LOOP( set, off ) + 1 ];
//}
	}

	static final void PUTSIZE(final int[] set, final int off, final int size) {
//if( BPI == 32 ) {
		int s = set[ off ];
		s &= 0xffff;
		s |= (size << 16);
		set[ off ] = s;
//} else {
//		set[ off + LOOP( set, off ) + 1 ] = size;
//}
	}

	static final int NELEM(final Jset set) {
		return (BPI * LOOP( set ));
	}
	static final int NELEM(final int[] set, final int off) {
		return (BPI * LOOP( set, off ));
	}
	static final int NELEM(final int[] set) {
		return (BPI * LOOP( set ));
	}

	static final int LOOPINIT(final int size) {
		return ((size <= BPI) ? 1 : WHICH_WORD( size - 1 ));
	}

	/* FLAGS store general information about the set */
	static final void SET(final Jset set, final int flag) {
		set.base[ set.offset ] |= flag;
	}
	static final void SET(final int[] set, final int off, final int flag) {
		set[ off ] |= flag;
	}
	static final void SET(final int[] set, final int flag) {
		set[ 0 ] |= flag;
	}

	static final void RESET(final Jset set, final int flag) {
		set.base[ set.offset ] &= ~ flag;
	}
	static final void RESET(final int[] set, final int off, final int flag) {
		set[ off ] &= ~ flag;
	}
	static final void RESET(final int[] set, final int flag) {
		set[ 0 ] &= ~ flag;
	}

	static final boolean TESTP(final int[] set, final int off, final int flag) {
		return (set[ off ] & flag) != 0;
	}
	static final boolean TESTP(final int[] set, final int flag) {
		return (set[ 0 ] & flag) != 0;
	}
	static final boolean TESTP(final Jset set, final int flag) {
		return (set.base[ set.offset ] & flag) != 0;
	}

	/* Flag definitions are ... */
	/** cube is prime */
	static final int PRIME     = 0x8000;
	/** cube cannot be essential prime */
	static final int NONESSEN  = 0x4000;
	/** cube is still active */
	static final int ACTIVE    = 0x2000;
	/** cube is redundant(at this point) */
	static final int REDUND    = 0x1000;
	/** cube has been covered */
	static final int COVERED   = 0x0800;
	/** cube is relatively essential */
	static final int RELESSEN  = 0x0400;

	/** Check for set membership */
	static final boolean is_in_set(final int[] set, final int off, final int e) {
		return (set[ off + WHICH_WORD( e ) ] & (1 << WHICH_BIT( e ))) != 0;
	}
	static final boolean is_in_set(final int[] set, final int e) {
		return (set[ WHICH_WORD( e ) ] & (1 << WHICH_BIT( e ))) != 0;
	}

	/** Remove set element */
	static final void set_remove(final int[] set, final int off, final int e) {
		set[ off + WHICH_WORD( e ) ] &= ~ (1 << WHICH_BIT( e ));
	}
	static final void set_remove(final int[] set, final int e) {
		set[ WHICH_WORD( e ) ] &= ~ (1 << WHICH_BIT( e ));
	}

	/** Insert set element */
	static final void set_insert(final int[] set, final int off, final int e) {
		set[ off + WHICH_WORD( e ) ] |= 1 << WHICH_BIT( e );
	}
	static final void set_insert(final int[] set, final int e) {
		set[ WHICH_WORD( e ) ] |= 1 << WHICH_BIT( e );
	}

/*----- END OF set.h ----- */

	/* Define a boolean type */
	// #define bool int
	static final int FALSE = 0;
	static final int TRUE  = 1;
	static final int MAYBE = 2;

	static String print_bool(final int x) {
		return (x == 0 ? "FALSE" : (x == 1 ? "TRUE" : "MAYBE"));
	}

	/* Map many cube/cover types/routines into equivalent set types/routines */

	static final int[] new_cube() {
		return Jset.set_new( Jglobals.cube.size );
	}

	static final Jset_family new_cover(final int i) {
		return Jset.sf_new( i, Jglobals.cube.size );
	}

	static final void free_cover(final Jset_family r) {
		Jset.sf_free( r );
	}

	/** This is a hack which I wish I hadn't done, but too painful to change */
	static final int CUBELISTSIZE(final Jset[] T) {
		return T[ 1 ].offset - 3;// CUBELISTSIZE(T) (((pcube *) T[1] - T) - 3)
	}

	/* The pla_type field describes the input and output format of the PLA */
	static final int F_type         = 1;
	static final int D_type         = 2;
	static final int R_type         = 4;
	static final int PLEASURE_type  = 8;              /* output format */
	static final int EQNTOTT_type   = 16;              /* output format algebraic eqns */
	static final int KISS_type	               = 128;		/* output format kiss */
	static final int CONSTRAINTS_type	       = 256;	/* output the constraints (numeric) */
	static final int SYMBOLIC_CONSTRAINTS_type = 512;	/* output the constraints (symbolic) */
	static final int FD_type  = (F_type | D_type);
	static final int FR_type  = (F_type | R_type);
	static final int DR_type  = (D_type | R_type);
	static final int FDR_type = (F_type | D_type | R_type);

	/* Definitions for the debug variable */
	static final int COMPL   = 0x0001;
	static final int ESSEN   = 0x0002;
	static final int EXPAND  = 0x0004;
	static final int EXPAND1 = 0x0008;
	static final int GASP    = 0x0010;
	static final int IRRED   = 0x0020;
	static final int REDUCE  = 0x0040;
	static final int REDUCE1 = 0x0080;
	static final int SPARSE  = 0x0100;
	static final int TAUT    = 0x0200;
	static final int EXACT   = 0x0400;
	static final int MINCOV  = 0x0800;
	static final int MINCOV1 = 0x1000;
	static final int SHARP   = 0x2000;
	static final int IRRED1  = 0x4000;

	/* Define constants used for recording program statistics */
	static final int TIME_COUNT     = 22;
	static final int READ_TIME      = 0;
	static final int COMPL_TIME     = 1;
	static final int ONSET_TIME     = 2;// is not used
	static final int ESSEN_TIME     = 3;
	static final int EXPAND_TIME    = 4;
	static final int IRRED_TIME     = 5;
	static final int REDUCE_TIME    = 6;
	static final int GEXPAND_TIME   = 7;
	static final int GIRRED_TIME    = 8;
	static final int GREDUCE_TIME   = 9;
	static final int PRIMES_TIME    = 10;
	static final int MINCOV_TIME    = 11;
	static final int MV_REDUCE_TIME = 12;
	static final int RAISE_IN_TIME  = 13;
	static final int VERIFY_TIME    = 14;
	static final int WRITE_TIME     = 15;
	static final int FCC_TIME       = 16;
	static final int ETR_TIME       = 17;
	static final int ETRAUX_TIME    = 18;
	private static final int SIGMA_TIME     = 19;// is not used
	private static final int UCOMP_TIME     = 20;// is not used
	private static final int BW_TIME        = 21;// is not used

	/* For those who like to think about PLAs, macros to get at inputs/outputs */
	// #define NUMINPUTS       cube.num_binary_vars// is not used
	// #define NUMOUTPUTS      cube.part_size[cube.num_vars - 1]// is not used

	static boolean POSITIVE_PHASE(final JpPLA PLA, final int pos) {// is not used
		final Jcube_struct cube = Jglobals.cube;// java
		return is_in_set( PLA.phase, cube.first_part[ cube.output ] + pos );
	}

	static String INLABEL(final JpPLA PLA, final int var) {
		return PLA.label[ Jglobals.cube.first_part[ var ] + 1 ];
	}

	static String OUTLABEL(final JpPLA PLA, final int pos) {
		return PLA.label[ Jglobals.cube.first_part[ Jglobals.cube.output ] + pos ];
	}

	static int GETINPUT(final int[] c, final int off_c, final int pos) {
		return ((c[ off_c + WHICH_WORD( pos << 1 ) ] >>> WHICH_BIT( pos << 1 )) & 3);
	}

	static boolean GETOUTPUT(final int[] c, final int pos) {
		return is_in_set( c, Jglobals.cube.first_part[ Jglobals.cube.output ] + pos);
	}

	static void PUTINPUT(final int[] c, int pos, final int value) {
		pos <<= 1;
		c[ WHICH_WORD( pos ) ] = (c[ WHICH_WORD( pos ) ] & ~(3 << WHICH_BIT( pos ))) | (value << WHICH_BIT( pos ));
	}

	static void PUTOUTPUT(final int[] c, final int pos, final int value) {
		c[ WHICH_WORD( pos ) ] = (c[ WHICH_WORD( pos ) ] & (1 << WHICH_BIT( pos ))) | (value << WHICH_BIT( pos ));
	}

	static final int TWO  = 3;
	static final int DASH = 3;
	static final int ONE  = 2;
	static final int ZERO = 1;

	/* function declarations */

	static Jset_family espresso(Jset_family F, final Jset_family D1, final Jset_family R) {
		final Jcube_struct cube = Jglobals.cube;// java
		Jset_family E, D, Fsave;
		final Jcost_t cost = new Jcost_t();
		final Jcost_t best_cost = new Jcost_t();

// begin:
		do {
			Fsave = Jset.sf_save( F );		/* save original function */
			D = Jset.sf_save( D1 );		/* make a scratch copy of D */

			/* Setup has always been a problem */
			if( Jglobals.recompute_onset ) {
			    {
					final long t = System.currentTimeMillis();
					E = Jcompl.simplify( Jcofactor.cube1list( F ) );
					if( Jglobals.trace ) {
						Jcvrmisc.print_trace( E, "SIMPLIFY   ", System.currentTimeMillis() - t);
					}
				}
				free_cover( F );
				F = E;
			}
			Jcvrmisc.cover_cost( F, cost );
			if( Jglobals.unwrap_onset && (cube.part_size[cube.num_vars - 1] > 1 )
					&& (cost.out != cost.cubes * cube.part_size[ cube.num_vars - 1 ])
					&& (cost.out < 5000) ) {
			    {
					final long t = System.currentTimeMillis();
					F = Jcontain.sf_contain( Jcvrm.unravel( F, cube.num_vars - 1 ) );
					if( Jglobals.trace ) {
						Jcvrmisc.print_trace( F, "SETUP      ", System.currentTimeMillis() - t);
					}
				}
			}

			/* Initial expand and irredundant */
			final int[] data = F.data;// java
			for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
				RESET( data, p, PRIME );
			}
		    {
				final long t = System.currentTimeMillis();
				F = Jexpand.expand( F, R, false );
				Jcvrmisc.totals( t, EXPAND_TIME, F, cost );
			}
		    {
				final long t = System.currentTimeMillis();
				F = Jirred.irredundant( F, D );
				Jcvrmisc.totals( t, IRRED_TIME, F, cost );
			}

			if( ! Jglobals.single_expand ) {
				if( Jglobals.remove_essential ) {
				    {
						final long t = System.currentTimeMillis();
						final Jset_family f_holder[] = new Jset_family[] { F };// java
						final Jset_family d_holder[] = new Jset_family[] { D };// java
						E = Jessen.essential( f_holder, d_holder );
						F = f_holder[ 0 ];// java
						D = d_holder[ 0 ];// java
						Jcvrmisc.totals( t, ESSEN_TIME, E, cost );
					}
				} else {
					E = new_cover( 0 );
				}

				Jcvrmisc.cover_cost( F, cost );
				do {
					/* Repeat inner loop until solution becomes "stable" */
					do {
						Jcvrmisc.copy_cost( cost, best_cost );
					    {
							final long t = System.currentTimeMillis();
							F = Jreduce.reduce( F, D );
							Jcvrmisc.totals( t, REDUCE_TIME, F, cost );
						}
					    {
							final long t = System.currentTimeMillis();
							F = Jexpand.expand( F, R, false );
							Jcvrmisc.totals( t, EXPAND_TIME, F, cost );
						}
					    {
							final long t = System.currentTimeMillis();
							F = Jirred.irredundant( F, D );
							Jcvrmisc.totals( t, IRRED_TIME, F, cost );
						}
					} while( cost.cubes < best_cost.cubes );

					/* Perturb solution to see if we can continue to iterate */
					Jcvrmisc.copy_cost( cost, best_cost );
					if( Jglobals.use_super_gasp ) {
						F = Jgasp.super_gasp( F, D, R, cost );
						if( cost.cubes >= best_cost.cubes ) {
							break;
						}
					} else {
						F = Jgasp.last_gasp( F, D, R, cost );
					}

				} while( cost.cubes < best_cost.cubes ||
						(cost.cubes == best_cost.cubes && cost.total < best_cost.total) );

				/* Append the essential cubes to F */
				F = Jset.sf_append( F, E );                /* disposes of E */
				if( Jglobals.trace ) {
					Jcvrmisc.size_stamp( F, "ADJUST     " );
				}
			}

			/* Free the D which we used */
			free_cover( D );

			/* Attempt to make the PLA matrix sparse */
			if( ! Jglobals.skip_make_sparse ) {
				F = Jsparse.make_sparse( F, D1, R );
			}

			/*
			*  Check to make sure function is actually smaller !!
			*  This can only happen because of the initial unravel.  If we fail,
			*  then run the whole thing again without the unravel.
			*/
			if( Fsave.count < F.count ) {
				free_cover( F );
				F = Fsave;
				Jglobals.unwrap_onset = false;
				// goto begin;
			} else {
				free_cover( Fsave );
				break;
			}
		} while( true );
		return F;
	}

	/**
	 *  util_print_time -- massage a long which represents a time interval in
	 *  milliseconds, into a string suitable for output
	 *
	 *  Hack for IBM/PC -- avoids using floating point
	 */
	@SuppressWarnings("boxing")
	static String print_time(final long t) {
	    return String.format( "%d.%02d sec", t / 1000, (t % 1000) / 10 );
	}
}