package espresso;

/**
    module: essen.c
    purpose: Find essential primes in a multiple-valued function
*/
final class Jessen {

	/**
	    essential -- return a cover consisting of the cubes of F which are
	    essential prime implicants (with respect to F u D); Further, remove
	    these cubes from the ON-set F, and add them to the OFF-set D.

	    Sometimes EXPAND can determine that a cube is not an essential prime.
	    If so, it will set the "NONESSEN" flag in the cube.

	    We count on IRREDUNDANT to have set the flag RELESSEN to indicate
	    that a prime was relatively essential (i.e., covers some minterm
	    not contained in any other prime in the current cover), or to have
	    reset the flag to indicate that a prime was relatively redundant
	    (i.e., all minterms covered by other primes in the current cover).
	    Of course, after executing irredundant, all of the primes in the
	    cover are relatively essential, but we can mark the primes which
	    were redundant at the start of irredundant and avoid an extra check
	    on these primes for essentiality.
	*/
	static Jset_family essential(final Jset_family[] Fp, final Jset_family[] Dp) {
		final Jset_family F = Fp[ 0 ], D = Dp[ 0 ];

		/* set all cubes in F active */
		Jset.sf_active( F );

		/* Might as well start out with some cubes in E */
		Jset_family E = Jespresso.new_cover( 10 );

		final int[] data = F.data;// java
		for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			/* don't test a prime which EXPAND says is nonessential */
			if( ! Jespresso.TESTP( data, p, Jespresso.NONESSEN ) ) {
				/* only test a prime which was relatively essential */
				if( Jespresso.TESTP( data, p, Jespresso.RELESSEN ) ) {
					/* Check essentiality */
					if( Jessen.essen_cube( F, D, data, p ) ) {
						if( (Jglobals.debug & Jespresso.ESSEN) != 0 ) {
							System.out.printf("ESSENTIAL: %s\n", Jcvrout.pc1( data, p ) );
						}
						E = Jset.sf_addset( E, data, p );
						Jespresso.RESET( data, p, Jespresso.ACTIVE );
						F.active_count--;
					}
				}
			}
		}

		Fp[ 0 ] = Jset.sf_inactive( F );               /* delete the inactive cubes from F */
		Dp[ 0 ] = Jset.sf_join( D, E );		/* add the essentials to D */
		Jset.sf_free( D );
		return E;
	}

	/**
	essen_cube -- check if a single cube is essential or not

	The prime c is essential iff

	consensus((F u D) # c, c) u D

	does not contain c.
	*/
	static boolean essen_cube(final Jset_family F, final Jset_family D, final int[] c, final int off_c) {

		/* Append F and D together, and take the sharp-consensus with c */
		final Jset_family FD = Jset.sf_join( F, D );
		final Jset_family H = Jessen.cb_consensus( FD, c, off_c );
		Jespresso.free_cover( FD );

		/* Add the don't care set, and see if this covers c */
		Jset[] H1 = Jcofactor.cube2list( H, D );
		final boolean essen = ! Jirred.cube_is_covered( H1, c, off_c );
		H1 = null;

		Jespresso.free_cover( H );
		return essen;
	}

	/**
	*  cb_consensus -- compute consensus(T # c, c)
	*/
	static Jset_family cb_consensus(final Jset_family T, final int[] c, final int off_c) {
		Jset_family R = Jespresso.new_cover( T.count << 1 );
		int[] temp = Jespresso.new_cube();
		final int[] data = T.data;// java
		for( int p = 0, last = p + T.count * T.wsize; p < last; p += T.wsize ) {
			if( data != c || p != off_c ) {
				switch( Jsetc.cdist01( data, p, c, off_c ) ) {
				case 0:
					/* distance-0 needs special care */
					R = cb_consensus_dist0( R, data, p, c, off_c );
					break;

				case 1:
					/* distance-1 is easy because no sharping required */
					Jsetc.consensus( temp, 0, data, p, c, off_c );
					R = Jset.sf_addset( R, temp );
					break;
				}
			}
		}
		temp = null;
		return R;
	}

	/*
	*  form the sharp-consensus for p and c when they intersect
	*  What we are forming is consensus(p # c, c).
	*/
	static Jset_family cb_consensus_dist0(Jset_family R, final int[] p, final int off_p, final int[] c, final int off_c) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] p_diff_c = cube.temp[0], p_and_c = cube.temp[1];

		/* If c contains p, then this gives us no information for essential test */
		if( Jset.setp_implies( p, off_p, c, off_c ) ) {
			return R;
		}

		/* For the multiple-valued variables */
		int[] temp = Jespresso.new_cube();
		boolean got_one = false;
		Jset.set_diff( p_diff_c, p, off_p, c, off_c );
		Jset.set_and( p_and_c, p, off_p, c, off_c );

		for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
			/* Check if c(var) is contained in p(var) -- if so, no news */
			final int[] mask = cube.var_mask[var];
			if( ! Jset.setp_disjoint( p_diff_c, mask ) ) {
				Jset.set_merge( temp, c, off_c, p_and_c, mask );
				R = Jset.sf_addset( R, temp );
				got_one = true;
			}
		}

		/* if no cube so far, add one for the intersection */
		if( ! got_one && cube.num_binary_vars > 0 ) {
			/* Add a single cube for the intersection of p and c */
			Jset.set_and( temp, p, off_p, c, off_c );
			R = Jset.sf_addset( R, temp );
		}

		temp = null;
		return R;
	}
}