package espresso;

import java.util.Arrays;
import java.util.Comparator;

/** Module:essentiality.c
 *	contains routines for performing essentiality test and reduction
 * Routines:
 * pcover ess_test_and_reduction():
 *	determines essentiality of the given signature cube and returns
 * 	cover of the signature cube (Hammered signature cube) if found
 *	inessential.
 * pcover aux_ess_test_and_reduction():
 *	core compuation routine which determines the essentiality of
 *	the signature cube and performs reduction of inessential signature
 *	cubes by recursively descending down the cube-tree.
 */
final class Jessentiality implements Comparator<JVAR> {

	/* Yuk, Yuk, Yuk!! More Globals. It seems recursive routines have unholy
	alliance with globals */

	/** List of raised variables  in cube c */
	static int[] c_free_list;
	/** active size of the above list */
	static int c_free_count;

	/** List of subset of raised variables in cube c
	which are raised in each cube of offset R */
	static int[] r_free_list;
	/** active size of the above list */
	static int r_free_count;
	/** current position in  the list above */
	static int r_head;

	/** c_free_list - r_free_list */
	static int[] reduced_c_free_list;
	/** active size of the above list */
	static int reduced_c_free_count;

	/** List of unate variables in the reduced_c_free_list */
	static JVAR[] unate_list;
	/** active size of the above list */
	static int unate_count;

	/** List of binate variables in the reduced_c_free_list */
	static JVAR[] binate_list;
	/** active size of the above list */
	static int binate_count;

	/** permutation of reduced c_free_count determining
	static ordering of variables */
	static int[] variable_order;
	/** active size of the above list */
	static int variable_count;
	/** current position in  the list above */
	static int variable_head;

	/* The passive size allocated on the first call to etr_order is equal
	to the number of binary variables */
	/* Clearly not the most optimized usage of memory. Not worth saving
	few pennies when the total memory budget is quite large */

	/** A global bag to collect the signature cubes in the cover
	of inessential signature cube */
	static Jset_family COVER;

	static int num_binary_vars;
	/** etr_order:
	* What does it do:
	*	Performs performs ordering of variables before calling
	*	essentiality test and reduction routine
	* Inputs:
	*	R: Cover of the Off-set.
	*	c: cube to be reduced
	* Output:
	*	COVER: signature cube cover of given inessential signature cube
	* Strategy: As many cubes in the cover as possible.
	* 	-> As deep a search tree recursion as possible
	*	-> variable ordering
	*	-> static ordering to minimize computation
	*		1. Free
	*		2. Freer Unate
	*		3. Freer Binate
	*/
	static Jset_family etr_order(final Jset_family F, final Jset_family E, final Jset_family R, final int[] c, final int[] d) {
		num_binary_vars = Jglobals.cube.num_binary_vars;
		c_free_list = new int[ num_binary_vars ];
		r_free_list = new int[ num_binary_vars ];
		reduced_c_free_list = new int[  num_binary_vars ];
		unate_list = new JVAR[ num_binary_vars ];
		binate_list = new JVAR[ num_binary_vars ];

		variable_order = new int[ num_binary_vars ];

		if( null == c_free_list || null == r_free_list || null == reduced_c_free_list ||
				null == unate_list || null == binate_list || null == variable_order ) {
			System.err.println("etr_order:alloc");
			Jmain.closeStreams();
			System.exit( 1 );
		}

		/* 1.Identify free variables of cube c */
		c_free_count = 0;
		for( int v = 0; v < num_binary_vars; v++ ) {
			final int e0 = v << 1;
			final int e1 = e0 + 1;
			if( Jespresso.is_in_set( d, e0 ) && Jespresso.is_in_set( d, e1 ) ) {
				c_free_list[ c_free_count++ ] = v;
			}
		}

		/* 2.Identify corresponding free variables of R */
		r_head = 0;
		r_free_count = 0;
		reduced_c_free_count = 0;
		for( int i = 0; i < c_free_count; i++ ) {
			final int v = c_free_list[ i ];
			final int e0 = v << 1;
			final int e1 = e0 + 1;
			boolean free_var = true;
			final int[] data = R.data;// java
			for( int r = 0, lastr = r + R.count * R.wsize; r < lastr; r += R.wsize ) {
				if( ! Jespresso.is_in_set( data, r, e0 ) || ! Jespresso.is_in_set( data, r, e1 ) ) {
					free_var = false;
					break;
				}
			}
			if( free_var ) {
				r_free_list[ r_free_count++ ] = v;
			}
			else{
				reduced_c_free_list[ reduced_c_free_count++ ] = v;
			}
		}

		/* 3.Identify unate and binate variables and sort them in the
		decreasing free_count */
		unate_count = 0;
		binate_count = 0;
		for( int i = 0; i < reduced_c_free_count; i++ ) {
			final int v = reduced_c_free_list[ i ];
			final int e0 = v << 1;
			final int e1 = e0 + 1;
			int even_count = 0;
			int odd_count = 0;
			int free_count = 0;
			final int[] data = R.data;// java
			for( int r = 0, lastr = r + R.count * R.wsize; r < lastr; r += R.wsize ) {
				final boolean odd = Jespresso.is_in_set( data, r, e0 );
				final boolean even = Jespresso.is_in_set( data, r, e1 );
				if( odd && even ) {
					free_count++;
				}
				else if( odd ) {
					odd_count++;
				}
				else{
					even_count++;
				}
			}
			final JVAR p = new JVAR();
			if( odd_count == 0 || even_count == 0 ) {
				unate_list[ unate_count++ ] = p;
				p.variable = v;
				p.free_count = free_count;
			} else {
				binate_list[ binate_count++ ] = p;
				p.variable = v;
				p.free_count = free_count;
			}
		}

		final Comparator<JVAR> ascending = new Jessentiality();
		Arrays.sort( unate_list, 0, unate_count, ascending );
		Arrays.sort( binate_list, 0, binate_count, ascending );

		variable_head = 0;
		variable_count = 0;
		for( int i = 0; i < binate_count; i++ ) {
			variable_order[ variable_count++ ] = binate_list[ i ].variable;
		}
		for( int i = 0; i < unate_count; i++ ) {
			variable_order[ variable_count++ ] = unate_list[ i ].variable;
		}

		/* 4.Recursively go down the tree defined by r_free_list,
		invoking "etr" at the leaves */

		COVER = Jespresso.new_cover( 10 );
		Jblack_white.setup_bw( R, c );
		Jespresso.SET( c, Jespresso.NONESSEN );

		aux_etr_order( F, E, R, c, d );

		Jblack_white.free_bw();
		c_free_list = null;
		r_free_list = null;
		reduced_c_free_list = null;
		unate_list = null;
		binate_list = null;
		variable_order = null;

		return COVER;
	}

	@Override
	public final int compare(final JVAR p1, final JVAR p2) {
	// private static int ascending(final JVAR p1, final JVAR p2) {
		final int f1 = p1.free_count;
		final int f2 = p2.free_count;

		if( f1 > f2 ) {
			return 1;
		} else if( f1 < f2 ) {
			return -1;
		} else {
			return 0;
		}
	}

	/**
	* aux_etr_order
	* Objective:main recursive routine for reducing the inessential signature cube.
	*	Very similar to aux_ess_test_and_reduction;
	* Input:
	*	c: signature cube;
	*	d: cube contained in the signature cube;
	*	E: Extended don't care set. DC + identified ESC;
	*	R: OFFSET cover;
	*/
	static void aux_etr_order(final Jset_family F, final Jset_family E, final Jset_family R, final int[] c, final int[] d) {
		int v_index,e0,e1;

		/* Special Cases */
		Jset[] local_dc = Jcofactor.cube3list( F, E, COVER );
		if( Jirred.cube_is_covered( local_dc, d, 0 ) ) {
			local_dc = null;
			return;
		}
		if( ! Jblack_white.black_white() ) {
			local_dc = null;
			int[] sigma_d = Jsigma.get_sigma( R, d, 0 );
			Jset.sf_addset( COVER, sigma_d );
			sigma_d = null;
			return;
		}
		if( variable_head == variable_count ) {
			final Jset_family minterms = get_mins( d );
			final int[] data = minterms.data;// java
			for( int d_minterm = 0, i = 0; i < minterms.count; d_minterm += minterms.wsize, i++ ) {
				if( Jirred.cube_is_covered( local_dc, data, d_minterm ) ) {
					continue;
				}
				int[] sigma_d = Jsigma.get_sigma( R, data, d_minterm );
				if( Jset.setp_equal( sigma_d, c ) ) {
					Jespresso.RESET( c, Jespresso.NONESSEN );
					sigma_d = null;
					break;
				}
				Jset.sf_addset( COVER, sigma_d );
				sigma_d = null;
			}
			Jset.sf_free( minterms );
			local_dc = null;
			return;
		}// else {
			v_index = variable_order[ variable_head ];
		//}
		local_dc = null;

		e0 = (v_index << 1);
		e1 = e0 + 1;

		variable_head++;

		Jespresso.set_remove( d, e1 );
		Jblack_white.reset_black_list();
		Jblack_white.split_list( R, e0 );
		Jblack_white.push_black_list();
		{
			final long t = System.currentTimeMillis();
			aux_etr_order( F, E, R, c, d );
			Jutil_signature.s_totals( t, Jespresso.ETRAUX_TIME );
		}
		if( ! Jespresso.TESTP( c, Jespresso.NONESSEN ) ) {
			return;
		}
		Jblack_white.pop_black_list();
		Jblack_white.merge_list();
		Jespresso.set_insert( d, e1 );

		Jespresso.set_remove( d, e0 );
		Jblack_white.reset_black_list();
		Jblack_white.split_list( R, e1 );
		Jblack_white.push_black_list();
		{
			final long t = System.currentTimeMillis();
			aux_etr_order( F, E, R, c, d );
			Jutil_signature.s_totals( t, Jespresso.ETRAUX_TIME );
		}
		if( ! Jespresso.TESTP( c, Jespresso.NONESSEN ) ) {
			return;
		}
		Jblack_white.pop_black_list();
		Jblack_white.merge_list();
		Jespresso.set_insert( d, e0 );

		variable_head--;

		return;
	}

	static Jset_family get_mins(final int[] c) {
		final Jcube_struct cube = Jglobals.cube;// java

		final Jset_family minterms = Jespresso.new_cover( 1 );
		int[] d_minterm = Jespresso.new_cube();
		Jset.set_copy( d_minterm, c );
		Jset.set_and( d_minterm, d_minterm, cube.binary_mask );
		for( int i = cube.num_binary_vars; i < cube.num_vars; i++ ) {
			for( int j = cube.first_part[i]; j <= cube.last_part[ i ]; j++ ) {
				if( Jespresso.is_in_set( c, j ) ) {
					Jespresso.set_insert( d_minterm, j );
					Jset.sf_addset( minterms, d_minterm );
					Jespresso.set_remove( d_minterm, j );
				}
			}
		}
		d_minterm = null;
		return minterms;
	}

	@SuppressWarnings("boxing")
	static void print_list(final int n, final int[] x, final String name) {
		System.out.printf("%s:\n",name);
		for( int i = 0; i < n; i++ ) {
			System.out.printf("%d%c", x[ i ], ((i + 1) % 10) != 0 ? '\t' : '\n');
		}
		System.out.printf("\n");
	}
}