package espresso;

import java.io.PrintStream;

final class Jexact {

	/**
	 *  minimize_exact -- main entry point for exact minimization
	 *
	 *  Global flags which affect this routine are:
	 *
	 *      debug
	 *      skip_make_sparse
	 */
	static Jset_family minimize_exact(final Jset_family F, final Jset_family D, final Jset_family R, final boolean exact_cover) {
		return do_minimize( F, D, R, exact_cover, /*weighted*/ false );
	}

	static Jset_family minimize_exact_literals(final Jset_family F, final Jset_family D, final Jset_family R, final boolean exact_cover) {
		return do_minimize( F, D, R, exact_cover, /*weighted*/ true );
	}

	private static Jset_family do_minimize(Jset_family F, final Jset_family D, final Jset_family R, final boolean exact_cover, final boolean weighted) {
		final int debug_save = Jglobals.debug;

		if( (Jglobals.debug & Jespresso.EXACT) != 0 ) {
			Jglobals.debug |= (Jespresso.IRRED | Jespresso.MINCOV);
		}
// if defined(sun) || defined(bsd4_2)			/* hack ... */
//		if( (Jglobals.debug & Jespresso.MINCOV) != 0 ) {
//			setlinebuf( System.out );
//		}
// endif
		final int level = (Jglobals.debug & Jespresso.MINCOV) != 0 ? 4 : 0;
		final boolean heur = ! exact_cover;

		/* Generate all prime implicants */
	    {
			final long t = System.currentTimeMillis();
			F = Jprimes.primes_consensus( Jcofactor.cube2list( F, D ) );
			if( Jglobals.trace ) {
				Jcvrmisc.print_trace( F, "PRIMES     ", System.currentTimeMillis() - t );
			}
		}

		/* Setup the prime implicant table */
	    final Jset_family E, Rt, Rp;
	    {
	    	final Jset_family[] pE = new Jset_family[ 1 ];// java helper
	    	final Jset_family[] pRt = new Jset_family[ 1 ];// java helper
	    	final Jset_family[] pRp = new Jset_family[ 1 ];// java helper
			final long t = System.currentTimeMillis();
			Jirred.irred_split_cover( F, D, pE, pRt, pRp );
			E = pE[ 0 ];// java
			Rt = pRt[ 0 ];// java
			Rp = pRp[ 0 ];// java
			if( Jglobals.trace ) {
				Jcvrmisc.print_trace( E, "ESSENTIALS ", System.currentTimeMillis() - t );
			}
		}
	    final Jsm_matrix table;
	    {
			final long t = System.currentTimeMillis();
			table = Jirred.irred_derive_table( D, E, Rp );
			if( Jglobals.trace ) {
				Jcvrmisc.print_trace( Rp, "PI-TABLE   ", System.currentTimeMillis() - t );
			}
		}

		/* Solve either a weighted or nonweighted covering problem */
		int[] weights;
		if( weighted ) {
			/* correct only for all 2-valued variables */
			weights = new int[ F.count ];
			final int[] data = Rp.data;// java
		    for( int p = 0, last = p + Rp.count * Rp.wsize; p < last; p += Rp.wsize ) {
				weights[ Jespresso.SIZE( data, p ) ] = Jglobals.cube.size - Jset.set_ord( data, p );
			}
		} else {
			weights = null;
		}
		final Jsm_row cover;
	    {
			final long t = System.currentTimeMillis();
			cover = Jmincov.sm_minimum_cover( table, weights, heur, level );
			if( Jglobals.trace ) {
				Jcvrmisc.print_trace( F, "MINCOV     ", System.currentTimeMillis() - t );
			}
		}
		weights = null;

		if( (Jglobals.debug & Jespresso.EXACT) != 0 ) {
			dump_irredundant( E, Rt, Rp, table );
		}

		/* Form the result cover */
		Jset_family newF = Jespresso.new_cover( 100 );
		final int[] data = E.data;// java
	    for( int p = 0, last = p + E.count * E.wsize; p < last; p += E.wsize ) {
			newF = Jset.sf_addset( newF, data, p );
		}
		for( Jsm_element pe = cover.first_col; pe != null; pe = pe.next_col ) {
			newF = Jset.sf_addset( newF, F.data, F.wsize * pe.col_num );
		}

		Jespresso.free_cover( E );
		Jespresso.free_cover( Rt );
		Jespresso.free_cover( Rp );
		Jmatrix.sm_free( table );
		Jrows.sm_row_free( cover );
		Jespresso.free_cover( F );

		/* Attempt to make the results more sparse */
		Jglobals.debug &= ~ (Jespresso.IRRED | Jespresso.SHARP | Jespresso.MINCOV);
		if( ! Jglobals.skip_make_sparse && R != null ) {
			newF = Jsparse.make_sparse( newF, D, R );
		}

		Jglobals.debug = debug_save;
		return newF;
	}

	private static void dump_irredundant(final Jset_family E, final Jset_family Rt, final Jset_family Rp, final Jsm_matrix table) {
		PrintStream fp_pi_table = null, fp_primes = null;
		if( Jglobals.filename == null || Jglobals.filename.compareTo("(stdin)") == 0 ) {
			fp_pi_table = fp_primes = System.out;
		} else {
			String file = String.format( "%s.primes", Jglobals.filename );
			try {
				fp_primes = new PrintStream( file, "UTF-8" );
			} catch( final Exception fe ) {
				System.err.printf("espresso: Unable to open %s\n", file );
				fp_primes = System.out;
			}
			file = String.format( "%s.pi", Jglobals.filename );
			try {
				fp_pi_table = new PrintStream( file, "UTF-8" );
			} catch( final Exception fe ) {
				System.err.printf("espresso: Unable to open %s\n", file );
				fp_pi_table = System.out;
			}
		}

		final JpPLA PLA = Jcvrin.new_PLA();
		Jcvrin.PLA_labels( PLA );

		Jcvrout.fpr_header( fp_primes, PLA, Jespresso.F_type );
		Jcvrin.free_PLA( PLA );

		fp_primes.printf("# Essential primes are\n");

		int[] data = E.data;// java
	    for( int p = 0, last = p + E.count * E.wsize; p < last; p += E.wsize ) {
	    	fp_primes.printf("%s\n", Jcvrout.pc1( data, p ) );
		}
		fp_primes.printf("# Totally redundant primes are\n");

		data = Rt.data;// java
	    for( int p = 0, last = p + Rt.count * Rt.wsize; p < last; p += Rt.wsize ) {
	    	fp_primes.printf("%s\n", Jcvrout.pc1( data, p ) );
		}
		fp_primes.printf("# Partially redundant primes are\n");

		data = Rp.data;// java
		for( int p = 0, last = p + Rp.count * Rp.wsize; p < last; p += Rp.wsize ) {
			fp_primes.printf("%s\n", Jcvrout.pc1( data, p ) );
		}
		if( fp_primes != System.out ) {
			fp_primes.close();
		}

		Jmatrix.sm_write( fp_pi_table, table );
		if( fp_pi_table != System.out ) {
			fp_pi_table.close();
		}
	}
}