package espresso;
/*
    module: expand.c
    purpose: Perform the Espresso-II Expansion Step

    The idea is to take each nonprime cube of the on-set and expand it
    into a prime implicant such that we can cover as many other cubes
    of the on-set.  If no cube of the on-set can be covered, then we
    expand each cube into a large prime implicant by transforming the
    problem into a minimum covering problem which is solved by the
    heuristics of minimum_cover.

    These routines revolve around having a representation of the
    OFF-set.  (In contrast to the Espresso-II manuscript, we do NOT
    require an "unwrapped" version of the OFF-set).

    Some conventions on variable names:

	SUPER_CUBE is the supercube of all cubes which can be covered
	by an expansion of the cube being expanded

	OVEREXPANDED_CUBE is the cube which would result from expanding
	all parts which can expand individually of the cube being expanded

	RAISE is the current expansion of the current cube

	FREESET is the set of parts which haven't been raised or lowered yet.

	INIT_LOWER is a set of parts to be removed from the free parts before
	starting the expansion
*/

import java.util.Comparator;

final class Jexpand {

	/**
	expand -- expand each nonprime cube of F into a prime implicant

	If nonsparse is true, only the non-sparse variables will be expanded;
	this is done by forcing all of the sparse variables out of the free set.
	 *
	 * @param F
	 * @param R
	 * @param nonsparse expand non-sparse variables only
	 * @return
	 */
	@SuppressWarnings("boxing")
	static Jset_family expand(Jset_family F, final Jset_family R, final boolean nonsparse) {

		/* Order the cubes according to "chewing-away from the edges" of mini */
		if( Jglobals.use_random_order ) {
			F = Jcvrm.random_order( F );
		} else {
			final Comparator<Jset> ascend = new Jascend_order();
			F = Jcvrm.mini_sort( F, ascend );
		}

		/* Allocate memory for variables needed by expand1() */
		int[] RAISE = Jespresso.new_cube();
		int[] FREESET = Jespresso.new_cube();
		int[] INIT_LOWER = Jespresso.new_cube();
		int[] SUPER_CUBE = Jespresso.new_cube();
		int[] OVEREXPANDED_CUBE = Jespresso.new_cube();

		/* Setup the initial lowering set (differs only for nonsparse) */
		if( nonsparse ) {
			final Jcube_struct cube = Jglobals.cube;// java
			for( int var = 0; var < cube.num_vars; var++ ) {
				if( cube.sparse[ var ] ) {
					Jset.set_or( INIT_LOWER, INIT_LOWER, cube.var_mask[ var ] );
				}
			}
		}

		/* Mark all cubes as not covered, and maybe essential */
		final int[] data = F.data;// java
		for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			Jespresso.RESET( data, p, Jespresso.COVERED );
			Jespresso.RESET( data, p, Jespresso.NONESSEN );
		}

		/* Try to expand each nonprime and noncovered cube */
		final int num_covered[] = new int[1];// java holder

		for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			/* do not expand if PRIME or if covered by previous expansion */
			if( ! Jespresso.TESTP( data, p, Jespresso.PRIME ) && ! Jespresso.TESTP( data, p, Jespresso.COVERED ) ) {

				/* expand the cube p, result is RAISE */
				Jexpand.expand1( R, F, RAISE, FREESET, OVEREXPANDED_CUBE, SUPER_CUBE, INIT_LOWER, num_covered, data, p );
				if( (Jglobals.debug & Jespresso.EXPAND) != 0 ) {
					System.out.printf("EXPAND: %s (covered %d)\n", Jcvrout.pc1( data, p ), num_covered[0] );
				}
				Jset.set_copy( data, p, RAISE );
				Jespresso.SET( data, p, Jespresso.PRIME );
				Jespresso.RESET( data, p, Jespresso.COVERED );		/* not really necessary */

				/* See if we generated an inessential prime */
				if( num_covered[0] == 0 && ! Jset.setp_equal( data, p, OVEREXPANDED_CUBE ) ) {
					Jespresso.SET( data, p, Jespresso.NONESSEN );
				}
			}
		}

		/* Delete any cubes of F which became covered during the expansion */
		F.active_count = 0;
		boolean change = false;

		for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			if( Jespresso.TESTP( data, p, Jespresso.COVERED ) ) {
				Jespresso.RESET( data, p, Jespresso.ACTIVE );
				change = true;
			} else {
				Jespresso.SET( data, p, Jespresso.ACTIVE );
				F.active_count++;
			}
		}
		if( change ) {
			F = Jset.sf_inactive( F );
		}

		RAISE = null;
		FREESET = null;
		INIT_LOWER = null;
		SUPER_CUBE = null;
		OVEREXPANDED_CUBE = null;
		return F;
	}

	/**
	expand1 -- Expand a single cube against the OFF-set
	 *
	 *
	 * @param BB Blocking matrix (OFF-set)
	 * @param CC Covering matrix (ON-set)
	 * @param RAISE The current parts which have been raised
	 * @param FREESET The current parts which are free
	 * @param OVEREXPANDED_CUBE Overexpanded cube of c
	 * @param SUPER_CUBE Supercube of all cubes of CC we cover
	 * @param INIT_LOWER Parts to initially remove from FREESET
	 * @param num_covered Number of cubes of CC which are covered
	 * @param c The cube to be expanded
	 */
	static void expand1(final Jset_family BB, final Jset_family CC, final int[] RAISE, final int[] FREESET, final int[] OVEREXPANDED_CUBE, final int[] SUPER_CUBE,
		final int[] INIT_LOWER, final int[] num_covered, final int[] c, final int off_c)
	{
		if( (Jglobals.debug & Jespresso.EXPAND1) != 0 ) {
			System.out.printf("\nEXPAND1:    \t%s\n", Jcvrout.pc1( c, off_c ) );
		}

		/* initialize BB and CC */
		Jespresso.SET( c, off_c, Jespresso.PRIME );		/* don't try to cover ourself */
		setup_BB_CC( BB, CC );

		/* initialize count of # cubes covered, and the supercube of them */
		num_covered[0] = 0;
		Jset.set_copy( SUPER_CUBE, c, off_c );

		/* Initialize the lowering, raising and unassigned sets */
		Jset.set_copy( RAISE, c, off_c );
		Jset.set_diff( FREESET, Jglobals.cube.fullset, RAISE );

		/* If some parts are forced into lowering set, remove them */
		if( ! Jset.setp_empty( INIT_LOWER ) ) {
			Jset.set_diff( FREESET, FREESET, INIT_LOWER );
			elim_lowering( BB, CC, RAISE, FREESET );
		}

		/* Determine what can be raised, and return the over-expanded cube */
		essen_parts( BB, CC, RAISE, FREESET );
		Jset.set_or( OVEREXPANDED_CUBE, RAISE, FREESET );

		/* While there are still cubes which can be covered, cover them ! */
		if( CC.active_count > 0 ) {
			select_feasible( BB, CC, RAISE, FREESET, SUPER_CUBE, num_covered );
		}

		/* While there are still cubes covered by the overexpanded cube ... */
		while( CC.active_count > 0 ) {
			final int bestindex = most_frequent( CC, FREESET );
			Jespresso.set_insert( RAISE, bestindex );
			Jespresso.set_remove( FREESET, bestindex );
			essen_parts( BB, CC, RAISE, FREESET );
		}

		/* Finally, when all else fails, choose the largest possible prime */
		/* We will loop only if we decide unravelling OFF-set is too expensive */
		while( BB.active_count > 0 ) {
			mincov( BB, RAISE, FREESET );
		}

		/* Raise any remaining free coordinates */
		Jset.set_or( RAISE, RAISE, FREESET );
	}

	/**
	essen_parts -- determine which parts are forced into the lowering
	set to insure that the cube be orthognal to the OFF-set.

	If any cube of the OFF-set is distance 1 from the raising cube,
	then we must lower all parts of the conflicting variable.  (If the
	cube is distance 0, we detect this error here.)

	If there are essentially lowered parts, we can remove from consideration
	any cubes of the OFF-set which are more than distance 1 from the
	overexpanded cube of RAISE.
	*/
	static void essen_parts(final Jset_family BB, final Jset_family CC, final int[] RAISE, final int[] FREESET) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] r = RAISE;
		final int[] xlower = cube.temp[0];
		Jset.set_copy( xlower, cube.emptyset );

		int dist;
		final int[] data = BB.data;// java
next_label:
		for( int p = 0, lastp = p + BB.count * BB.wsize; p < lastp; p += BB.wsize ) {
			if( Jespresso.TESTP( data, p, Jespresso.ACTIVE ) ) {
if( Jespresso.NO_INLINE ) {
				if( (dist = Jsetc.cdist01( data, p, r, 0 )) > 1 ) {
					continue;// goto exit_if;
				}
} else {
				{
					int last;
					dist = 0;
					if( (last = cube.inword) != -1 ) {
						int x = data[ p + last ] & r[ last ];
						if( (x = ~(x | x >>> 1 ) & cube.inmask) != 0 ) {
							if( (dist = Jespresso.count_ones( x ) ) > 1 ) {
								continue next_label;// goto exit_if;
							}
						}
						for( int w = 1; w < last; w++ ) {
							x = data[ p + w ] & r[ w ];
							if( (x = ~(x | x >>> 1) & Jespresso.DISJOINT) != 0 ) {
								if( dist == 1 || (dist += Jespresso.count_ones( x )) > 1 ) {
									continue next_label;// goto exit_if;
								}
							}
						}
					}
				}
				{
nextvar_label:
					for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
						final int[] mask = cube.var_mask[ var ];
						final int last = cube.last_word[ var ];
						for( int w = cube.first_word[ var ]; w <= last; w++ ) {
							if( (data[ p + w ] & r[ w ] & mask[ w ]) != 0 ) {
								continue nextvar_label;// goto nextvar;
							}
						}
						if( ++dist > 1 ) {
							continue next_label;// goto exit_if;
						}
			// nextvar:
							;
					}
				}
}
				if( dist == 0 ) {
					Jcvrmisc.fatal("ON-set and OFF-set are not orthogonal");
				} else {
					Jsetc.force_lower( xlower, 0, data, p, r, 0 );
					BB.active_count--;
					Jespresso.RESET( data, p, Jespresso.ACTIVE );
				}
//	exit_if: ;
			}
		}

		if( ! Jset.setp_empty( xlower ) ) {
			Jset.set_diff( FREESET, FREESET, xlower );/* remove from free set */
			elim_lowering( BB, CC, RAISE, FREESET );
		}

		if( (Jglobals.debug & Jespresso.EXPAND1) != 0 ) {
			System.out.printf("ESSEN_PARTS:\tRAISE=%s FREESET=%s\n", Jcvrout.pc1( RAISE, 0 ), Jcvrout.pc2( FREESET ) );
		}
	}

	/**
	essen_raising -- determine which parts may always be added to
	the raising set without restricting further expansions

	General rule: if some part is not blocked by any cube of BB, then
	this part can always be raised.
	*/
	static void essen_raising(final Jset_family BB, final int[] RAISE, final int[] FREESET) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] xraise = cube.temp[0];

		/* Form union of all cubes of BB, and then take complement wrt FREESET */
		Jset.set_copy( xraise, cube.emptyset );

		final int[] data = BB.data;// java
		for( int p = 0, last = p + BB.count * BB.wsize; p < last; p += BB.wsize ) {
			if( Jespresso.TESTP( data, p, Jespresso.ACTIVE ) ) {
				Jset.set_or( xraise, xraise, data, p );
			}
		}
		Jset.set_diff( xraise, FREESET, xraise );

		Jset.set_or( RAISE, RAISE, xraise );         /* add to raising set */
		Jset.set_diff( FREESET, FREESET, xraise );       /* remove from free set */

		if( (Jglobals.debug & Jespresso.EXPAND1) != 0 ) {
			System.out.printf("ESSEN_RAISING:\tRAISE=%s FREESET=%s\n", Jcvrout.pc1( RAISE, 0 ), Jcvrout.pc2( FREESET ) );
		}
	}

	/**
	elim_lowering -- after removing parts from FREESET, we can reduce the
	size of both BB and CC.

	We mark as inactive any cube of BB which does not intersect the
	overexpanded cube (i.e., RAISE + FREESET).  Likewise, we remove
	from CC any cube which is not covered by the overexpanded cube.
	*/
	static void elim_lowering(final Jset_family BB, final Jset_family CC, final int[] RAISE, final int[] FREESET) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] r = cube.temp[0];
		Jset.set_or( r, RAISE, FREESET );

		/* Remove sets of BB which are orthogonal to future expansions */
		int[] data = BB.data;// java
next_label:
		for( int p = 0, last = p + BB.count * BB.wsize; p < last; p += BB.wsize ) {
			if( Jespresso.TESTP( data, p, Jespresso.ACTIVE ) ) {
if( Jespresso.NO_INLINE ) {
				if( ! Jsetc.cdist0( data, p, r, 0 ) ) {
					BB.active_count--;
					Jespresso.RESET( data, p, Jespresso.ACTIVE );
				}
} else {
				{
					int lastw;
					if( (lastw = cube.inword ) != -1 ) {
						int x = data[ p + lastw ] & r[ lastw ];
						if( (~(x | x >>> 1 ) & cube.inmask) != 0 ) {
							BB.active_count--;
							Jespresso.RESET( data, p, Jespresso.ACTIVE );
							continue next_label;
							// goto false;
						}
						for( int w = 1; w < lastw; w++ ) {
							x = data[ p + w ] & r[ w ];
							if( (~(x | x >>> 1 ) & Jespresso.DISJOINT) != 0 ) {
								BB.active_count--;
								Jespresso.RESET( data, p, Jespresso.ACTIVE );
								continue next_label;
								// goto false;
							}
						}
					}
				}
				{
nextvar_label:
					for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
						final int[] mask = cube.var_mask[ var ];
						final int lastw = cube.last_word[ var ];
						for( int w = cube.first_word[ var ]; w <= lastw; w++ ) {
							if( (data[ p + w ] & r[ w ] & mask[ w ]) != 0 ) {
								continue nextvar_label;// goto nextvar;
							}
						}
						BB.active_count--;
						Jespresso.RESET( data, p, Jespresso.ACTIVE );
						continue next_label;
						// goto false;
	// nextvar:			;
					}
				}
				// continue;
	// false:
}
				// BB.active_count--;
				// Jespresso.RESET( data, p, Jespresso.ACTIVE );
			}
		}

		/* Remove sets of CC which cannot be covered by future expansions */
		if( CC != null ) {
			data = CC.data;// java
			for( int p = 0, last = p + CC.count * CC.wsize; p < last; p += CC.wsize ) {
				if( Jespresso.TESTP( data, p, Jespresso.ACTIVE ) ) {
					if( ! Jset.setp_implies( data, p, r ) ) {
						CC.active_count--;
						Jespresso.RESET( data, p, Jespresso.ACTIVE );
					}
				}
			}
		}
	}

	/**
	most_frequent -- When all else fails, select a reasonable part to raise
	The active cubes of CC are the cubes which are covered by the
	overexpanded cube of the original cube (however, we know that none
	of them can actually be covered by a feasible expansion of the
	original cube).  We resort to the MINI strategy of selecting to
	raise the part which will cover the same part in the most cubes of CC.
	*/
	@SuppressWarnings("boxing")
	static int most_frequent(final Jset_family CC, final int[] FREESET) {
		final Jcube_struct cube = Jglobals.cube;// java

		/* Count occurences of each variable */
		int[] count = new int[ cube.size ];// java: already zeroed
		//for( int i = 0; i < cube.size; i++ ) {
		//	count[i] = 0;
		//}
		if( CC != null ) {
			final int[] data = CC.data;// java
			for( int p = 0, last = p + CC.count * CC.wsize; p < last; p += CC.wsize ) {
				if( Jespresso.TESTP( data, p, Jespresso.ACTIVE ) ) {
					Jset.set_adjcnt( data, p, count, 1 );
				}
			}
		}

		/* Now find which free part occurs most often */
		int best_part, best_count;
		best_count = best_part = -1;
		for( int i = 0; i < cube.size; i++ ) {
			if( Jespresso.is_in_set( FREESET, i ) && count[i] > best_count ) {
				best_part = i;
				best_count = count[ i ];
			}
		}
		count = null;

		if( (Jglobals.debug & Jespresso.EXPAND1) != 0 ) {
			System.out.printf("MOST_FREQUENT:\tbest=%d FREESET=%s\n", best_part, Jcvrout.pc2( FREESET ) );
		}
		return best_part;
	}

	/**
	setup_BB_CC -- set up the blocking and covering set families;

	Note that the blocking family is merely the set of cubes of R, and
	that CC is the set of cubes of F which might possibly be covered
	(i.e., nonprime cubes, and cubes not already covered)
	*/
	static void setup_BB_CC(final Jset_family BB, final Jset_family CC) {
		/* Create the block and cover set families */
		BB.active_count = BB.count;
		int[] data = BB.data;// java
		for( int p = 0, last = p + BB.count * BB.wsize; p < last; p += BB.wsize ) {
			Jespresso.SET( data, p, Jespresso.ACTIVE );
		}

		if( CC != null ) {
			CC.active_count = CC.count;

			data = CC.data;// java
			for( int p = 0, last = p + CC.count * CC.wsize; p < last; p += CC.wsize ) {
				if( Jespresso.TESTP( data, p, Jespresso.COVERED ) || Jespresso.TESTP( data, p, Jespresso.PRIME ) ) {
					CC.active_count--;
					Jespresso.RESET( data, p, Jespresso.ACTIVE );
				} else {
					Jespresso.SET( data, p, Jespresso.ACTIVE );
				}
			}
		}
	}

	/**
	select_feasible -- Determine if there are cubes which can be covered,
	and if so, raise those parts necessary to cover as many as possible.

	We really don't check to maximize the number that can be covered;
	instead, we check, for each fcc, how many other fcc remain fcc
	after expanding to cover the fcc.  (Essentially one-level lookahead).
	*/
	@SuppressWarnings("boxing")
	static void select_feasible(final Jset_family BB, final Jset_family CC, final int[] RAISE, final int[] FREESET, final int[] SUPER_CUBE, final int[] num_covered) {

		/*  Start out with all cubes covered by the over-expanded cube as
		*  the "possibly" feasibly-covered cubes (pfcc)
		*/
		Jset[] feas = new Jset[ CC.active_count ];
		int numfeas = 0;

		final int[] data = CC.data;// java
		for( int p = 0, last = p + CC.count * CC.wsize; p < last; p += CC.wsize ) {
			if( Jespresso.TESTP( data, p, Jespresso.ACTIVE ) ) {
				feas[ numfeas++ ] = new Jset( data, p );
			}
		}

		/* Setup extra cubes to record parts forced low after a covering */
		Jset[] feas_new_lower = new Jset[ CC.active_count ];
		final Jset_family new_lower = Jespresso.new_cover( numfeas );
		for( int i = 0; i < numfeas; i++ ) {
			feas_new_lower[i] = new Jset( new_lower.data, new_lower.wsize * i );// GETSET(new_lower, i);
		}

// loop:
		do {
			/* Find the essentially raised parts -- this might cover some cubes
			for us, without having to find out if they are fcc or not
			*/
			essen_raising( BB, RAISE, FREESET );

			/* Now check all "possibly" feasibly covered cubes to check feasibility */
			final int lastfeas = numfeas;
			numfeas = 0;
			for( int i = 0; i < lastfeas; i++ ) {
				final Jset p = feas[ i ];

				/* Check active because essen_parts might have removed it */
				if( Jespresso.TESTP( p, Jespresso.ACTIVE ) ) {

					/*  See if the cube is already covered by RAISE --
					*  this can happen because of essen_raising() or because of
					*  the previous "loop"
					*/
					if( Jset.setp_implies( p.base, p.offset, RAISE ) ) {
						num_covered[0]++;
						Jset.set_or( SUPER_CUBE, SUPER_CUBE, p.base, p.offset );
						CC.active_count--;
						Jespresso.RESET( p, Jespresso.ACTIVE );
						Jespresso.SET( p, Jespresso.COVERED );
						/* otherwise, test if it is feasibly covered */
					} else {
						final Jset t = feas_new_lower[ numfeas ];
						if( Jexpand.feasibly_covered( BB, p.base, p.offset, RAISE, t.base, t.offset ) ) {
							feas[ numfeas ] = p;			/* save the fcc */
							numfeas++;
						}
					}
				}
			}
			if( (Jglobals.debug & Jespresso.EXPAND1) != 0 ) {
				System.out.printf("SELECT_FEASIBLE: started with %d pfcc, ended with %d fcc\n",
						lastfeas, numfeas );
			}

			/* Exit here if there are no feasibly covered cubes */
			if( numfeas == 0) {
				feas = null;
				feas_new_lower = null;
				Jespresso.free_cover( new_lower );
				return;
			}

			/* Now find which is the best feasibly covered cube */
			Jset bestfeas = feas[0];// java: to suppress the warning may have been not initialized
			int bestcount = 0;
			int bestsize = 9999;
			for( int i = 0; i < numfeas; i++ ) {
				Jset t = feas[ i ];
				final int size = Jset.set_dist( t.base, t.offset, FREESET );	/* # of newly raised parts */
				int count = 0;	/* # of other cubes which remain fcc after raising */

	// #define NEW
	// #ifdef NEW
				for( int j = 0; j < numfeas; j++ ) {
					t = feas_new_lower[i];
					final Jset t1 = feas[j];
					if( Jset.setp_disjoint( t.base, t.offset, t1.base, t1.offset ) ) {
						count++;
					}
				}
	// #else
	//			for( int j = 0; j < numfeas; j++ )
	//				if( Jset.setp_implies( feas[j], feas[i] ) )
	//					count++;
	// #endif
				if( count > bestcount ) {
					bestcount = count;
					bestfeas = feas[ i ];
					bestsize = size;
				} else if( count == bestcount && size < bestsize ) {
					bestfeas = feas[ i ];
					bestsize = size;
				}
			}

			/* Add the necessary parts to the raising set */
			Jset.set_or( RAISE, RAISE, bestfeas.base, bestfeas.offset );
			Jset.set_diff( FREESET, FREESET, RAISE );
			if( (Jglobals.debug & Jespresso.EXPAND1) != 0 ) {
				System.out.printf("FEASIBLE:  \tRAISE=%s FREESET=%s\n", Jcvrout.pc1( RAISE, 0 ), Jcvrout.pc2( FREESET ) );
			}
			essen_parts( BB, CC, RAISE, FREESET );
			// goto loop;
		} while( true );
		/* NOTREACHED */
	}

	/**
	feasibly_covered -- determine if the cube c is feasibly covered
	(i.e., if it is possible to raise all of the necessary variables
	while still insuring orthogonality with R).  Also, if c is feasibly
	covered, then compute the new set of parts which are forced into
	the lowering set.
	*/
	static boolean feasibly_covered(final Jset_family BB, final int[] c, final int off_c, final int[] RAISE, final int[] new_lower, final int off_lower) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] r = cube.temp[ 0 ];
		Jset.set_or( r, RAISE, c, off_c );
		Jset.set_copy( new_lower, off_lower, cube.emptyset );
		int dist;
		final int[] data = BB.data;// java
exit_if_label:
		for( int p = 0, lastp = p + BB.count * BB.wsize; p < lastp; p += BB.wsize ) {
			if( Jespresso.TESTP( data, p, Jespresso.ACTIVE ) ) {
if( Jespresso.NO_INLINE ) {
				if( (dist = Jsetc.cdist01( data, p, r, 0 )) > 1 ) {
					continue;// goto exit_if;
				}
} else {
				{
					int last;
					dist = 0;
					if( (last = cube.inword) != -1 ) {
						int x = data[ p + last ] & r[ last ];
						if( (x = ~(x | x >>> 1) & cube.inmask) != 0 ) {
							if( (dist = Jespresso.count_ones( x )) > 1 ) {
								continue;// goto exit_if;
							}
						}
						for( int w = 1; w < last; w++ ) {
							x = data[ p + w ] & r[ w ];
							if( (x = ~(x | x >>> 1) & Jespresso.DISJOINT) != 0 ) {
								if( dist == 1 || (dist += Jespresso.count_ones( x )) > 1 ) {
									continue exit_if_label;// goto exit_if;
								}
							}
						}
					}
				}
				{
nextvar_label:
					for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
						final int[] mask = cube.var_mask[ var ];
						final int last = cube.last_word[ var ];
						for( int w = cube.first_word[ var ]; w <= last; w++ ) {
							if( (data[ p + w ] & r[ w ] & mask[ w ]) != 0 )  {
								continue nextvar_label;// goto nextvar;
							}
						}
						if( ++dist > 1 ) {
							continue exit_if_label;// goto exit_if;
// nextvar:		;
						}
					}
				}
}
				if( dist == 0 ) {
					return false;
				}// else {
					Jsetc.force_lower( new_lower, off_lower, data, p, r, 0 );
//	exit_if: ;
				//}
			}
		}
		return true;
	}

	/**
	mincov -- transform the problem of expanding a cube to a maximally-
	large prime implicant into the problem of selecting a minimum
	cardinality cover over a family of sets.

	When we get to this point, we must unravel the remaining off-set.
	This may be painful.
	*/
	static void mincov(final Jset_family BB, final int[] RAISE, final int[] FREESET) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] xraise = cube.temp[0];

if( Jespresso.RANDOM_MINCOV ) {
		int dist = Jespresso.random.nextInt() % Jset.set_ord( FREESET );
		int var;
		for( var = 0; var < cube.size && dist >= 0; var++) {
			if( Jespresso.is_in_set( FREESET, var ) ) {
				dist--;
			}
		}

		Jespresso.set_insert( RAISE, var );
		Jespresso.set_remove( FREESET, var );
		essen_parts( BB, /*CC*/ null, RAISE, FREESET );
} else {

		/* Create B which are those cubes which we must avoid intersecting */
		Jset_family B = Jespresso.new_cover( BB.active_count );
		int[] data = BB.data;// java
		for( int p = 0, last = p + BB.count * BB.wsize; p < last; p += BB.wsize ) {
			if( Jespresso.TESTP( data, p, Jespresso.ACTIVE ) ) {
				final int plower = B.wsize * B.count++;
				Jset.set_copy( B.data, plower, cube.emptyset );
				Jsetc.force_lower( B.data, plower, data, p, RAISE, 0 );
			}
		}

		/* Determine how many sets it will blow up into after the unravel */
		int nset = 0;
		data = B.data;// java
		for( int p = 0, last = p + B.count * B.wsize; p < last; p += B.wsize ) {
			int expansion = 1;
			for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
				int dist;
				if( (dist = Jset.set_dist( data, p, cube.var_mask[ var ])) > 1 ) {
					expansion *= dist;
					if( expansion > 500 ) {
						// goto heuristic_mincov;
						Jset.sf_free( B );
						/* most_frequent will pick first free part */
						Jespresso.set_insert( RAISE, most_frequent(/*CC*/ null, FREESET ) );
						Jset.set_diff( FREESET, FREESET, RAISE );
						essen_parts( BB, /*CC*/ null, RAISE, FREESET );
						return;
					}
				}
			}
			nset += expansion;
			if( nset > 500 ) {
				// goto heuristic_mincov;
				Jset.sf_free( B );
				/* most_frequent will pick first free part */
				Jespresso.set_insert( RAISE, most_frequent(/*CC*/ null, FREESET ) );
				Jset.set_diff( FREESET, FREESET, RAISE );
				essen_parts( BB, /*CC*/ null, RAISE, FREESET );
				return;
			}
		}

		B = Jcvrm.unravel( B, cube.num_binary_vars );
		int[] xlower = Jsminterf.do_sm_minimum_cover( B );

		/* Add any remaining free parts to the raising set */
		Jset.set_diff( xraise, FREESET, xlower );
		Jset.set_or( RAISE, RAISE, xraise );
		Jset.set_copy( FREESET, cube.emptyset );	/* free set is empty */
		BB.active_count = 0;			/* BB satisfied */
		if( (Jglobals.debug & Jespresso.EXPAND1) != 0 ) {
			System.out.printf("MINCOV:    \tRAISE=%s FREESET=%s\n", Jcvrout.pc1( RAISE, 0 ), Jcvrout.pc2( FREESET ) );
		}
		Jset.sf_free( B );
		xlower = null;
		return;

// heuristic_mincov:
//		Jset.sf_free( B );
		/* most_frequent will pick first free part */
//		Jset.set_insert( RAISE, most_frequent(/*CC*/ null, FREESET ) );
//		Jset.set_diff( FREESET, FREESET, RAISE );
//		essen_parts( BB, /*CC*/ null, RAISE, FREESET );
//		return;
}
	}

	/**
	find_all_primes -- find all of the primes which cover the
	currently reduced BB
	*/
	static Jset_family find_all_primes(final Jset_family BB, final int[] RAISE, final int[] FREESET) {
		Jset_family B1;

		if( BB.active_count == 0) {
			B1 = Jespresso.new_cover( 1 );
			final int[] p_base = B1.data;// GETSET(B1, B1.count++);
			final int p = B1.wsize * B1.count++;
			Jset.set_copy( p_base, p, RAISE );
			Jespresso.SET( p_base, p, Jespresso.PRIME );
		} else {
			final Jcube_struct cube = Jglobals.cube;// java
			Jset_family B = Jespresso.new_cover( BB.active_count );
			int data[] = BB.data;// java
			for( int p = 0, last = p + BB.count * BB.wsize; p < last; p += BB.wsize ) {
				if( Jespresso.TESTP( data, p, Jespresso.ACTIVE ) ) {
					final int plower = B.wsize * B.count++;
					Jset.set_copy( B.data, plower, cube.emptyset );
					Jsetc.force_lower( B.data, plower, data, p, RAISE, 0 );
				}
			}
			B = Jcontain.sf_rev_contain( Jcvrm.unravel( B, cube.num_binary_vars ) );
			B1 = Junate.exact_minimum_cover( B );
			data = B1.data;// java
			for( int p = 0, last = p + B1.count * B1.wsize; p < last; p += B1.wsize ) {
				Jset.set_diff( data, p, FREESET, data, p );
				Jset.set_or( data, p, data, p, RAISE );
				Jespresso.SET( data, p, Jespresso.PRIME );
			}
			Jespresso.free_cover( B );
		}
		return B1;
	}

	/**
	all_primes -- foreach cube in F, generate all of the primes
	which cover the cube.
	*/
	static Jset_family all_primes(final Jset_family F, final Jset_family R) {
		final Jcube_struct cube = Jglobals.cube;// java
		int[] FREESET = Jespresso.new_cube();
		int[] RAISE = Jespresso.new_cube();
		Jset_family Fall_primes = Jespresso.new_cover( F.count );

		final int[] data = F.data;// java
		for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			if( Jespresso.TESTP( data, p, Jespresso.PRIME ) ) {
				Fall_primes = Jset.sf_addset( Fall_primes, data, p );
			} else {
				/* Setup for call to essential parts */
				Jset.set_copy( RAISE, data, p );
				Jset.set_diff( FREESET, cube.fullset, RAISE );
				setup_BB_CC(R, /* CC */ null );
				essen_parts(R, /* CC */ null, RAISE, FREESET );

				/* Find all of the primes, and add them to the prime set */
				final Jset_family B1 = find_all_primes( R, RAISE, FREESET );
				Fall_primes = Jset.sf_append( Fall_primes, B1 );
			}
		}

		RAISE = null;
		FREESET = null;
		return Fall_primes;
	}
}