package espresso;

/**
    module: gasp.c

    The "last_gasp" heuristic computes the reduction of each cube in
    the cover (without replacement) and then performs an expansion of
    these cubes.  The cubes which expand to cover some other cube are
    added to the original cover and irredundant finds a minimal subset.

    If one of the reduced cubes expands to cover some other reduced
    cube, then the new prime thus generated is a candidate for reducing
    the size of the cover.

    super_gasp is a variation on this strategy which extracts a minimal
    subset from the set of all prime implicants which cover all
    maximally reduced cubes.
*/
final class Jgasp {


	/**
	 *  reduce_gasp -- compute the maximal reduction of each cube of F
	 *
	 *  If a cube does not reduce, it remains prime; otherwise, it is marked
	 *  as nonprime.   If the cube is redundant (should NEVER happen here) we
	 *  just crap out ...
	 *
	 *  A cover with all of the cubes of F is returned.  Those that did
	 *  reduce are marked "NONPRIME"; those that reduced are marked "PRIME".
	 *  The cubes are in the same order as in F.
	 */
	static Jset_family reduce_gasp(final Jset_family F, final Jset_family D) {
		Jset_family G = Jespresso.new_cover( F.count );
		Jset[] FD = Jcofactor.cube2list( F, D );

		/* Reduce cubes of F without replacement */
		final int[] data = F.data;// java
		for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			int[] cunder = Jreduce.reduce_cube( FD, data, p );
			if( Jset.setp_empty( cunder ) ) {
				Jcvrmisc.fatal("empty reduction in reduce_gasp, shouldn't happen");
			} else if( Jset.setp_equal( cunder, data, p ) ) {
				Jespresso.SET( cunder, Jespresso.PRIME );			/* just to make sure */
				G = Jset.sf_addset( G, data, p );		/* it did not reduce ... */
			} else {
				Jespresso.RESET( cunder, Jespresso.PRIME );		/* it reduced ... */
				G = Jset.sf_addset( G, cunder );
			}
			if( (Jglobals.debug & Jespresso.GASP) != 0 ) {
				System.out.printf("REDUCE_GASP: %s reduced to %s\n", Jcvrout.pc1( data, p ), Jcvrout.pc2( cunder ) );
			}
			cunder = null;
		}

		FD = null;
		return G;
	}

	/**
	*  expand_gasp -- expand each nonprime cube of F into a prime implicant
	*
	*  The gasp strategy differs in that only those cubes which expand to
	*  cover some other cube are saved; also, all cubes are expanded
	*  regardless of whether they become covered or not.
	*/
	static Jset_family expand_gasp(final Jset_family F, final Jset_family D, final Jset_family R, final Jset_family Foriginal) {
		/* Try to expand each nonprime and noncovered cube */
		final Jset_family G[] = new Jset_family[1];// java holder
		G[ 0 ] = Jespresso.new_cover( 10 );
		for( int c1index = 0; c1index < F.count; c1index++ ) {
			expand1_gasp( F, D, R, Foriginal, c1index, G );
		}
		G[0] = Jcontain.sf_dupl( G[0] );
		G[0] = Jexpand.expand( G[0], R, /*nonsparse*/ false );	/* Make them prime ! */
		return G[0];
	}

	/**
	 * expand1 -- Expand a single cube against the OFF-set, using the gasp strategy
	 *
	 * @param F reduced cubes of ON-set
	 * @param D DC-set
	 * @param R OFF-set
	 * @param Foriginal ON-set before reduction (same order as F)
	 * @param c1index which index of F (or Freduced) to be checked
	 * @param G
	 */
	static void expand1_gasp(final Jset_family F, final Jset_family D, final Jset_family R, final Jset_family Foriginal, final int c1index, final Jset_family[] G) {
		if( (Jglobals.debug & Jespresso.EXPAND1) != 0 ) {
			System.out.printf("\nEXPAND1_GASP:    \t%s\n", Jcvrout.pc1( F.data, F.wsize * c1index ) );
		}

		int[] RAISE = Jespresso.new_cube();
		int[] FREESET = Jespresso.new_cube();
		int[] temp = Jespresso.new_cube();

		/* Initialize the OFF-set */
		R.active_count = R.count;
		int[] data = R.data;// java
		for( int p = 0, last = p + R.count * R.wsize; p < last; p += R.wsize ) {
			Jespresso.SET( data, p, Jespresso.ACTIVE );
		}
		/* Initialize the reduced ON-set, all nonprime cubes become active */
		F.active_count = F.count;
		data = F.data;// java
		for( int c2under = 0, c2index = 0; c2index < F.count; c2under += F.wsize, c2index++ ) {
			if( c1index == c2index || Jespresso.TESTP( data, c2under, Jespresso.PRIME ) ) {
				F.active_count--;
				Jespresso.RESET( data, c2under, Jespresso.ACTIVE );
			} else {
				Jespresso.SET( data, c2under, Jespresso.ACTIVE );
			}
		}

		/* Initialize the raising and unassigned sets */
		Jset.set_copy( RAISE, F.data, F.wsize * c1index );
		Jset.set_diff( FREESET, Jglobals.cube.fullset, RAISE );

		/* Determine parts which must be lowered */
		Jexpand.essen_parts( R, F, RAISE, FREESET );

		/* Determine parts which can always be raised */
		Jexpand.essen_raising( R, RAISE, FREESET );

		/* See which, if any, of the reduced cubes we can cover */
		data = F.data;// java
		for( int c2under = 0, c2index = 0; c2index < F.count; c2under += F.wsize, c2index++ ) {
			if (Jespresso.TESTP( data, c2under, Jespresso.ACTIVE ) ) {
				/* See if this cube can be covered by an expansion */
				if( Jset.setp_implies( data, c2under, RAISE ) ||
						Jexpand.feasibly_covered( R, data, c2under, RAISE, temp, 0 ) ) {

					/* See if c1under can expanded to cover c2 reduced against
					* (F - c1) u c1under; if so, c2 can definitely be removed !
					*/

					/* Copy F and replace c1 with c1under */
					final Jset_family F1 = Jset.sf_save( Foriginal );
					Jset.set_copy( F1.data, F1.wsize * c1index, F.data, F.wsize * c1index );

					/* Reduce c2 against ((F - c1) u c1under) */
					Jset[] FD = Jcofactor.cube2list( F1, D );
					int[] c2essential = Jreduce.reduce_cube( FD, F1.data, F1.wsize * c2index );
					FD = null;
					Jset.sf_free( F1 );

					/* See if c2essential is covered by an expansion of c1under */
					if( Jexpand.feasibly_covered( R, c2essential, 0, RAISE, temp, 0 ) ) {
						Jset.set_or( temp, RAISE, c2essential );
						Jespresso.RESET( temp, Jespresso.PRIME );		/* cube not prime */
						G[0] = Jset.sf_addset( G[0], temp);
					}
					c2essential = null;
				}
			}
		}

		RAISE = null;
		FREESET = null;
		temp = null;
	}

	/** irred_gasp -- Add new primes to F and find an irredundant subset
	 *
	 * @param F
	 * @param D
	 * @param G G is disposed of
	 */
	static Jset_family irred_gasp(Jset_family F, final Jset_family D, final Jset_family G) {
		if( G.count != 0 ) {
			F = Jirred.irredundant( Jset.sf_append( F, G ), D );
		} else {
			Jespresso.free_cover( G );
		}
		return F;
	}

	/** last_gasp */
	static Jset_family last_gasp(Jset_family F, final Jset_family D, final Jset_family R, final Jcost_t cost) {
		Jset_family G, G1;

		{
			final long t = System.currentTimeMillis();
			G = reduce_gasp( F, D );
			Jcvrmisc.totals( t, Jespresso.GREDUCE_TIME, G, cost );
		}
		{
			final long t = System.currentTimeMillis();
			G1 = expand_gasp( G, D, R, F );
			Jcvrmisc.totals( t, Jespresso.GEXPAND_TIME, G1, cost );
		}
		Jespresso.free_cover( G );
		{
			final long t = System.currentTimeMillis();
			F = irred_gasp( F, D, G1 );
			Jcvrmisc.totals( t, Jespresso.GIRRED_TIME, F, cost );
		}
		return F;
	}

	/** super_gasp */
	static Jset_family super_gasp(Jset_family F, final Jset_family D, final Jset_family R, final Jcost_t cost) {
		Jset_family G, G1;

		{
			final long t = System.currentTimeMillis();
			G = reduce_gasp( F, D );
			Jcvrmisc.totals( t, Jespresso.GREDUCE_TIME, G, cost );
		}
		{
			final long t = System.currentTimeMillis();
			G1 = Jexpand.all_primes( G, R );
			Jcvrmisc.totals( t, Jespresso.GEXPAND_TIME, G1, cost );
		}
		Jespresso.free_cover( G );
		{
			final long t = System.currentTimeMillis();
			G = Jcontain.sf_dupl( Jset.sf_append( F, G1 ) );
			if( Jglobals.trace ) {
				Jcvrmisc.print_trace( G, "NEWPRIMES", System.currentTimeMillis() - t );
			}
		}
		{
			final long t = System.currentTimeMillis();
			F = Jirred.irredundant( G, D );
			Jcvrmisc.totals( t, Jespresso.IRRED_TIME, F, cost );
		}
		return F;
	}
}