package espresso;

final class Jmincov {
	private static final String CLASS_NAME = "Jmincov";

	// mincov_int.h
	static final int WEIGHT(final int[] weight, final int col) {
		return (weight == null ? 1 : weight[ col] );
	}
	// end mincov_int.h
	/* mincov.c */

	private static final boolean USE_GIMPEL = true;
	private static final boolean USE_INDEP_SET = true;

	@SuppressWarnings("boxing")
	private static void fail(final String file, final int line, final String why ) {
		System.err.printf("Fatal error: file %s, line %d\n%s\n", file, line, why );
		System.out.flush();
		Jmain.closeStreams();
		System.exit( 1 );
	}

	/**
	 *
	 * @param A
	 * @param weight
	 * @param heuristic set to 1 for a heuristic covering
	 * @param debug_level how deep in the recursion to provide info
	 * @return
	 */
	@SuppressWarnings("boxing")
	static Jsm_row sm_minimum_cover(final Jsm_matrix A, final int[] weight, final boolean heuristic, final int debug_level ) {

		/* Avoid sillyness */
		if( A.nrows <= 0 ) {
			return Jrows.sm_row_alloc();		/* easy to cover */
		}

		/* Initialize debugging structure */
		final Jstats_t stats = new Jstats_t();
		stats.start_time = System.currentTimeMillis();
		stats.debug = debug_level > 0;
		stats.max_print_depth = debug_level;
		stats.max_depth = -1;
		stats.nodes = 0;
		stats.component = stats.comp_count = 0;
		stats.gimpel = stats.gimpel_count = 0;
		stats.no_branching = heuristic;
		stats.lower_bound = -1;

		/* Check the matrix sparsity */
		int nelem = 0;
		for( Jsm_row prow = A.first_row; prow != null; prow = prow.next_row ) {
			nelem += prow.length;
		}
		final double sparsity = (double) nelem / (double) (A.nrows * A.ncols);

		/* Determine an upper bound on the solution */
		int bound = 1;
		for( Jsm_col pcol = A.first_col; pcol != null; pcol = pcol.next_col) {
			bound += WEIGHT( weight, pcol.col_num );
		}

		/* Perform the covering */
		final Jsolution select = Jsolution.solution_alloc();
		final Jsm_matrix dup_A = Jmatrix.sm_dup( A );
		final Jsolution best = sm_mincov( dup_A, select, weight, 0, bound, 0, stats );
		Jmatrix.sm_free( dup_A );
		Jsolution.solution_free( select );

		if( stats.debug ) {
			if( stats.no_branching ) {
				System.out.printf("**** heuristic covering ...\n");
				System.out.printf("lower bound = %d\n", stats.lower_bound);
			}
			System.out.printf("matrix     = %d by %d with %d elements (%4.3f%%)\n",
					A.nrows, A.ncols, nelem, sparsity * 100.0);
			System.out.printf("cover size = %d elements\n", best.row.length);
			System.out.printf("cover cost = %d\n", best.cost);
			System.out.printf("time       = %s\n",
					Jespresso.print_time( System.currentTimeMillis() - stats.start_time ) );
			System.out.printf("components = %d\n", stats.comp_count);
			System.out.printf("gimpel     = %d\n", stats.gimpel_count);
			System.out.printf("nodes      = %d\n", stats.nodes);
			System.out.printf("max_depth  = %d\n", stats.max_depth);
		}

		final Jsm_row sol = Jrows.sm_row_dup( best.row );
		if( ! verify_cover( A, sol ) ) {
			fail( CLASS_NAME, 92, "mincov: internal error -- cover verification failed\n");
		}
		Jsolution.solution_free( best );
		return sol;
	}

	/*
	 *  Find the best cover for 'A' (given that 'select' already selected);
	 *
	 *    - abort search if a solution cannot be found which beats 'bound'
	 *
	 *    - if any solution meets 'lower_bound', then it is the optimum solution
	 *      and can be returned without further work.
	 */

	@SuppressWarnings("boxing")
	static Jsolution sm_mincov(final Jsm_matrix A, final Jsolution select, final int[] weight, final int lb, int bound, final int depth, final Jstats_t stats)
	{
		final Jsm_matrix[] L = new Jsm_matrix[1];// java helper
		final Jsm_matrix[] R = new Jsm_matrix[1];// java helper
		int pick, lb_new;

		/* Start out with some debugging information */
		stats.nodes++;
		if( depth > stats.max_depth ) {
			stats.max_depth = depth;
		}
		final boolean debug = stats.debug && (depth <= stats.max_print_depth);

		/* Apply row dominance, column dominance, and select essentials */
		select_essential( A, select, weight, bound );
		if( select.cost >= bound ) {
			return null;
		}

		/* See if gimpel's reduction technique applies ... */
if( USE_GIMPEL ) {
		if( weight == null ) {	/* hack until we fix it */
			final Jsolution[] best = new Jsolution[ 1 ];// java holder
			if( Jgimpel.gimpel_reduce( A, select, weight, lb, bound, depth, stats, best ) ) {
				return best[ 0 ];
			}
		}
}

if( USE_INDEP_SET ) {
		/* Determine bound from here to final solution using independent-set */
		final Jsolution indep = Jindep.sm_maximal_independent_set( A, weight );

		/* make sure the lower bound is monotonically increasing */
		lb_new = Math.max( select.cost + indep.cost, lb );
		pick = select_column( A, weight, indep );
		Jsolution.solution_free( indep );
} else {
		lb_new = select.cost + (A.nrows > 0 ? 1 : 0);
		pick = select_column( A, weight, null );
}

		if( depth == 0 ) {
			stats.lower_bound = lb_new + stats.gimpel;
		}

		if( debug ) {
			System.out.printf("ABSMIN[%2d]%s", depth, stats.component != 0 ? "*" : " ");
			System.out.printf(" %3dx%3d sel=%3d bnd=%3d lb=%3d %12s ",
					A.nrows, A.ncols, select.cost + stats.gimpel,
					bound + stats.gimpel, lb_new + stats.gimpel,
					Jespresso.print_time( System.currentTimeMillis() - stats.start_time) );
		}

		/* Check for bounding based on no better solution possible */
		Jsolution best;
		if( lb_new >= bound ) {
			if( debug ) {
				System.out.printf("bounded\n");
			}
			best = null;

		/* Check for new best solution */
		} else if( A.nrows == 0 ) {
			best = Jsolution.solution_dup( select );
			if( debug ) {
				System.out.printf("BEST\n");
			}
			if( stats.debug && stats.component == 0 ) {
				System.out.printf("new 'best' solution %d at level %d (time is %s)\n",
						best.cost + stats.gimpel, depth,
						Jespresso.print_time( System.currentTimeMillis() - stats.start_time ) );
			}


			/* Check for a partition of the problem */
		} else if( Jpart.sm_block_partition( A, L, R ) ) {
			/* Make L the smaller problem */
			if( L[0].ncols > R[0].ncols ) {
				final Jsm_matrix A1 = L[0];
				L[0] = R[0];
				R[0] = A1;
			}
			if( debug ) {
				System.out.printf("comp %d %d\n", L[0].nrows, R[0].nrows);
			}
			stats.comp_count++;

			/* Solve problem for L */
			final Jsolution select1 = Jsolution.solution_alloc();
			stats.component++;
			final Jsolution best1 = sm_mincov( L[0], select1, weight, 0, bound - select.cost, depth + 1, stats );
			stats.component--;
			Jsolution.solution_free( select1 );
			Jmatrix.sm_free( L[0] );

			/* Add best solution to the selected set */
			if( best1 == null ) {
				best = null;
			} else {
				for( Jsm_element p = best1.row.first_col; p != null; p = p.next_col ) {
					Jsolution.solution_add( select, weight, p.col_num );
				}
				Jsolution.solution_free( best1 );

				/* recur for the remaining block */
				best = sm_mincov( R[0], select, weight, lb_new, bound, depth+1, stats );
			}
			Jmatrix.sm_free( R[0] );

			/* We've tried as hard as possible, but now we must split and recur */
		} else {
			if( debug ) {
				System.out.printf("pick=%d\n", pick);
			}

			/* Assume we choose this column to be in the covering set */
			final Jsm_matrix A1 = Jmatrix.sm_dup( A );
			final Jsolution select1 = Jsolution.solution_dup( select );
			Jsolution.solution_accept( select1, A1, weight, pick );
			final Jsolution best1 = sm_mincov( A1, select1, weight, lb_new, bound, depth+1, stats );
			Jsolution.solution_free( select1 );
			Jmatrix.sm_free( A1 );

			/* Update the upper bound if we found a better solution */
			if( best1 != null && bound > best1.cost ) {
				bound = best1.cost;
			}

			/* See if this is a heuristic covering (no branching) */
			if( stats.no_branching ) {
				return best1;
			}

			/* Check for reaching lower bound -- if so, don't actually branch */
			if( best1 != null && best1.cost == lb_new ) {
				return best1;
			}

			/* Now assume we cannot have that column */
			final Jsm_matrix A2 = Jmatrix.sm_dup( A );
			final Jsolution select2 = Jsolution.solution_dup( select );
			Jsolution.solution_reject( select2, A2, weight, pick );
			final Jsolution best2 = sm_mincov( A2, select2, weight, lb_new, bound, depth+1, stats );
			Jsolution.solution_free( select2 );
			Jmatrix.sm_free( A2 );

			best = Jsolution.solution_choose_best( best1, best2 );
		}

		return best;
	}

	static int select_column(final Jsm_matrix A, final int[] weight, final Jsolution indep) {

		final Jsm_row indep_cols = Jrows.sm_row_alloc();
		if( indep != null ) {
			/* Find which columns are in the independent sets */
			for( Jsm_element p = indep.row.first_col; p != null; p = p.next_col ) {
				final Jsm_row prow = Jsparse.sm_get_row( A, p.col_num );
				for( Jsm_element p1 = prow.first_col; p1 != null; p1 = p1.next_col ) {
					Jrows.sm_row_insert( indep_cols, p1.col_num );
				}
			}
		} else {
			/* select out of all columns */
			for( Jsm_col pcol = A.first_col; pcol != null; pcol = pcol.next_col) {
				Jrows.sm_row_insert( indep_cols, pcol.col_num );
			}
		}

		/* Find the best column */
		int best_col = -1;
		double best = -1;

		/* Consider only columns which are in some independent row */
		for( Jsm_element p1 = indep_cols.first_col; p1 != null; p1 = p1.next_col ) {
			final Jsm_col pcol = Jsparse.sm_get_col( A, p1.col_num );

			/* Compute the total 'value' of all things covered by the column */
			double w = 0.0;
			for( Jsm_element p = pcol.first_row; p != null; p = p.next_row ) {
				final Jsm_row prow = Jsparse.sm_get_row( A, p.row_num );
				w += 1.0 / ((double) prow.length - 1.0);
			}

			/* divide this by the relative cost of choosing this column */
			w = w / (double) WEIGHT( weight, pcol.col_num );

			/* maximize this ratio */
			if( w  > best ) {
				best_col = pcol.col_num;
				best = w;
			}
		}

		Jrows.sm_row_free( indep_cols );
		return best_col;
	}

	/**
	 *
	 * @param A
	 * @param select
	 * @param weight
	 * @param bound must beat this solution
	 */
	static void select_essential(final Jsm_matrix A, final Jsolution select, final int[] weight, final int bound) {
		int delcols;
		int delrows, essen_count;

		do {
			/*  Check for dominated columns  */
			delcols = Jdominate.sm_col_dominance( A, weight );

			/*  Find the rows with only 1 element (the essentials) */
			final Jsm_row essen = Jrows.sm_row_alloc();
			for( Jsm_row prow = A.first_row; prow != null; prow = prow.next_row ) {
				if( prow.length == 1 ) {
					Jrows.sm_row_insert( essen, prow.first_col.col_num );
				}
			}

			/* Select all of the elements */
			for( Jsm_element p = essen.first_col; p != null; p = p.next_col ) {
				Jsolution.solution_accept( select, A, weight, p.col_num );
				/* Make sure solution still looks good */
				if( select.cost >= bound ) {
					Jrows.sm_row_free( essen );
					return;
				}
			}
			essen_count = essen.length;
			Jrows.sm_row_free( essen );

			/*  Check for dominated rows  */
			delrows = Jdominate.sm_row_dominance( A );

		} while( delcols > 0 || delrows > 0 || essen_count > 0 );
	}

	static boolean verify_cover(final Jsm_matrix A, final Jsm_row cover) {
		for( Jsm_row prow = A.first_row; prow != null; prow = prow.next_row ) {
			if( ! Jrows.sm_row_intersects( prow, cover ) ) {
				return false;
			}
		}
		return true;
	}
}