package espresso;

/**
 *   Phase assignment technique (T. Sasao):
 *
 *      1. create a function with 2*m outputs which implements the
 *      original function and its complement for each output
 *
 *      2. minimize this function
 *
 *      3. choose the minimum number of prime implicants from the
 *      result of step 2 which are needed to realize either a function
 *      or its complement for each output
 *
 *  Step 3 is performed in a rather crude way -- by simply multiplying
 *  out a large expression of the form:
 *
 *      I = (ab + cdef)(acd + bgh) ...
 *
 *  which is a product of m expressions where each expression has two
 *  product terms -- one representing which primes are needed for the
 *  function, and one representing which primes are needed for the
 *  complement.  The largest product term resulting shows which primes
 *  to keep to implement one function or the other for each output.
 *  For problems with many outputs, this may grind to a
 *  halt.
 *
 *  Untried: form complement of I and use unate_complement ...
 *
 *  I have unsuccessfully tried several modifications to the basic
 *  algorithm.  The first is quite simple: use Sasao's technique, but
 *  only commit to a single output at a time (rather than all
 *  outputs).  The goal would be that the later minimizations can "take
 *  into account" the partial assignment at each step.  This is
 *  expensive (m+1 minimizations rather than 2), and the results are
 *  discouraging.
 *
 *  The second modification is rather complicated.  The result from the
 *  minimization in step 2 is guaranteed to be minimal.  Hence, for
 *  each output, the set of primes with a 1 in that output are both
 *  necessary and sufficient to implement the function.  Espresso
 *  achieves the minimality using the routine MAKE_SPARSE.  The
 *  modification is to prevent MAKE_SPARSE from running.  Hence, there
 *  are potentially many subsets of the set of primes with a 1 in a
 *  column which can be used to implement that output.  We use
 *  IRREDUNDANT to enumerate all possible subsets and then proceed as
 *  before.
 */
final class Jopo {
	private static boolean opo_no_make_sparse;
	private static boolean opo_repeated;
	private static boolean opo_exact;

	static void phase_assignment(final JpPLA PLA, final int opo_strategy) {
		opo_no_make_sparse = (opo_strategy % 2) != 0;
		Jglobals.skip_make_sparse = opo_no_make_sparse;
		opo_repeated = ((opo_strategy / 2) % 2) != 0;
		opo_exact = (((opo_strategy / 4) % 2) != 0);

		/* Determine a phase assignment */
		if( PLA.phase != null ) {
			PLA.phase = null;
		}

		if( opo_repeated ) {
			PLA.phase = Jset.set_save( Jglobals.cube.fullset );
			repeated_phase_assignment( PLA );
		} else {
			PLA.phase = find_phase( PLA, 0, null );
		}

		/* Now minimize with this assignment */
		Jglobals.skip_make_sparse = false;
		set_phase( PLA );
		minimize( PLA );
	}

	/**
	*  repeated_phase_assignment -- an alternate strategy which commits
	*  to a single phase assignment a step at a time.  Performs m + 1
	*  minimizations !
	*/
	@SuppressWarnings("boxing")
	static void repeated_phase_assignment(final JpPLA PLA) {
		final Jcube_struct cube = Jglobals.cube;// java

		for( int i = 0; i < cube.part_size[ cube.output ]; i++ ) {

			/* Find best assignment for all undecided outputs */
			int[] phase = find_phase( PLA, i, PLA.phase );

			/* Commit for only a single output ... */
			if( ! Jespresso.is_in_set( phase, cube.first_part[cube.output] + i ) ) {
				Jespresso.set_remove( PLA.phase, cube.first_part[cube.output] + i );
			}

			if( Jglobals.trace || Jglobals.summary ) {
				System.out.printf("\nOPO loop for output #%d\n", i);
				System.out.printf("PLA->phase is %s\n", Jcvrout.pc1( PLA.phase, 0 ) );
				System.out.printf("phase      is %s\n", Jcvrout.pc1( phase, 0 ) );
			}
			phase = null;
		}
	}


	/**
	*  find_phase -- find a phase assignment for the PLA for all outputs starting
	*  with output number first_output.
	*/
	static int[] find_phase(final JpPLA PLA, final int first_output, final int[] phase1) {
		final Jcube_struct cube = Jglobals.cube;// java

		final int[] phase = Jset.set_save( cube.fullset );

		/* setup the double-phase characteristic function, resize the cube */
		final JpPLA PLA1 = Jcvrin.new_PLA();
		PLA1.F = Jset.sf_save( PLA.F );
		PLA1.R = Jset.sf_save( PLA.R );
		PLA1.D = Jset.sf_save( PLA.D );
		if( phase1 != null ) {
			PLA1.phase = Jset.set_save( phase1 );
			set_phase( PLA1 );
		}
		{
			final long t = System.currentTimeMillis();
			output_phase_setup( PLA1, first_output );
			if( Jglobals.summary ) {
				Jcvrmisc.print_trace( PLA1.F, "OPO-SETUP ", System.currentTimeMillis() - t );
			}
		}

		/* minimize the double-phase function */
		minimize( PLA1 );

		/* set the proper phases according to what gives a minimum solution */
		{
			final long t = System.currentTimeMillis();
			PLA1.F = opo( phase, PLA1.F, PLA1.D, PLA1.R, first_output );
			if( Jglobals.summary ) {
				Jcvrmisc.print_trace( PLA1.F, "OPO       ", System.currentTimeMillis() - t );
			}
		}
		Jcvrin.free_PLA( PLA1 );

		/* set the cube structure to reflect the old size */
		Jcubestr.setdown_cube();
		cube.part_size[ cube.output ] -= (cube.part_size[ cube.output ] - first_output) >>> 1;
		Jcubestr.cube_setup();

		return phase;
	}

	/**
	*  opo -- multiply the expression out to determine a minimum subset of
	*  primes.
	*/

	/*ARGSUSED*/
	static Jset_family opo(final int[] phase, final Jset_family T, final Jset_family D, final Jset_family R, final int first_output) {
		final Jcube_struct cube = Jglobals.cube;// java

		/* must select all primes for outputs [0 .. first_output-1] */
		int[] select = Jset.set_full( T.count );
		for( int output = 0; output < first_output; output++ ) {
			final int ind = cube.first_part[cube.output] + output;
			final int[] data = T.data;// java
			for( int p = 0, i = 0; i < T.count; p += T.wsize, i++ ) {
				if( Jespresso.is_in_set( data, p, ind ) ) {
					Jespresso.set_remove( select, i );
				}
			}
		}

		/* Recursively perform the intersections */
		final int offset = (cube.part_size[cube.output] - first_output) / 2;
		final int last_output = first_output + offset - 1;
		final Jset_family temp = opo_recur( T, D, select, offset, first_output, last_output );

		/* largest set is on top -- select primes which are inferred from it */
		final int[] pdest = temp.data;
		Jset_family T1 = Jespresso.new_cover( T.count );
		int[] data = T.data;// java
		for( int p = 0, i = 0; i < T.count; p += T.wsize, i++ ) {
			if( ! Jespresso.is_in_set( pdest, i ) ) {
				T1 = Jset.sf_addset( T1, data, p );
			}
		}

		select = null;
		Jset.sf_free( temp );

		/* finding phases is difficult -- see which functions are not covered */
		final Jset_family T2 = Jcompl.complement( Jcofactor.cube1list( T1 ) );
		int[] not_covered = Jespresso.new_cube();
		int[] tmp = Jespresso.new_cube();

		final int[] data2 = T2.data;// java
		data = T.data;// java
		for( int p = 0, last = p + T.count * T.wsize; p < last; p += T.wsize ) {
			for( int p1 = 0, last1 = p1 + T2.count * T2.wsize; p1 < last1; p1 += T2.wsize ) {
				if( Jsetc.cdist0( data, p, data2, p1 ) ) {
					Jset.set_and( tmp, data, p, data2, p1 );
					Jset.set_or( not_covered, not_covered, tmp );
				}
			}
		}
		Jespresso.free_cover( T );
		Jespresso.free_cover( T2 );
		tmp = null;

		/* Now reflect the phase choice in a single cube */
		for( int output = first_output; output <= last_output; output++ ) {
			final int ind = cube.first_part[ cube.output ] + output;
			if( Jespresso.is_in_set( not_covered, ind ) ) {
				if( Jespresso.is_in_set( not_covered, ind + offset ) ) {
					Jcvrmisc.fatal("error in output phase assignment");
				} else {
					Jespresso.set_remove( phase, ind );
				}
			}
		}
		not_covered = null;
		return T1;
	}

	static int level = 0;
	@SuppressWarnings("boxing")
	static Jset_family opo_recur(final Jset_family T, final Jset_family D, final int[] select, final int offset, final int first, final int last) {
		int middle;
		Jset_family sl, sr, temp;

		level++;
		if( first == last ) {
/* #if 0
			if( opo_no_make_sparse ) {
				temp = form_cover_table(T, D, select, first, first + offset);
			} else {
				temp = opo_leaf(T, select, first, first + offset);
			}
#else */
			temp = opo_leaf( T, select, offset, first, first + offset );
// #endif
		} else {
			middle = (first + last) / 2;
			sl = opo_recur( T, D, select, offset, first, middle );
			sr = opo_recur( T, D, select, offset, middle + 1, last );
			temp = Junate.unate_intersect( sl, sr, level == 1 );
			if( Jglobals.trace ) {
				System.out.printf("# OPO[%d]: %4d = %4d x %4d, time = %s\n", level - 1,
						temp.count, sl.count, sr.count, Jespresso.print_time( System.currentTimeMillis() ) );
				System.out.flush();
			}
			Jespresso.free_cover( sl );
			Jespresso.free_cover( sr );
		}
		level--;
		return temp;
	}


	static Jset_family opo_leaf(final Jset_family T, final int[] select, final int off, int out1, int out2) {
		final Jcube_struct cube = Jglobals.cube;// java

		out1 += cube.first_part[ cube.output ];
		out2 += cube.first_part[ cube.output ];

		/* Allocate space for the result */
		final Jset_family temp = Jset.sf_new( 2, T.count );

		/* Find which primes are needed for the ON-set of this fct */
		int[] p_base = temp.data;// java
		int pdest = temp.wsize * temp.count++;
		Jset.set_copy( p_base, pdest, select, off );

		final int[] data = T.data;// java
		for( int p = 0, i = 0; i < T.count; p += T.wsize, i++ ) {
			if( Jespresso.is_in_set( data, p, out1 ) ) {
				Jespresso.set_remove( p_base, pdest, i );
			}
		}

		/* Find which primes are needed for the OFF-set of this fct */
		p_base = temp.data;// java
		pdest = temp.wsize * temp.count++;
		Jset.set_copy( p_base, pdest, select, off );

		for( int p = 0, i = 0; i < T.count; p += T.wsize, i++ ) {
			if( Jespresso.is_in_set( data, p, out2 ) ) {
				Jespresso.set_remove( p_base, pdest, i );
			}
		}

		return temp;
	}

	/**
	 *
	 * @param F
	 * @param D
	 * @param select
	 * @param f indices of f and fbar in the output part
	 * @param fbar indices of f and fbar in the output part
	 * @return
	 */
/* #if 0
	static pset_family form_cover_table(final Jset_family F, final Jset_family D, final int[] select, final int f, final int fbar) {
		final int i;
		final pcube p;
		pset_family f_table, fbar_table;

		// setup required for fcube_is_covered
		Rp_size = F.count;
		Rp_start = set_new(Rp_size);
		foreachi_set(F, i, p) {
			PUTSIZE(p, i);
		}
		foreachi_set(D, i, p) {
			RESET(p, REDUND);
		}

		f_table = find_covers(F, D, select, f);
		fbar_table = find_covers(F, D, select, fbar);
		f_table = sf_append(f_table, fbar_table);

		set_free(Rp_start);
		return f_table;
	}

	static Jset_family find_covers(final Jset_family F, final Jset_family D, final int[] select, final int n) {
		register pset p, last, new;
		final pcover F1;
		pcube *Flist;
		final pset_family f_table, table;
		final int i;

		n += cube.first_part[cube.output];

		// save cubes in this output, and remove the output variable
		F1 = new_cover(F.count);
		foreach_set(F, last, p)
			if( is_in_set(p, n) ) {
				new = GETSET(F1, F1.count++);
				set_or(new, p, cube.var_mask[cube.output]);
				PUTSIZE(new, SIZE(p));
				SET(new, REDUND);
			}

		// Find ways (sop form) to fail to cover output indexed by n
		Flist = cube2list(F1, D);
		table = sf_new(10, Rp_size);
		foreach_set(F1, last, p) {
			set_fill(Rp_start, Rp_size);
			set_remove(Rp_start, SIZE(p));
			table = sf_append(table, fcube_is_covered(Flist, p));
			RESET(p, REDUND);
		}
		set_fill(Rp_start, Rp_size);
		foreach_set(table, last, p) {
			set_diff(p, Rp_start, p);
		}

		// complement this to get possible ways to cover the function
		for( i = 0; i < Rp_size; i++ ) {
			if( ! is_in_set(select, i) ) {
				p = set_new(Rp_size);
				set_insert(p, i);
				table = sf_addset(table, p);
				set_free(p);
			}
		}
		f_table = unate_compl(table);

		// what a pain, but we need bitwise complement of this
		set_fill(Rp_start, Rp_size);
		foreach_set(f_table, last, p) {
			set_diff(p, Rp_start, p);
		}

		free_cubelist(Flist);
		sf_free(F1);
		return f_table;
	}
#endif */

	/**
	*  Take a PLA (ON-set, OFF-set and DC-set) and create the
	*  "double-phase characteristic function" which is merely a new
	*  function which has twice as many outputs and realizes both the
	*  function and the complement.
	*
	*  The cube structure is assumed to represent the PLA upon entering.
	*  It will be modified to represent the double-phase function upon
	*  exit.
	*
	*  Only the outputs numbered starting with "first_output" are
	*  duplicated in the output part
	*/
	static void output_phase_setup(final JpPLA PLA, final int first_output) {
		final Jcube_struct cube = Jglobals.cube;// java

		if( cube.output == -1 ) {
			Jcvrmisc.fatal("output_phase_setup: must have an output");
		}

		final Jset_family F = PLA.F;
		final Jset_family D = PLA.D;
		final Jset_family R = PLA.R;
		final int first_part = cube.first_part[cube.output] + first_output;
		final int last_part = cube.last_part[cube.output];
		final int offset = cube.part_size[cube.output] - first_output;

		/* Change the output size, setup the cube structure */
		Jcubestr.setdown_cube();
		cube.part_size[ cube.output ] += offset;
		Jcubestr.cube_setup();

		/* Create a mask to select that part of the cube which isn't changing */
		int[] mask = Jset.set_save( cube.fullset );
		for( int i = first_part; i < cube.size; i++ ) {
			Jespresso.set_remove( mask, i );
		}
		int[] mask1 = Jset.set_save( mask );
		for( int i = cube.first_part[cube.output]; i < first_part; i++ ) {
			Jespresso.set_remove( mask1, i );
		}

		PLA.F = Jespresso.new_cover( F.count + R.count );
		PLA.R = Jespresso.new_cover( F.count + R.count );
		PLA.D = Jespresso.new_cover( D.count );

		int data[] = F.data;// java
		for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			final int[] pf_base = PLA.F.data;
			final int pf = PLA.F.wsize * (PLA.F).count++;
			final int[] pr_base = PLA.R.data;
			final int pr = PLA.R.wsize * (PLA.R).count++;
			Jset.set_and( pf_base, pf, mask, data, p );
			Jset.set_and( pr_base, pr, mask1, data, p );
			for( int i = first_part; i <= last_part; i++ ) {
				if( Jespresso.is_in_set( data, p, i ) ) {
					Jespresso.set_insert( pf_base, pf, i );
				}
			}
			boolean save = false;
			for( int i = first_part; i <= last_part; i++ ) {
				if( Jespresso.is_in_set( data, p, i ) ) {
					save = true;
					Jespresso.set_insert( pr_base, pr, i + offset );
				}
			}
			if( ! save ) {
				PLA.R.count--;
			}
		}

		data = R.data;// java
		for( int p = 0, last = p + R.count * R.wsize; p < last; p += R.wsize ) {
			final int[] pf_base = PLA.F.data;
			final int pf = PLA.F.wsize * PLA.F.count++;
			final int[] pr_base = PLA.R.data;
			final int pr = PLA.R.wsize * PLA.R.count++;
			Jset.set_and( pf_base, pf, mask1, data, p );
			Jset.set_and( pr_base, pr, mask, data, p );
			boolean save = false;
			for( int i = first_part; i <= last_part; i++ ) {
				if( Jespresso.is_in_set( data, p, i ) ) {
					save = true;
					Jespresso.set_insert( pf_base, pf, i + offset );
				}
			}
			if( ! save ) {
				PLA.F.count--;
			}
			for( int i = first_part; i <= last_part; i++ ) {
				if( Jespresso.is_in_set( data, p, i ) ) {
					Jespresso.set_insert( pr_base, pr, i );
				}
			}
		}

		data = D.data;// java
		for( int p = 0, last = p + D.count * D.wsize; p < last; p += D.wsize ) {
			final int[] pf_base = PLA.D.data;
			final int pf = PLA.D.wsize * (PLA.D).count++;
			Jset.set_and( pf_base, pf, mask, data, p );
			for( int i = first_part; i <= last_part; i++) {
				if( Jespresso.is_in_set( data, p, i ) ) {
					Jespresso.set_insert( pf_base, pf, i );
					Jespresso.set_insert( pf_base, pf, i + offset );
				}
			}
		}

		Jespresso.free_cover( F );
		Jespresso.free_cover( D );
		Jespresso.free_cover( R );
		mask = null;
		mask1 = null;
	}

	/**
	*  set_phase -- given a "cube" which describes which phases of the output
	*  are to be implemented, compute the appropriate on-set and off-set
	*/
	static JpPLA set_phase(final JpPLA PLA) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] temp = cube.temp[0], phase = PLA.phase, phase1 = cube.temp[1];

		final int[] outmask = cube.var_mask[ cube.num_vars - 1 ];
		Jset.set_diff( phase1, outmask, phase );
		Jset.set_diff( temp, cube.fullset, outmask );
		Jset.set_or( phase1, temp, phase1 );
		final Jset_family F1 = Jespresso.new_cover( (PLA.F).count + (PLA.R).count );
		final Jset_family R1 = Jespresso.new_cover( (PLA.F).count + (PLA.R).count );

		int[] data = PLA.F.data;// java
		for( int p = 0, last = p + PLA.F.count * PLA.F.wsize; p < last; p += PLA.F.wsize ) {
			Jset.set_and( temp, data, p, phase );
			if( ! Jset.setp_disjoint( temp, outmask ) ) {
				Jset.set_copy( F1.data, F1.wsize * F1.count++, temp );
			}
			Jset.set_and( temp, data, p, phase1 );
			if( ! Jset.setp_disjoint( temp, outmask ) ) {
				Jset.set_copy( R1.data, R1.wsize * R1.count++, temp );
			}
		}

		data = PLA.R.data;// java
		for( int p = 0, last = p + PLA.R.count * PLA.R.wsize; p < last; p += PLA.R.wsize ) {
			Jset.set_and( temp, data, p, phase );
			if( ! Jset.setp_disjoint( temp, outmask ) ) {
				Jset.set_copy( R1.data, R1.wsize * R1.count++, temp );
			}
			Jset.set_and( temp, data, p, phase1 );
			if( ! Jset.setp_disjoint( temp, outmask ) ) {
				Jset.set_copy( F1.data, F1.wsize * F1.count++, temp );
			}
		}
		Jespresso.free_cover( PLA.F );
		Jespresso.free_cover( PLA.R );
		PLA.F = F1;
		PLA.R = R1;
		return PLA;
	}

	static void opoall(final JpPLA PLA, final int first_output, final int last_output, final int opo_strategy) {
		final Jcube_struct cube = Jglobals.cube;// java

		opo_exact = (opo_strategy != 0);

		PLA.phase = null;

		final int[] bestphase = Jset.set_save( cube.fullset );
		Jset_family best_F = Jset.sf_save( PLA.F );
		Jset_family best_D = Jset.sf_save( PLA.D );
		Jset_family best_R = Jset.sf_save( PLA.R );

		for( int i = 0, n = 1 << (last_output - first_output + 1); i < n; i++ ) {

			/* save the initial PLA covers */
			final Jset_family F = Jset.sf_save( PLA.F );
			final Jset_family D = Jset.sf_save( PLA.D );
			final Jset_family R = Jset.sf_save( PLA.R );

			/* compute the phase cube for this iteration */
			PLA.phase = Jset.set_save( cube.fullset );
			int num = i;
			for( int j = last_output; j >= first_output; j-- ) {
				if( num % 2 == 0 ) {
					final int ind = cube.first_part[cube.output] + j;
					Jespresso.set_remove( PLA.phase, ind );
				}
				num /= 2;
			}

			/* set the phase and minimize */
			set_phase( PLA );
			System.out.printf("# phase is %s\n", Jcvrout.pc1( PLA.phase, 0 ) );
			Jglobals.summary = true;
			minimize( PLA );

			/* see if this is the best so far */
			if( PLA.F.count < best_F.count ) {
				/* save new best solution */
				Jset.set_copy( bestphase, PLA.phase );
				Jset.sf_free( best_F );
				Jset.sf_free( best_D );
				Jset.sf_free( best_R );
				best_F = PLA.F;
				best_D = PLA.D;
				best_R = PLA.R;
			} else {
				/* throw away the solution */
				Jespresso.free_cover( PLA.F );
				Jespresso.free_cover( PLA.D );
				Jespresso.free_cover( PLA.R );
			}
			PLA.phase = null;

			/* restore the initial PLA covers */
			PLA.F = F;
			PLA.D = D;
			PLA.R = R;
		}

		/* one more minimization to restore the best answer */
		PLA.phase = bestphase;
		PLA.F = best_F;
		PLA.D = best_D;
		PLA.R = best_R;
	}

	private static void minimize(final JpPLA PLA) {
		if( opo_exact ) {
			final long t = System.currentTimeMillis();
			PLA.F = Jexact.minimize_exact( PLA.F, PLA.D, PLA.R, true );
			if( Jglobals.summary ) {
				Jcvrmisc.print_trace( PLA.F, "EXACT", System.currentTimeMillis() - t );
			}
		} else {
			final long t = System.currentTimeMillis();
			PLA.F = Jespresso.espresso( PLA.F, PLA.D, PLA.R );
			if( Jglobals.summary ) {
				Jcvrmisc.print_trace( PLA.F, "ESPRESSO  ", System.currentTimeMillis() - t );
			}
		}
	}
}