package espresso;

import java.util.Comparator;

/**
    module: reduce.c
    purpose: Perform the Espresso-II reduction step

    Reduction is a technique used to explore larger regions of the
    optimization space.  We replace each cube of F with a smaller
    cube while still maintaining a cover of the same logic function.
*/

final class Jreduce {

	private static boolean toggle = true;

	/**
	    reduce -- replace each cube in F with its reduction

	    The reduction of a cube is the smallest cube contained in the cube
	    which can replace the cube in the original cover without changing
	    the cover.  This is equivalent to the super cube of all of the
	    essential points in the cube.  This can be computed directly.

	    The problem is that the order in which the cubes are reduced can
	    greatly affect the final result.  We alternate between two ordering
	    strategies:

		(1) Order the cubes in ascending order of distance from the
		largest cube breaking ties by ordering cubes of equal distance
		in descending order of size (sort_reduce)

		(2) Order the cubes in descending order of the inner-product of
		the cube and the column sums (mini_sort)

	    The real workhorse of this section is the routine SCCC which is
	    used to find the Smallest Cube Containing the Complement of a cover.
	    Reduction as proposed by Espresso-II takes a cube and computes its
	    maximal reduction as the intersection between the cube and the
	    smallest cube containing the complement of (F u D - {c}) cofactored
	    against c.

	    As usual, the unate-recursive paradigm is used to compute SCCC.
	    The SCCC of a unate cover is trivial to compute, and thus we perform
	    Shannon Cofactor expansion attempting to drive the cover to be unate
	    as fast as possible.
	*/
	static Jset_family reduce(Jset_family F, final Jset_family D) {
		/* Order the cubes */
		if( Jglobals.use_random_order ) {
			F = Jcvrm.random_order( F );
		} else {
			final Comparator<Jset> descend = new Jdescend_order();
			F = toggle ? Jcvrm.sort_reduce( F ) : Jcvrm.mini_sort( F, descend );
			toggle = ! toggle;
		}

		/* Try to reduce each cube */
		Jset[] FD = Jcofactor.cube2list( F, D );

		final int[] data = F.data;// java
		for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			int[] cunder = reduce_cube( FD, data, p );		/* reduce the cube */
			if( Jset.setp_equal( cunder, data, p ) ) {            /* see if it actually did */
				Jespresso.SET( data, p, Jespresso.ACTIVE );	/* cube remains active */
				Jespresso.SET( data, p, Jespresso.PRIME );	/* cube remains prime ? */
			} else {
				if( (Jglobals.debug & Jespresso.REDUCE) != 0 ) {
					System.out.printf("REDUCE: %s to %s %s\n",
							Jcvrout.pc1( data, p ), Jcvrout.pc2( cunder ), Jespresso.print_time( System.currentTimeMillis() ) );
				}
				Jset.set_copy( data, p, cunder );                /* save reduced version */
				Jespresso.RESET( data, p, Jespresso.PRIME );                    /* cube is no longer prime */
				if( Jset.setp_empty( cunder ) ) {
					Jespresso.RESET( data, p, Jespresso.ACTIVE );
				} else {
					Jespresso.SET( data, p, Jespresso.ACTIVE );
				}                 /* cube is active */
			}
			cunder = null;
		}
		FD = null;

		/* Delete any cubes of F which reduced to the empty cube */
		return Jset.sf_inactive( F );
	}

	/** reduce_cube -- find the maximal reduction of a cube */
	static int[] reduce_cube(final Jset[] FD, final int[] p, final int off) {
		final int[] cunder = sccc( Jcofactor.cofactor( FD, p, off ) );
		Jset.set_and( cunder, cunder, p, off );
		return cunder;
	}

	private static int sccc_level = 0;

	/** sccc -- find Smallest Cube Containing the Complement of a cover
	 * @param T T will be disposed of
	 */
	@SuppressWarnings("boxing")
	static int[] sccc(final Jset[] T) {
		if( (Jglobals.debug & Jespresso.REDUCE1) != 0 ) {
			Jcvrout.debug_print( T, "SCCC", sccc_level++ );
		}

		final int[][] r = new int[1][];// java holder
		if( sccc_special_cases( T, r ) == Jespresso.MAYBE ) {
			final int[] cl = Jespresso.new_cube();
			final int[] cr = Jespresso.new_cube();
			final int best = Jcofactor.binate_split_select( T, cl, cr, Jespresso.REDUCE1 );
			r[0] = sccc_merge( sccc( Jcofactor.scofactor( T, cl, best ) ),
								sccc( Jcofactor.scofactor( T, cr, best ) ), cl, cr );
			T[0] = null;
		}

		if( (Jglobals.debug & Jespresso.REDUCE1) != 0 ) {
			System.out.printf("SCCC[%d]: result is %s\n", --sccc_level, Jcvrout.pc1( r[0], 0 ) );
		}
		return r[0];
	}

	/**
	 *
	 * @param left will be disposed of ...
	 * @param right will be disposed of ...
	 * @param cl will be disposed of ...
	 * @param cr will be disposed of ...
	 * @return
	 */
	static int[] sccc_merge(final int[] left, final int[] right, final int[] cl, final int[] cr) {
		Jset.set_and( left, left, cl );
		Jset.set_and( right, right, cr );
		Jset.set_or( left, left, right );
		// right = null;
		// cl = null;
		// cr = null;
		return left;
	}

	/**
	    sccc_cube -- find the smallest cube containing the complement of a cube

	    By DeMorgan's law and the fact that the smallest cube containing a
	    cover is the "or" of the positional cubes, it is simple to see that
	    the SCCC is the universe if the cube has more than two active
	    variables.  If there is only a single active variable, then the
	    SCCC is merely the bitwise complement of the cube in that
	    variable.  A last special case is no active variables, in which
	    case the SCCC is empty.

	    This is "anded" with the incoming cube result.
	*/
	static int[] sccc_cube(final int[] result, final int[] p) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] temp = cube.temp[0];
		int var;

		if( (var = Jsetc.cactive( p, 0 )) >= 0 ) {
			final int[] mask = cube.var_mask[var];
			Jset.set_xor( temp, p, mask );
			Jset.set_and( result, result, temp );
		}
		return result;
	}

	/**
	*   sccc_special_cases -- check the special cases for sccc
	* @param T will be disposed if answer is determined
	* @param result returned only if answer determined
	*/
	static int sccc_special_cases(final Jset[] T, final int[][] result) {
		final Jcube_struct cube = Jglobals.cube;// java
		final Jcdata_struct cdata = Jglobals.cdata;// java
		Jset p;
		final int[] temp = cube.temp[1];
		final Jset cof = T[0];

		/* empty cover => complement is universe => SCCC is universe */
		if( T[2] == null ) {
			result[0] = Jset.set_save( cube.fullset );
			T[0] = null;
			return Jespresso.TRUE;
		}

		/* row of 1's => complement is empty => SCCC is empty */
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			if( Jsetc.full_row( p, cof ) ) {
				result[0] = Jespresso.new_cube();
				T[0] = null;
				return Jespresso.TRUE;
			}
		}

		/* Collect column counts, determine unate variables, etc. */
		Jcofactor.massive_count( T );

		/* If cover is unate (or single cube), apply simple rules to find SCCCU */
		if( cdata.vars_unate == cdata.vars_active || T[3] == null ) {
			result[0] = Jset.set_save( cube.fullset );
			for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
				Jset.set_or( temp, p.base, p.offset, cof.base, cof.offset );
				sccc_cube( result[0], temp );
			}
			T[0] = null;
			return Jespresso.TRUE;
		}

		/* Check for column of 0's (which can be easily factored( */
		int[] ceil = Jset.set_save( cof );
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			Jset.set_or( ceil, ceil, p.base, p.offset );
		}
		if( ! Jset.setp_equal( ceil, cube.fullset ) ) {
			result[0] = sccc_cube( Jset.set_save( cube.fullset ), ceil );
			if( Jset.setp_equal( result[0], cube.fullset ) ) {
				ceil = null;
			} else {
				result[0] = sccc_merge( sccc( Jcofactor.cofactor( T, ceil, 0 ) ),
						Jset.set_save( cube.fullset ), ceil, result[0] );
			}
			T[0] = null;
			return Jespresso.TRUE;
		}
		ceil = null;

		/* Single active column at this point => tautology => SCCC is empty */
		if( cdata.vars_active == 1 ) {
			result[0] = Jespresso.new_cube();
			T[0] = null;
			return Jespresso.TRUE;
		}

		/* Check for components */
		final Jset[][] A = new Jset[1][];// java holder
		final Jset[][] B = new Jset[1][];// java holder
		if( cdata.var_zeros[cdata.best] < Jespresso.CUBELISTSIZE( T ) / 2 ) {
			if( Jcvrm.cubelist_partition( T, A, B, Jglobals.debug & Jespresso.REDUCE1 ) == 0 ) {
				return Jespresso.MAYBE;
			}// else {
				T[0] = null;
				result[0] = sccc( A[0] );
				ceil = sccc( B[0] );
				Jset.set_and( result[0], result[0], ceil );
				ceil = null;
				return Jespresso.TRUE;
			//}
		}

		/* Not much we can do about it */
		return Jespresso.MAYBE;
	}
}