package espresso;

/**
    setc.c -- massive bit-hacking for performing special "cube"-type
    operations on a set

    The basic trick used for binary valued variables is the following:

    If a[w] and b[w] contain a full word of binary variables, then:

     1) to get the full word of their intersection, we use

	    x = a[w] & b[w];


     2) to see if the intersection is null in any variables, we examine

	    x = ~(x | x >> 1) & DISJOINT;

	this will have a single 1 in each binary variable for which
	the intersection is null.  In particular, if this is zero,
	then there are no disjoint variables; or, if this is nonzero,
	then there is at least one disjoint variable.  A "count_ones"
	over x will tell in how many variables they have an null
	intersection.


     3) to get a mask which selects the disjoint variables, we use

	    (x | x << 1)

	this provides a selector which can be used to see where
	they have an null intersection


    cdist       return distance between two cubes
    cdist0      return true if two cubes are distance 0 apart
    cdist01     return distance, or 2 if distance exceeds 1
    consensus   compute consensus of two cubes distance 1 apart
    force_lower expand hack (for now), related to consensus
*/

final class Jsetc {

	/** see if the cube has a full row of 1's (with respect to cof) */
	static boolean full_row(final Jset p, final Jset cof) {
		final Jcube_struct cube = Jglobals.cube;// java
		int i = Jespresso.LOOP( p );
		do {
			if( (p.base[ p.offset + i] | cof.base[ cof.offset + i ]) != cube.fullset[i] ) {
				return false;
			}
		} while( --i > 0 );
		return true;
	}

	/**
	 * java: the second variant for other input data
	 * @param p
	 * @param off_p
	 * @param cof
	 * @param off_cof
	 * @return
	 */
	static boolean full_row(final int[] p, final int off_p, final int[] cof, final int off_cof) {
		final Jcube_struct cube = Jglobals.cube;// java
		int i = Jespresso.LOOP( p, off_p );
		do {
			if( (p[ off_p + i] | cof[ off_cof + i ]) != cube.fullset[i] ) {
				return false;
			}
		} while( --i > 0 );
		return true;
	}

	/**
	cdist0 -- return TRUE if a and b are distance 0 apart
	*/
	static boolean cdist0(final int[] a, final int off_a, final int[] b, final int off_b) {
		final Jcube_struct cube = Jglobals.cube;// java
		{  /* Check binary variables */
			int last;
			if( (last = cube.inword) != -1 ) {

				/* Check the partial word of binary variables */
				int x = a[ off_a + last ] & b[ off_b + last ];
				if( (~(x | x >>> 1) & cube.inmask) != 0 ) {
					return false;
				}               /* disjoint in some variable */

				/* Check the full words of binary variables */
				for( int w = 1; w < last; w++ ) {
					x = a[ off_a + w ] & b[ off_b + w ];
					if( (~(x | x >>> 1) & Jespresso.DISJOINT) != 0 ) {
						return false;
					}           /* disjoint in some variable */
				}
			}
		}

		{  /* Check the multiple-valued variables */
			int last;
nextvar_label:
			for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
				final int[] mask = cube.var_mask[ var ]; last = cube.last_word[ var ];
				for( int w = cube.first_word[ var ]; w <= last; w++ ) {
					if( (a[ off_a + w ] & b[ off_b + w ] & mask[ w ]) != 0 ) {
						continue nextvar_label;// goto nextvar;
					}
				}
				return false;           /* disjoint in this variable */
//	nextvar: ;
			}
		}
		return true;
	}

	/**
	cdist01 -- return the "distance" between two cubes (defined as the
	number of null variables in their intersection).  If the distance
	exceeds 1, the value 2 is returned.
	*/
	static int cdist01(final int[] a, final int off_a, final int[] b, final int off_b) {
		final Jcube_struct cube = Jglobals.cube;// java
		int dist = 0;

		{  /* Check binary variables */
			int last;
			if( (last = cube.inword) != -1 ) {

				/* Check the partial word of binary variables */
				int x = a[ off_a + last ] & b[ off_b + last ];
				if( (x = ~ (x | x >>> 1) & cube.inmask) != 0 ) {
					if( (dist = Jespresso.count_ones( x )) > 1 ) {
						return 2;
					}
				}

				/* Check the full words of binary variables */
				for( int w = 1; w < last; w++ ) {
					x = a[ off_a + w ] & b[ off_b + w ];
					if( (x = ~ (x | x >>> 1) & Jespresso.DISJOINT) != 0 ) {
						if( dist == 1 || (dist += Jespresso.count_ones( x )) > 1 ) {
							return 2;
						}
					}
				}
			}
		}

		{  /* Check the multiple-valued variables */
nextvar_label:
			for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
				final int[] mask = cube.var_mask[ var ];
				final int last = cube.last_word[ var ];
				for( int w = cube.first_word[var]; w <= last; w++ ) {
					if( (a[ off_a + w ] & b[ off_b + w ] & mask[ w ]) != 0 ) {
						// goto nextvar;
						continue nextvar_label;
					}
				}
				if( ++dist > 1 ) {
					return 2;
				}
	// nextvar: ;
			}
		}
		return dist;
	}

	/**
	cdist -- return the "distance" between two cubes (defined as the
	number of null variables in their intersection).
	*/
	static int cdist(final int[] a, final int off_a, final int[] b, final int off_b) {
		final Jcube_struct cube = Jglobals.cube;// java
		int dist = 0;

		{  /* Check binary variables */
			int w, last;
			if( (last = cube.inword) != -1 ) {

				/* Check the partial word of binary variables */
				int x = a[ off_a + last ] & b[ off_b + last ];
				if( (x = ~ (x | x >>> 1) & cube.inmask) != 0 ) {
					dist = Jespresso.count_ones( x );
				}

				/* Check the full words of binary variables */
				for( w = 1; w < last; w++ ) {
					x = a[ off_a + w ] & b[ off_b + w ];
					if( (x = ~ (x | x >>> 1) & Jespresso.DISJOINT) != 0 ) {
						dist += Jespresso.count_ones( x );
					}
				}
			}
		}

		{  /* Check the multiple-valued variables */
nextvar_label:
			for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
				final int[] mask = cube.var_mask[ var ];
				final int last = cube.last_word[ var ];
				for( int w = cube.first_word[ var ]; w <= last; w++ ) {
					if( (a[ off_a + w ] & b[ off_b + w ] & mask[ w ]) != 0 ) {
						continue nextvar_label;
					}
				}
				dist++;
//	nextvar: ;
			}
		}
		return dist;
	}

	/** force_lower -- Determine which variables of a do not intersect b. */
	static void force_lower(final int[] xlower, final int off_x, final int[] a, final int off_a, final int[] b, final int off_b) {
		final Jcube_struct cube = Jglobals.cube;// java
		{  /* Check binary variables (if any) */
			int last;
			if( (last = cube.inword) != -1 ) {

				/* Check the partial word of binary variables */
				int x = a[ off_a + last ] & b[ off_b + last ];
				if( (x = ~(x | x >>> 1) & cube.inmask) != 0 ) {
					xlower[ off_x + last ] |= (x | (x << 1)) & a[ off_a + last ];
				}

				/* Check the full words of binary variables */
				for( int w = 1; w < last; w++ ) {
					x = a[ off_a + w ] & b[ off_b + w ];
					if( (x = ~(x | x >>> 1) & Jespresso.DISJOINT) != 0 ) {
						xlower[ off_x + w ] |= (x | (x << 1)) & a[ off_a + w ];
					}
				}
			}
		}

		{  /* Check the multiple-valued variables */
nextvar_label:
			for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
				final int[] mask = cube.var_mask[var];
				final int last = cube.last_word[var];
				for( int w = cube.first_word[var]; w <= last; w++ ) {
					if( (a[ off_a + w ] & b[ off_b + w ] & mask[ w ]) != 0 ) {
						// goto nextvar;
						continue nextvar_label;
					}
				}
				for( int w = cube.first_word[var]; w <= last; w++ ) {
					xlower[ off_x + w ] |= a[ off_a + w ] & mask[ w ];
				}
// nextvar: ;
			}
		}
		// return xlower;
	}

	/**
	consensus -- multiple-valued consensus

	Although this looks very messy, the idea is to compute for r the
	"and" of the cubes a and b for each variable, unless the "and" is
	null in a variable, in which case the "or" of a and b is computed
	for this variable.

	Because we don't check how many variables are null in the
	intersection of a and b, the returned value for r really only
	represents the consensus when a and b are distance 1 apart.
	*/
	static void consensus(final int[] r, final int off_r, final int[] a, final int off_a, final int[] b, final int off_b) {
		final Jcube_struct cube = Jglobals.cube;// java
		Jset.set_clear( r, off_r, cube.size );

		{  /* Check binary variables (if any) */
			int last, x;
			if( (last = cube.inword) != -1 ) {

				/* Check the partial word of binary variables */
				r[ off_r + last ] = x = a[ off_a + last ] & b[ off_b + last ];
				if( (x = ~(x | x >>> 1) & cube.inmask) != 0 ) {
					r[ off_r + last ] |= (x | (x << 1)) & (a[ off_a + last ] | b[ off_b + last ]);
				}

				/* Check the full words of binary variables */
				for( int w = 1; w < last; w++ ) {
					r[ off_r + w ] = x = a[ off_a + w ] & b[ off_b + w ];
					if( (x = ~(x | x >>> 1) & Jespresso.DISJOINT) != 0 ) {
						r[ off_r + w ] |= (x | (x << 1)) & (a[ off_a + w ] | b[ off_b + w ]);
					}
				}
			}
		}
		{  /* Check the multiple-valued variables */
			int x;
			for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
				final int[] mask = cube.var_mask[var];
				final int last = cube.last_word[var];
				boolean empty = true;
				for( int w = cube.first_word[var]; w <= last; w++) {
					if( ((x = a[ off_a + w ]) & b[ off_b + w] & mask[w]) != 0 ) {
						empty = false;
						r[ off_r + w ] |= x;
					}
				}
				if( empty ) {
					for( int w = cube.first_word[var]; w <= last; w++) {
						r[ off_r + w ] |= mask[w] & (a[ off_a + w ] | b[ off_b + w ]);
					}
				}
			}
		}
	}

	/**
	cactive -- return the index of the single active variable in
	the cube, or return -1 if there are none or more than 2.
	*/
	static int cactive(final int[] a, final int off) {
		final Jcube_struct cube = Jglobals.cube;// java
		int active = -1, dist = 0;

		{  /* Check binary variables */
			int last;
			if( (last = cube.inword) != -1 ) {

				/* Check the partial word of binary variables */
				int x = a[ off + last ];
				if( (x = ~ (x & x >>> 1) & cube.inmask) != 0 ) {
					if( (dist = Jespresso.count_ones( x )) > 1 ) {
						return -1;
					}		/* more than 2 active variables */
					active = (last - 1) * (Jespresso.BPI / 2) + (Jset.bit_index( x ) >>> 1);
				}

				/* Check the full words of binary variables */
				for( int w = 1; w < last; w++ ) {
					x = a[ off + w ];
					if( (x = ~ (x & x >>> 1) & Jespresso.DISJOINT) != 0 ) {
						if( (dist += Jespresso.count_ones( x )) > 1 ) {
							return -1;
						}		/* more than 2 active variables */
						active = (w - 1) * (Jespresso.BPI / 2) + (Jset.bit_index( x ) >>> 1);
					}
				}
			}
		}
		{  /* Check the multiple-valued variables */
			for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
				final int[] mask = cube.var_mask[var];
				final int last = cube.last_word[var];
				for( int w = cube.first_word[var]; w <= last; w++ ) {
					if( (mask[w] & ~ a[ off + w ]) != 0 ) {
						if( ++dist > 1 ) {
							return -1;
						}
						active = var;
						break;
					}
				}
			}
		}
		return active;
	}

	/**
	ccommon -- return TRUE if a and b are share "active" variables
	active variables include variables that are empty;
	*/
	static boolean ccommon(final int[] a, final int off_a, final int[] b, final int[] cof, final int off_c) {
		final Jcube_struct cube = Jglobals.cube;// java
		{  /* Check binary variables */
			int last;
			int w;
			if( (last = cube.inword) != -1 ) {

				/* Check the partial word of binary variables */
				int x = a[ off_a + last ] | cof[ off_c + last ];
				int y = b[ last ] | cof[ off_c + last ];
				if( (~(x & x >>> 1) & ~(y & y >>> 1) & cube.inmask) != 0 ) {
					return true;
				}

				/* Check the full words of binary variables */
				for( w = 1; w < last; w++ ) {
					x = a[ off_a + w ] | cof[ off_c + w ];
					y = b[ w ] | cof[ off_c + w ];
					if( (~(x & x >>> 1) & ~(y & y >>> 1) & Jespresso.DISJOINT) != 0 ) {
						return true;
					}
				}
			}
		}
		{  /* Check the multiple-valued variables */
			for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
				final int[] mask = cube.var_mask[var];
				final int last = cube.last_word[var];
				/* Check for some part missing from a */
				for( int w = cube.first_word[var]; w <= last; w++ ) {
					if( (mask[w] & ~a[ off_a + w ] & ~cof[ off_c + w ]) != 0 ) {

						/* If so, check for some part missing from b */
						for( w = cube.first_word[var]; w <= last; w++ ) {
							if( (mask[w] & ~b[ w ] & ~cof[ off_c + w ]) != 0 ) {
								return true;
							}
						}            /* both active */
						break;
					}
				}
			}
		}
		return false;
	}

	/* desc1 -- comparison (without indirection) for descending sort */
	/** also has effect of handling NULL pointers,and a NULL pointer has smallest
	order */
	static int desc1(final Jset a, final Jset b ) {
		if( a == null  ) {
			return (b == null) ? 0 : 1;
		} else if( b == null ) {
			return -1;
		}
		if( Jespresso.SIZE( a ) > Jespresso.SIZE( b ) ) {
			return -1;
		} else if( Jespresso.SIZE( a ) < Jespresso.SIZE( b ) ) {
			return 1;
		} else {
			int i = Jespresso.LOOP( a );
			do {
				final long x1 = ((long)a.base[ a.offset + i ] & 0xffffffffL);
				final long x2 = ((long)b.base[ b.offset + i ] & 0xffffffffL);
				if( x1 > x2 ) {
					return -1;
				} else if( x1 < x2 ) {
					return 1;
				}
			} while( --i > 0 );
		}
		return 0;
	}
}