package espresso;

/**
    module: sparse.c

    make_sparse is a last-step cleanup to reduce the total number
    of literals in the cover.

    This is done by reducing the "sparse" variables (using a modified
    version of irredundant rather than reduce), followed by expanding
    the "dense" variables (using modified version of expand).
*/
final class Jsparse {
	// sparse.h
	static Jsm_col sm_get_col(final Jsm_matrix A, final int colnum) {
		return ((colnum >= 0 && colnum < A.cols_size) ? A.cols[ colnum ] : null);
	}

	static Jsm_row sm_get_row(final Jsm_matrix A, final int rownum) {
		return ((rownum >= 0 && rownum < A.rows_size) ? A.rows[ rownum ] : null);
	}

	// #define sm_foreach_row(A, prow)
	//	for( prow = A.first_row; prow != null; prow = prow.next_row )

	//#define sm_foreach_col(A, pcol)
	//	for( pcol = A.first_col; pcol != null; pcol = pcol.next_col)

	//#define sm_foreach_row_element(prow, p)
	//	for( p = prow.first_col; p != null; p = p.next_col )

	//#define sm_foreach_col_element(pcol, p)
	//	for( p = pcol.first_row; p != null; p = p.next_row )

	// #define sm_put(x, val)
	//	(x.user_word = (char *) val)

	//#define sm_get(type, x)
	//	((type) (x.user_word))
	// end sprse.h

	static Jset_family make_sparse(Jset_family F, final Jset_family D, final Jset_family R) {
		final Jcost_t cost = new Jcost_t();
		final Jcost_t best_cost = new Jcost_t();

		Jcvrmisc.cover_cost( F, best_cost );

		do {
			{
				final long t = System.currentTimeMillis();
				F = mv_reduce( F, D );
				Jcvrmisc.totals( t, Jespresso.MV_REDUCE_TIME, F, cost );
			}
			if( cost.total == best_cost.total) {
				break;
			}
			Jcvrmisc.copy_cost( cost, best_cost );

			{
				final long t = System.currentTimeMillis();
				F = Jexpand.expand( F, R, true );
				Jcvrmisc.totals( t, Jespresso.RAISE_IN_TIME, F, cost );
			}
			if( cost.total == best_cost.total ) {
				break;
			}
			Jcvrmisc.copy_cost( cost, best_cost );
		} while( Jglobals.force_irredundant );

		return F;
	}

	/**
	    mv_reduce -- perform an "optimal" reduction of the variables which
	    we desire to be sparse

	    This could be done using "reduce" and then saving just the desired
	    part of the reduction.  Instead, this version uses IRRED to find
	    which cubes of an output are redundant.  Note that this gets around
	    the cube-ordering problem.

	    In normal use, it is expected that the cover is irredundant and
	    hence no cubes will be reduced to the empty cube (however, this is
	    checked for and such cubes will be deleted)
	*/
	static Jset_family mv_reduce(Jset_family F, final Jset_family D) {
		/* loop for each multiple-valued variable */
		for( int var = 0; var < Jglobals.cube.num_vars; var++ ) {

			if( Jglobals.cube.sparse[var] ) {

				/* loop for each part of the variable */
				for( int i = Jglobals.cube.first_part[var]; i <= Jglobals.cube.last_part[var]; i++ ) {

					/* remember mapping of F1 cubes back to F cubes */
					Jset[] F_cube_table = new Jset[ F.count ];

					/* 'cofactor' against part #i of variable #var */
					final Jset_family F1 = Jespresso.new_cover( F.count );
					int[] data = F.data;// java
					for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
						if( Jespresso.is_in_set( data, p, i ) ) {
							F_cube_table[ F1.count ] = new Jset( data, p );
							final int p1 = F1.wsize * F1.count++;
							Jset.set_diff( F1.data, p1, data, p, Jglobals.cube.var_mask[ var ] );
							Jespresso.set_insert( F1.data, p1, i );
						}
					}

					/* 'cofactor' against part #i of variable #var */
					/* not really necessary -- just more efficient ? */
					final Jset_family D1 = Jespresso.new_cover( D.count );
					data = D.data;// java
					for( int p = 0, last = p + D.count * D.wsize; p < last; p += D.wsize ) {
						if( Jespresso.is_in_set( data, p, i ) ) {
							final int p1 = D1.wsize * D1.count++;
							Jset.set_diff( D1.data, p1, data, p, Jglobals.cube.var_mask[ var ] );
							Jespresso.set_insert( D1.data, p1, i );
						}
					}

					Jirred.mark_irredundant( F1, D1 );

					/* now remove part i from cubes which are redundant */
					int index = 0;
					data = F1.data;// java
					for( int p1 = 0, last = p1 + F1.count * F1.wsize; p1 < last; p1 += F1.wsize ) {
						if( ! Jespresso.TESTP( data, p1, Jespresso.ACTIVE ) ) {
							final Jset p = F_cube_table[ index ];

							/*   don't reduce a variable which is full
							*   (unless it is the output variable)
							*/
							if( var == Jglobals.cube.num_vars - 1 ||
									! Jset.setp_implies( Jglobals.cube.var_mask[ var ], p.base, p.offset ) ) {
								Jespresso.set_remove( p.base, p.offset, i );
							}
							Jespresso.RESET( p, Jespresso.PRIME );
						}
						index++;
					}

					Jespresso.free_cover( F1 );
					Jespresso.free_cover( D1 );
					F_cube_table = null;
				}
			}
		}

		/* Check if any cubes disappeared */
		Jset.sf_active( F );
		for( int var = 0; var < Jglobals.cube.num_vars; var++ ) {
			if( Jglobals.cube.sparse[ var ] ) {
				final int[] data = F.data;// java
				for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
					if( Jset.setp_disjoint( data, p, Jglobals.cube.var_mask[ var ] ) ) {
						Jespresso.RESET( data, p, Jespresso.ACTIVE );
						F.active_count--;
					}
				}
			}
		}

		if( F.count != F.active_count ) {
			F = Jset.sf_inactive( F );
		}
		return F;
	}
}