package espresso;

final class Jverify {

	/**
	 *  verify -- check that all minterms of F are contained in (Fold u Dold)
	 *  and that all minterms of Fold are contained in (F u Dold).
	 */
	static boolean verify(final Jset_family F, final Jset_family Fold, final Jset_family Dold ) {
		boolean verify_error = false;

		/* Make sure the function didn't grow too large */
		Jset[] FD = Jcofactor.cube2list( Fold, Dold );
		int[] data = F.data;// java
		for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			if( ! Jirred.cube_is_covered( FD, data, p ) ) {
				System.out.printf("some minterm in F is not covered by Fold u Dold\n");
				verify_error = true;
				if( Jglobals.verbose_debug ) {
					System.out.printf("%s\n", Jcvrout.pc1( data, p ) );
				} else {
					break;
				}
			}
		}

		/* Make sure minimized function covers the original function */
		FD = Jcofactor.cube2list( F, Dold );
		data = Fold.data;// java
		for( int p = 0, last = p + Fold.count * Fold.wsize; p < last; p += Fold.wsize ) {
			if( ! Jirred.cube_is_covered( FD, data, p ) ) {
				System.out.printf("some minterm in Fold is not covered by F u Dold\n");
				verify_error = true;
				if( Jglobals.verbose_debug ) {
					System.out.printf("%s\n", Jcvrout.pc1( data, p ) );
				} else {
					break;
				}
			}
		}
		FD = null;

		return verify_error;
	}

	/**
	 *  PLA_verify -- verify that two PLA's are identical
	 *
	 *  If names are given, row and column permutations are done to make
	 *  the comparison meaningful.
	 *
	 */
	static boolean PLA_verify(final JpPLA PLA1, final JpPLA PLA2 ) {
		/* Check if both have names given; if so, attempt to permute to
		* match the names
		*/
		if( PLA1.label != null && PLA1.label[0] != null &&
				PLA2.label != null && PLA2.label[0] != null ) {
			PLA_permute( PLA1, PLA2 );
		} else {
			System.err.printf("Warning: cannot permute columns without names\n");
			return true;
		}

		if( PLA1.F.sf_size != PLA2.F.sf_size ) {
			System.err.printf("PLA_verify: PLA's are not the same size\n");
			return true;
		}

		return verify( PLA2.F, PLA1.F, PLA1.D );
	}

	/**
	 *  Permute the columns of PLA1 so that they match the order of PLA2
	 *  Discard any columns of PLA1 which are not in PLA2
	 *  Association is strictly by the names of the columns of the cover.
	 */
	private static void PLA_permute(final JpPLA PLA1, final JpPLA PLA2 ) {
		/* determine which columns of PLA1 to save, and place these in the list
		 * "permute"; the order in this list is the final output order
		 */
		int npermute = 0;
		int[] permute = new int[ PLA2.F.sf_size ];
		for( int i = 0; i < PLA2.F.sf_size; i++ ) {
			final String labi = PLA2.label[ i ];
			for( int j = 0; j < PLA1.F.sf_size; j++ ) {
				if( labi.compareTo( PLA1.label[ j ] ) == 0 ) {
					permute[ npermute++ ] = j;
					break;
				}
			}
		}

		/* permute columns */
		if( PLA1.F != null ) {
			PLA1.F = Jset.sf_permute( PLA1.F, permute, npermute );
		}
		if( PLA1.R != null ) {
			PLA1.R = Jset.sf_permute( PLA1.R, permute, npermute );
		}
		if( PLA1.D != null ) {
			PLA1.D = Jset.sf_permute( PLA1.D, permute, npermute );
		}

		/* permute the labels */
		final Jcube_struct cube = Jglobals.cube;// java
		final String[] label = new String[ cube.size ];// java: nulled
		for( int i = 0; i < npermute; i++ ) {
			label[ i ] = PLA1.label[ permute[ i ] ];
		}
		/* for( int i = npermute; i < cube.size; i++ ) {
			label[ i ] = null;
		}*/
		// PLA1.label = null;
		PLA1.label = label;

		permute = null;
	}

	/**
	 *  check_consistency -- test that the ON-set, OFF-set and DC-set form
	 *  a partition of the boolean space.
	 */
	static boolean check_consistency(final JpPLA PLA ) {
		boolean verify_error = false;

		Jset_family T = Jsharp.cv_intersect( PLA.F, PLA.D );
		if( T.count == 0 ) {
			System.out.printf("ON-SET and DC-SET are disjoint\n");
		} else {
			System.out.printf("Some minterm(s) belong to both the ON-SET and DC-SET !\n");
			if( Jglobals.verbose_debug ) {
				Jcvrout.cprint( T );
			}
			verify_error = true;
		}
		System.out.flush();
		Jespresso.free_cover( T );

		T = Jsharp.cv_intersect( PLA.F, PLA.R );
		if( T.count == 0 ) {
			System.out.printf("ON-SET and OFF-SET are disjoint\n");
		} else {
			System.out.printf("Some minterm(s) belong to both the ON-SET and OFF-SET !\n");
			if( Jglobals.verbose_debug ) {
				Jcvrout.cprint( T );
			}
			verify_error = true;
		}
		System.out.flush();
		Jespresso.free_cover( T );

		T = Jsharp.cv_intersect( PLA.D, PLA.R );
		if( T.count == 0 ) {
			System.out.printf("DC-SET and OFF-SET are disjoint\n");
		} else {
			System.out.printf("Some minterm(s) belong to both the OFF-SET and DC-SET !\n");
			if( Jglobals.verbose_debug ) {
				Jcvrout.cprint( T );
			}
			verify_error = true;
		}
		System.out.flush();
		Jespresso.free_cover( T );

		if( Jirred.tautology( Jcofactor.cube3list( PLA.F, PLA.D, PLA.R ) ) != 0 ) {
			System.out.printf("Union of ON-SET, OFF-SET and DC-SET is the universe\n");
		} else {
			T = Jcompl.complement( Jcofactor.cube3list( PLA.F, PLA.D, PLA.R ) );
			System.out.printf("There are minterms left unspecified !\n");
			if( Jglobals.verbose_debug ) {
				Jcvrout.cprint( T );
			}
			verify_error = true;
			Jespresso.free_cover( T );
		}
		System.out.flush();
		return verify_error;
	}
}